@@ -46,8 +46,13 @@ namespace System.Windows.Forms {
 #if NET_2_0
 		private SizeF		auto_scale_dimensions;
 		private AutoScaleMode	auto_scale_mode;
+		private bool		auto_scale_mode_set;
+		private bool		auto_scale_pending;
+		private bool		is_auto_scaling;
 #endif
 
+		internal bool validation_failed; //track whether validation was cancelled by a validating control
+
 		#region Public Constructors
 		public ContainerControl() {
 			active_control = null;
@@ -55,7 +60,7 @@ namespace System.Windows.Forms {
 			ControlRemoved += new ControlEventHandler(OnControlRemoved);
 #if NET_2_0
 			auto_scale_dimensions = SizeF.Empty;
-			auto_scale_mode = AutoScaleMode.None;
+			auto_scale_mode = AutoScaleMode.Inherit;
 #endif
 		}
 		#endregion	// Public Constructors
@@ -79,89 +84,95 @@ namespace System.Windows.Forms {
 
 				// Fire the enter and leave events if possible
 				Form form = FindForm ();
-				if (form != null) {
-					Control active = GetMostDeeplyNestedActiveControl (form);
-					Control common_container = GetCommonContainer (active, value);
-					ArrayList chain = new ArrayList ();
-					ArrayList validation_chain = new ArrayList ();
-					Control walk = active;
+				Control active = GetMostDeeplyNestedActiveControl (form == null ? this : form);
+				Control common_ancestor = GetCommonContainer (active, value);
+				ArrayList chain = new ArrayList ();
+				ArrayList validation_chain = new ArrayList ();
+				Control walk = active;
 
-					// we split this up into three steps:
-					//    1. walk up the tree (if we need to) to our root, firing leave events.
-					//    2. validate.
-					//    3. walk down the tree (if we need to), firing enter events.
+				// we split this up into three steps:
+				//    1. walk up the tree (if we need to) to our root, firing leave events.
+				//    2. validate.
+				//    3. walk down the tree (if we need to), firing enter events.
 
-					// "our root" is either the common container of the current active
-					// control and the new active control, or the current active control,
-					// or the new active control.  That is, it's either one of these three
-					// configurations:
+				// "our root" is either the common ancestor of the current active
+				// control and the new active control, or the current active control,
+				// or the new active control.  That is, it's either one of these three
+				// configurations:
 
-					//  (1)     root	    (2)  new   	      (3)  current
-					//          /  \	       	/   \		    /  	\
-					//	  ...  	...	      ...   ...		  ...  	...
-					//     	  /      \	      /	       			  \
-					//     current 	 new   	   current   			  new
+				//  (1)     root	    (2)  new   	      (3)  current
+				//          /  \	       	/   \		    /  	\
+				//	  ...  	...	      ...   ...		  ...  	...
+				//     	  /      \	      /	       			  \
+				//     current 	 new   	   current   			  new
 
 
-					// note (3) doesn't require any upward walking, and no leave events are generated.
-					//      (2) doesn't require any downward walking, and no enter events are generated.
+				// note (3) doesn't require any upward walking, and no leave events are generated.
+				//      (2) doesn't require any downward walking, and no enter events are generated.
 
-					// as we walk up the tree, we generate a list of all controls which cause
-					// validation.  After firing the leave events, we invoke (in order starting from
-					// the most deeply nested) their Validating event.  If one sets CancelEventArgs.Cancel
-					// to true, we ignore the control the user wanted to set ActiveControl to, and use
-					// the Validating control instead.
+				// as we walk up the tree, we generate a list of all controls which cause
+				// validation.  After firing the leave events, we invoke (in order starting from
+				// the most deeply nested) their Validating event.  If one sets CancelEventArgs.Cancel
+				// to true, we ignore the control the user wanted to set ActiveControl to, and use
+				// the Validating control instead.
 
-					bool fire_enter = true;
-					Control root = common_container;
+				bool fire_enter = true;
+				Control root = common_ancestor;
 
-					// Generate the leave messages
-					while (walk != common_container) {
-						if (walk == value) {
-							root = value;
-							fire_enter = false;
-							break;
-						}
-						walk.FireLeave ();
-						/* clear our idea of the active control as we go back up */
-						if (walk is ContainerControl)
-							((ContainerControl)walk).active_control = null;
+				active_control = value;
 
-						if (walk.CausesValidation)
-							validation_chain.Add (walk);
-
-						walk = walk.Parent;
+				// Generate the leave messages
+				while (walk != common_ancestor && walk != null) {
+					if (walk == value) {
+						root = value;
+						fire_enter = false;
+						break;
 					}
+					walk.FireLeave ();
+					/* clear our idea of the active control as we go back up */
+					if (walk is ContainerControl)
+						((ContainerControl)walk).active_control = null;
 
-					for (int i = 0; i < validation_chain.Count; i ++) {
-						if (!ValidateControl ((Control)validation_chain[i])) {
-							value = (Control)validation_chain[i];
-							fire_enter = true;
-						}
-					}
+					if (walk.CausesValidation)
+						validation_chain.Add (walk);
 
-					if (fire_enter) {
-						walk = value;
-						while (walk != root) {
-							chain.Add (walk);
-							walk = walk.Parent;
-						}
+					walk = walk.Parent;
+				}
 
-						for (int i = chain.Count - 1; i >= 0; i--) {
-							walk = (Control) chain [i];
-							walk.FireEnter ();
-						}
-					}
-
-					walk = this;
-					while (walk != null) {
-						if (walk.Parent is ContainerControl)
-							((ContainerControl)walk.Parent).active_control = walk;
-						walk = walk.Parent;
+				validation_failed = false;
+				for (int i = 0; i < validation_chain.Count; i ++) {
+					if (!ValidateControl ((Control)validation_chain[i])) {
+						active_control = value = (Control)validation_chain[i];
+						fire_enter = true;
+						validation_failed = true;
 					}
 				}
 
-				active_control = value;
+				if (fire_enter) {
+					walk = value;
+					while (walk != root && walk != null) {
+						chain.Add (walk);
+						walk = walk.Parent;
+					}
+
+					if (root != null && walk == root && !(root is ContainerControl))
+						chain.Add (walk);
+
+					for (int i = chain.Count - 1; i >= 0; i--) {
+						walk = (Control) chain [i];
+						walk.FireEnter ();
+					}
+				}
+
+				walk = this;
+				Control ctl = this;
+				while (walk != null) {
+					if (walk.Parent is ContainerControl) {
+						((ContainerControl) walk.Parent).active_control = ctl;
+						ctl = walk.Parent;
+					}
+					walk = walk.Parent;
+				}
 
 				if (this is Form)
 					CheckAcceptButton();
@@ -170,7 +181,7 @@ namespace System.Windows.Forms {
 				ScrollControlIntoView(active_control);
 
 				// Let the control know it's selected
-				SendControlFocus (value);
+				SendControlFocus (active_control);
 			}
 		}
 
@@ -237,22 +248,25 @@ namespace System.Windows.Forms {
 			}
 
 			set {
-				auto_scale_dimensions = value;
+				if (auto_scale_dimensions != value) {
+					auto_scale_dimensions = value;
+					
+					PerformAutoScale ();
+				}
 			}
 		}
 
 		protected SizeF AutoScaleFactor {
 			get {
-				if (auto_scale_dimensions.IsEmpty) {
-					return new SizeF(1f, 1f);
-				}
+				if (auto_scale_dimensions.IsEmpty)
+					return new SizeF (1f, 1f);
+
 				return new SizeF(CurrentAutoScaleDimensions.Width / auto_scale_dimensions.Width,
 					CurrentAutoScaleDimensions.Height / auto_scale_dimensions.Height);
 			}
 		}
 
 
-		[MonoTODO("Call scaling method")]
 		[Browsable (false)]
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
@@ -261,10 +275,18 @@ namespace System.Windows.Forms {
 				return auto_scale_mode;
 			}
 			set {
+				if (this is Form)
+					(this as Form).AutoScale = false;
+
 				if (auto_scale_mode != value) {
 					auto_scale_mode = value;
 
-					// Trigger scaling
+					if (auto_scale_mode_set)
+						auto_scale_dimensions = SizeF.Empty;
+
+					auto_scale_mode_set = true;
+
+					PerformAutoScale ();
 				}
 			}
 		}
@@ -285,30 +307,19 @@ namespace System.Windows.Forms {
 		}
 
 #if NET_2_0
-		[MonoTODO("Revisit when System.Drawing.GDI.WindowsGraphics.GetTextMetrics is done or come up with other cross-plat avg. font width calc method")]
 		[Browsable (false)]
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
 		public SizeF CurrentAutoScaleDimensions {
 			get {
 				switch(auto_scale_mode) {
-					case AutoScaleMode.Dpi: {
-						Bitmap		bmp;
-						Graphics	g;
-						SizeF		size;
+					case AutoScaleMode.Dpi:
+						return TextRenderer.GetDpi ();
 
-						bmp = new Bitmap(1, 1, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
-						g = Graphics.FromImage(bmp);
-						size = new SizeF(g.DpiX, g.DpiY);
-						g.Dispose();
-						bmp.Dispose();
-						return size;
-					}
-
-					case AutoScaleMode.Font: {
-						// http://msdn2.microsoft.com/en-us/library/system.windows.forms.containercontrol.currentautoscaledimensions(VS.80).aspx
-						// Implement System.Drawing.GDI.WindowsGraphics.GetTextMetrics first...
-						break;
-					}
+					case AutoScaleMode.Font:
+						Size s = TextRenderer.MeasureText ("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890", Font);
+						int width = (int)Math.Round ((float)s.Width / 62f);
+						
+						return new SizeF (width, s.Height);
 				}
 
 				return auto_scale_dimensions;
@@ -337,6 +348,11 @@ namespace System.Windows.Forms {
 		#endregion	// Public Instance Properties
 
 		#region Protected Instance Methods
+#if NET_2_0
+		protected override bool CanEnableIme {
+			get { return false; }
+		}
+#endif
 		protected override CreateParams CreateParams {
 			get {
 				return base.CreateParams;
@@ -345,6 +361,53 @@ namespace System.Windows.Forms {
 		#endregion	// Public Instance Methods
 
 		#region Public Instance Methods
+#if NET_2_0
+		internal void PerformAutoScale (bool called_by_scale)
+		{
+			if ((AutoScaleMode == AutoScaleMode.Inherit) && !called_by_scale)
+				return;
+
+			if ((layout_suspended > 0) && !called_by_scale) {
+				auto_scale_pending = true;
+				return;
+			}
+			// Set this first so we don't get called again from
+			// PerformDelayedAutoScale after ResumeLayout
+			auto_scale_pending = false;
+
+			SizeF factor = AutoScaleFactor;
+			if (AutoScaleMode == AutoScaleMode.Inherit) {
+				ContainerControl cc = FindContainer (this.Parent);
+				if (cc != null)
+					factor = cc.AutoScaleFactor;
+			}
+			if (factor != new SizeF (1F, 1F)) {
+				is_auto_scaling = true;
+				SuspendLayout ();
+				Scale (factor);
+				ResumeLayout (false);
+				is_auto_scaling = false;
+			}
+
+			auto_scale_dimensions = CurrentAutoScaleDimensions;
+		}
+
+		public void PerformAutoScale ()
+		{
+			PerformAutoScale (false);
+		}
+
+		internal void PerformDelayedAutoScale ()
+		{
+			if (auto_scale_pending)
+				PerformAutoScale ();
+		}
+
+		internal bool IsAutoScaling {
+			get { return is_auto_scaling; }
+		}
+#endif
+
 		[MonoTODO]
 		static bool ValidateWarned;
 		public bool Validate() {
@@ -356,6 +419,36 @@ namespace System.Windows.Forms {
 			return true;
 		}
 
+#if NET_2_0
+		public bool Validate (bool checkAutoValidate)
+		{
+			if ((checkAutoValidate && (AutoValidate != AutoValidate.Disable)) || !checkAutoValidate)
+				return Validate ();
+				
+			return true;
+		}
+		
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public virtual bool ValidateChildren ()
+		{
+			return ValidateChildren (ValidationConstraints.Selectable);
+		}
+
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public virtual bool ValidateChildren (ValidationConstraints validationConstraints)
+		{
+			bool recurse = !((validationConstraints & ValidationConstraints.ImmediateChildren) == ValidationConstraints.ImmediateChildren);
+			
+			foreach (Control control in Controls)
+				if (!ValidateNestedControls (control, validationConstraints, recurse))
+					return false;
+
+			return true;
+		}
+#endif
+
 		bool IContainerControl.ActivateControl(Control control) {
 			return Select(control);
 		}
@@ -422,8 +515,10 @@ namespace System.Windows.Forms {
 
 			switch (key) {
 				case Keys.Tab: {
-					if (ProcessTabKey((Control.ModifierKeys & Keys.Shift) == 0)) {
-						return true;
+					if ((keyData & (Keys.Alt | Keys.Control)) == Keys.None) {
+						if (ProcessTabKey ((Control.ModifierKeys & Keys.Shift) == 0)) {
+							return true;
+						}
 					}
 					break;
 				}
@@ -460,11 +555,19 @@ namespace System.Windows.Forms {
 			wrapped = false;
 			c = active_control;
 
+#if NET_2_0
+			System.Collections.Generic.List<MenuStrip> strips = new System.Collections.Generic.List<MenuStrip> ();
+#endif
+
 			do {
 				c = GetNextControl(c, true);
+#if NET_2_0
+				if (c is MenuStrip)
+					strips.Add ((MenuStrip)c);
+#endif
 				if (c != null) {
 					// This is stupid. I want to be able to call c.ProcessMnemonic directly
-					if (c.CanSelect && c.ProcessControlMnemonic(charCode)) {
+					if (c.ProcessControlMnemonic(charCode)) {
 						return(true);
 					}
 					continue;
@@ -476,6 +579,14 @@ namespace System.Windows.Forms {
 				}
 			} while (c != active_control);
 
+#if NET_2_0
+			// No one has an explicit mnemonic for this key.
+			// Let MenuStrips have a chance at implicit mnemonics.
+			foreach (MenuStrip ms in strips)
+				if (ms.ProcessImplicitMnemonic (charCode))
+					return true;
+#endif
+
 			return false;
 		}
 
@@ -526,9 +637,9 @@ namespace System.Windows.Forms {
 		#region Internal Methods
 		internal void ChildControlRemoved (Control control)
 		{
-			if (control == active_control) {
+			if (control == active_control || control.Contains (active_control)) {
 				SelectNextControl (this, true, true, true, true);
-				if (control == active_control) {
+				if (control == active_control || control.Contains (active_control)) {
 					active_control = null;
 				}
 			}
@@ -537,6 +648,47 @@ namespace System.Windows.Forms {
 		{
 			// do nothing here, only called if it is a Form
 		}
+
+#if NET_2_0
+		private bool ValidateNestedControls (Control c, ValidationConstraints constraints, bool recurse)
+		{
+			bool validate_result = true;
+
+			if (!c.CausesValidation)
+				validate_result = true;
+			else if (!ValidateThisControl (c, constraints))
+				validate_result = true;
+			else if (!ValidateControl (c))
+				validate_result = false;
+
+			if (recurse)
+				foreach (Control control in c.Controls)
+					if (!ValidateNestedControls (control, constraints, recurse))
+						return false;
+
+			return validate_result;
+		}
+
+		private bool ValidateThisControl (Control c, ValidationConstraints constraints)
+		{
+			if (constraints == ValidationConstraints.None)
+				return true;
+
+			if ((constraints & ValidationConstraints.Enabled) == ValidationConstraints.Enabled && !c.Enabled)
+				return false;
+
+			if ((constraints & ValidationConstraints.Selectable) == ValidationConstraints.Selectable && !c.GetStyle (ControlStyles.Selectable))
+				return false;
+
+			if ((constraints & ValidationConstraints.TabStop) == ValidationConstraints.TabStop && !c.TabStop)
+				return false;
+
+			if ((constraints & ValidationConstraints.Visible) == ValidationConstraints.Visible && !c.Visible)
+				return false;
+
+			return true;
+		}
+#endif
 		#endregion	// Internal Methods
 
 #if NET_2_0
@@ -549,11 +701,14 @@ namespace System.Windows.Forms {
 		protected override void OnFontChanged (EventArgs e)
 		{
 			base.OnFontChanged (e);
+			
+			if (AutoScaleMode == AutoScaleMode.Font)
+				PerformAutoScale ();
 		}
 
-		protected override void OnLayout (LayoutEventArgs levent)
+		protected override void OnLayout (LayoutEventArgs e)
 		{
-			base.OnLayout (levent);
+			base.OnLayout (e);
 		}
 		
 		AutoValidate auto_validate = AutoValidate.Inherit;
@@ -575,6 +730,11 @@ namespace System.Windows.Forms {
 			}
 		}
 
+		internal bool ShouldSerializeAutoValidate ()
+		{
+			return this.AutoValidate != AutoValidate.Inherit;
+		}
+
 		static object OnValidateChanged = new object ();
 
 		protected virtual void OnAutoValidateChanged (EventArgs e)
