@@ -29,12 +29,14 @@
 using System.ComponentModel;
 using System.Drawing;
 using System.Runtime.InteropServices;
+using System.Collections;
+using System.Collections.Generic;
 
-namespace System.Windows.Forms {
-
-	// XXX [TypeConverter (typeof (DataGridRowConverter))]
-	public class DataGridViewRow : DataGridViewBand {
-
+namespace System.Windows.Forms
+{
+	[TypeConverter (typeof (DataGridViewRowConverter))]
+	public class DataGridViewRow : DataGridViewBand
+	{
 		private AccessibleObject accessibilityObject;
 		private DataGridViewCellCollection cells;
 		private ContextMenuStrip contextMenuStrip;
@@ -51,6 +53,9 @@ namespace System.Windows.Forms {
 			minimumHeight = 3;
 			height = -1;
 			headerCell = new DataGridViewRowHeaderCell();
+			headerCell.SetOwningRow (this);
+			accessibilityObject = new AccessibleObject ();
+			SetState (DataGridViewElementStates.Visible);
 		}
 
 		[Browsable (false)]
@@ -66,7 +71,12 @@ namespace System.Windows.Forms {
 
 		[DefaultValue (null)]
 		public override ContextMenuStrip ContextMenuStrip {
-			get { return contextMenuStrip; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Operation cannot be performed on a shared row.");
+					
+				return contextMenuStrip;
+			}
 			set {
 				if (contextMenuStrip != value) {
 					contextMenuStrip = value;
@@ -100,7 +110,12 @@ namespace System.Windows.Forms {
 
 		[Browsable (false)]
 		public override bool Displayed {
-			get { return base.Displayed; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the Displayed property of a shared row is not a valid operation.");
+					
+				return base.Displayed;
+			}
 		}
 
 		[DefaultValue (0)]
@@ -113,7 +128,12 @@ namespace System.Windows.Forms {
 		[DefaultValue ("")]
 		[NotifyParentProperty (true)]
 		public string ErrorText {
-			get { return errorText; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Operation cannot be performed on a shared row.");
+					
+				return errorText == null ? string.Empty : errorText;
+			}
 			set {
 				if (errorText != value) {
 					errorText = value;
@@ -126,7 +146,12 @@ namespace System.Windows.Forms {
 
 		[Browsable (false)]
 		public override bool Frozen {
-			get { return base.Frozen; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the Frozen property of a shared row is not a valid operation.");
+					
+				return base.Frozen;
+			}
 			set { base.Frozen = value; }
 		}
 
@@ -137,7 +162,10 @@ namespace System.Windows.Forms {
 			set {
 				if (headerCell != value) {
 					headerCell = value;
+					headerCell.SetOwningRow (this);
+					
 					if (DataGridView != null) {
+						headerCell.SetDataGridView (DataGridView);
 						DataGridView.OnRowHeaderCellChanged(new DataGridViewRowEventArgs(this));
 					}
 				}
@@ -152,7 +180,7 @@ namespace System.Windows.Forms {
 					if (DefaultCellStyle != null && DefaultCellStyle.Font != null) {
 						return DefaultCellStyle.Font.Height + 9;
 					}
-					if (InheritedStyle != null && InheritedStyle.Font != null) {
+					if (Index >= 0 && InheritedStyle != null && InheritedStyle.Font != null) {
 						return InheritedStyle.Font.Height + 9;
 					}
 					return System.Windows.Forms.Control.DefaultFont.Height + 9;
@@ -174,6 +202,9 @@ namespace System.Windows.Forms {
 
 		public override DataGridViewCellStyle InheritedStyle {
 			get {
+				if (Index == -1)
+					throw new InvalidOperationException ("Getting the InheritedStyle property of a shared row is not a valid operation.");
+					
 				if (DataGridView == null) {
 					return DefaultCellStyle;
 				}
@@ -194,13 +225,19 @@ namespace System.Windows.Forms {
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public bool IsNewRow {
 			get {
-				if (DataGridView != null && DataGridView.Rows[DataGridView.Rows.Count - 1] == this) {
+				if (DataGridView != null && DataGridView.Rows[DataGridView.Rows.Count - 1] == this && DataGridView.NewRowIndex == Index) {
 					return true;
 				}
 				return false;
 			}
 		}
 
+		internal bool IsShared {
+			get {
+				return Index == -1 && DataGridView != null;
+			}
+		}
+
 		[Browsable (false)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public int MinimumHeight {
@@ -222,24 +259,34 @@ namespace System.Windows.Forms {
 		[DefaultValue (false)]
 		[NotifyParentProperty (true)]
 		public override bool ReadOnly {
-			get { return base.ReadOnly; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the ReadOnly property of a shared row is not a valid operation.");
+					
+				if (DataGridView != null && DataGridView.ReadOnly)
+					return true;
+					
+				return base.ReadOnly;
+			}
 			set { base.ReadOnly = value; }
 		}
 
 		[NotifyParentProperty (true)]
 		public override DataGridViewTriState Resizable {
-			get { return base.Resizable; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the Resizable property of a shared row is not a valid operation.");
+					
+				return base.Resizable;
+			}
 			set { base.Resizable = value; }
 		}
 
 		public override bool Selected {
 			get {
-				if (Index == -1) {
-					throw new InvalidOperationException("The row is a shared row.");
-				}
-				if (DataGridView == null) {
-					throw new InvalidOperationException("The row has not been added to a DataGridView control.");
-				}
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the Selected property of a shared row is not a valid operation.");
+					
 				return base.Selected;
 			}
 			set {
@@ -250,19 +297,26 @@ namespace System.Windows.Forms {
 					throw new InvalidOperationException("The row has not been added to a DataGridView control.");
 				}
 				base.Selected = value;
-				foreach (DataGridViewCell cell in cells) {
-					cell.Selected = value;
-				}
 			}
 		}
 
 		public override DataGridViewElementStates State {
-			get { return base.State; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the State property of a shared row is not a valid operation.");
+					
+				return base.State;
+			}
 		}
 
 		[Browsable (false)]
 		public override bool Visible {
-			get { return base.Visible; }
+			get {
+				if (IsShared)
+					throw new InvalidOperationException ("Getting the Visible property of a shared row is not a valid operation.");
+					
+				return base.Visible;
+			}
 			set {
 				if (IsNewRow && value == false) {
 					throw new InvalidOperationException("Cant make invisible a new row.");
@@ -279,11 +333,18 @@ namespace System.Windows.Forms {
 
 		public override object Clone ()
 		{
-			DataGridViewRow row = (DataGridViewRow) MemberwiseClone();
-			row.cells = new DataGridViewCellCollection(row);
-			foreach (DataGridViewCell cell in cells) {
-				row.cells.Add(cell.Clone() as DataGridViewCell);
-			}
+			DataGridViewRow row = (DataGridViewRow)MemberwiseClone ();
+
+			row.HeaderCell = (DataGridViewRowHeaderCell)HeaderCell.Clone ();
+			row.SetIndex (-1);
+			
+			row.cells = new DataGridViewCellCollection (row);
+			
+			foreach (DataGridViewCell cell in cells)
+				row.cells.Add (cell.Clone () as DataGridViewCell);
+
+			row.SetDataGridView (null);
+
 			return row;
 		}
 
@@ -330,18 +391,62 @@ namespace System.Windows.Forms {
 
 		public string GetErrorText (int rowIndex)
 		{
-			return "";
+			return string.Empty;
 		}
 
 		public virtual int GetPreferredHeight (int rowIndex, DataGridViewAutoSizeRowMode autoSizeRowMode, bool fixedWidth)
 		{
-			throw new NotImplementedException();
+			DataGridViewRow row;
+			
+			if (DataGridView != null)
+				row = DataGridView.Rows.SharedRow (rowIndex);
+			else
+				row = this;
+				
+			int height = 0;
+			
+			foreach (DataGridViewCell cell in row.Cells)
+				height = Math.Max (height, cell.PreferredSize.Height);
+			
+			return height;
 		}
 
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
 		public virtual DataGridViewElementStates GetState (int rowIndex)
 		{
-			throw new NotImplementedException();
+			DataGridViewElementStates state = DataGridViewElementStates.None;
+			
+			if (rowIndex == -1) {
+				state |= DataGridViewElementStates.Displayed;
+				
+				if (DataGridView.ReadOnly)
+					state |= DataGridViewElementStates.ReadOnly;
+				if (DataGridView.AllowUserToResizeRows)
+					state |= DataGridViewElementStates.Resizable;
+				if (DataGridView.Visible)
+					state |= DataGridViewElementStates.Visible;
+			
+				return state;
+			}
+			
+			DataGridViewRow row = DataGridView.Rows[rowIndex];
+
+			if (row.Displayed)
+				state |= DataGridViewElementStates.Displayed;
+			if (row.Frozen)
+				state |= DataGridViewElementStates.Frozen;
+			if (row.ReadOnly)
+				state |= DataGridViewElementStates.ReadOnly;
+			if (row.Resizable == DataGridViewTriState.True || (row.Resizable == DataGridViewTriState.NotSet && DataGridView.AllowUserToResizeRows))
+				state |= DataGridViewElementStates.Resizable;
+			if (row.Resizable == DataGridViewTriState.True)
+				state |= DataGridViewElementStates.ResizableSet;
+			if (row.Selected)
+				state |= DataGridViewElementStates.Selected;
+			if (row.Visible)
+				state |= DataGridViewElementStates.Visible;
+				
+			return state;
 		}
 
 		public bool SetValues (params object[] values)
@@ -354,9 +459,14 @@ namespace System.Windows.Forms {
 			}
 			/////// COLUMNAS //////////
 			for (int i = 0; i < values.Length; i++) {
-				DataGridViewCell cell = new DataGridViewTextBoxCell();
+				DataGridViewCell cell;
+				if (cells.Count > i) {
+					cell = cells [i];
+				} else {
+					cell = new DataGridViewTextBoxCell ();
+					cells.Add (cell);
+				}
 				cell.Value = values[i];
-				cells.Add(cell);
 			}
 			
 			// XXX
@@ -387,22 +497,108 @@ namespace System.Windows.Forms {
 
 		protected internal virtual void Paint (Graphics graphics, Rectangle clipBounds, Rectangle rowBounds, int rowIndex, DataGridViewElementStates rowState, bool isFirstDisplayedRow, bool isLastVisibleRow)
 		{
+			DataGridViewCellStyle style;
+			
+			if (Index == -1)
+				style = DataGridView.RowsDefaultCellStyle;
+			else
+				style = InheritedStyle;
+				
+			DataGridViewRowPrePaintEventArgs pre = new DataGridViewRowPrePaintEventArgs (DataGridView, graphics, clipBounds, rowBounds, rowIndex, rowState, string.Empty, style, isFirstDisplayedRow, isLastVisibleRow);
+			pre.PaintParts = DataGridViewPaintParts.All;
+
+			DataGridView.OnRowPrePaint (pre);
+
+			// The user has elected for us to not do anything
+			if (pre.Handled)
+				return;
+
+			if (DataGridView.RowHeadersVisible)
+				PaintHeader (graphics, pre.ClipBounds, rowBounds, rowIndex, rowState, isFirstDisplayedRow, isLastVisibleRow, pre.PaintParts);
+			
+			PaintCells (graphics, pre.ClipBounds, rowBounds, rowIndex, rowState, isFirstDisplayedRow, isLastVisibleRow, pre.PaintParts);
+
+			DataGridViewRowPostPaintEventArgs post = new DataGridViewRowPostPaintEventArgs (DataGridView, graphics, pre.ClipBounds, rowBounds, rowIndex, rowState, pre.ErrorText, style, isFirstDisplayedRow, isLastVisibleRow);
+			DataGridView.OnRowPostPaint (post);
 		}
 
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
 		protected internal virtual void PaintCells (Graphics graphics, Rectangle clipBounds, Rectangle rowBounds, int rowIndex, DataGridViewElementStates rowState, bool isFirstDisplayedRow, bool isLastVisibleRow, DataGridViewPaintParts paintParts)
 		{
+			List<DataGridViewColumn> sortedColumns = DataGridView.Columns.ColumnDisplayIndexSortedArrayList;
+			
+			Rectangle bounds = rowBounds;
+			
+			// If row headers are visible, adjust our starting point
+			if (DataGridView.RowHeadersVisible) {
+				bounds.X += DataGridView.RowHeadersWidth;
+				bounds.Width -= DataGridView.RowHeadersWidth;
+			}
+			
+			for (int i = DataGridView.first_col_index; i < sortedColumns.Count; i++) {
+				DataGridViewColumn col = sortedColumns[i];
+				
+				if (!col.Visible)
+					continue;
+					
+				if (!col.Displayed)
+					break;
+					
+				bounds.Width = col.Width;
+				DataGridViewCell cell = Cells[col.Index];
+				
+				graphics.FillRectangle (Brushes.White, bounds);
+				
+				DataGridViewCellStyle style;
+
+				if (cell.RowIndex == -1)
+					style = DefaultCellStyle;
+				else
+					style = cell.InheritedStyle;
+
+				object value;
+				object formattedValue;
+				string errorText;
+				DataGridViewElementStates cellState;
+				
+				if (cell.RowIndex == -1) {
+					// TODO: Look up value if databound.
+					value = null;
+					formattedValue = null;
+					errorText = null;
+					cellState = cell.State;
+				} else {
+					value = cell.Value;
+					formattedValue = cell.FormattedValue;
+					errorText = cell.ErrorText;
+					cellState = cell.InheritedState;
+				}
+
+				DataGridViewAdvancedBorderStyle intermediateBorderStyle = (DataGridViewAdvancedBorderStyle)((ICloneable)DataGridView.AdvancedCellBorderStyle).Clone ();
+				DataGridViewAdvancedBorderStyle borderStyle = cell.AdjustCellBorderStyle (DataGridView.AdvancedCellBorderStyle, intermediateBorderStyle, true, true, cell.ColumnIndex == 0, cell.RowIndex == 0);
+				DataGridView.OnCellFormattingInternal (new DataGridViewCellFormattingEventArgs (cell.ColumnIndex, cell.RowIndex, value, cell.FormattedValueType, style));
+
+
+				cell.PaintWork (graphics, clipBounds, bounds, rowIndex, cellState, style, borderStyle, paintParts);
+				bounds.X += bounds.Width;
+			}
 		}
 
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
 		protected internal virtual void PaintHeader (Graphics graphics, Rectangle clipBounds, Rectangle rowBounds, int rowIndex, DataGridViewElementStates rowState, bool isFirstDisplayedRow, bool isLastVisibleRow, DataGridViewPaintParts paintParts)
 		{
+			rowBounds.Width = DataGridView.RowHeadersWidth;
+			graphics.FillRectangle (Brushes.White, rowBounds);
+	
+			HeaderCell.PaintWork (graphics, clipBounds, rowBounds, rowIndex, rowState, HeaderCell.InheritedStyle, DataGridView.AdvancedRowHeadersBorderStyle, paintParts);
 		}
 
 		internal override void SetDataGridView (DataGridView dataGridView)
 		{
 			base.SetDataGridView(dataGridView);
 			headerCell.SetDataGridView(dataGridView);
+			foreach (DataGridViewCell cell in cells)
+				cell.SetDataGridView (dataGridView);
 		}
 
 		internal override void SetState (DataGridViewElementStates state)
@@ -424,19 +620,15 @@ namespace System.Windows.Forms {
 			{
 			}
 
-			public DataGridViewRowAccessibleObject (DataGridViewRow row)
+			public DataGridViewRowAccessibleObject (DataGridViewRow owner)
 			{
-				this.dataGridViewRow = row;
+				this.dataGridViewRow = owner;
 			}
 
 			public override Rectangle Bounds {
 				get { throw new NotImplementedException(); }
 			}
 
-			public override string DefaultAction {
-				get { return "Edit"; }
-			}
-
 			public override string Name {
 				get { return "Index: " + dataGridViewRow.Index.ToString(); }
 			}
@@ -530,13 +722,12 @@ namespace System.Windows.Forms {
 						break;
 				}
 			}
-
 		}
-
-
-
 	}
-
+	
+	internal class DataGridViewRowConverter : TypeConverter
+	{
+	}
 }
 
 #endif
