@@ -33,6 +33,7 @@ using System.Drawing;
 using System.Drawing.Text;
 using System.Drawing.Imaging;
 using System.Runtime.InteropServices;
+using System.Windows.Forms.Theming;
 
 namespace System.Windows.Forms
 {
@@ -41,20 +42,21 @@ namespace System.Windows.Forms
 #if NET_2_0
 	[ClassInterface (ClassInterfaceType.AutoDispatch)]
 	[ComVisible (true)]
-	[ToolboxItem (true)]
+	[ToolboxItem ("System.Windows.Forms.Design.AutoSizeToolboxItem," + Consts.AssemblySystem_Design)]
 	[DefaultBindingProperty ("Text")]
 #endif
 	public class Label : Control
 	{
 		private bool autosize;
-#if NET_2_0
 		private bool auto_ellipsis;
-#endif
 		private Image image;
 		private bool render_transparent;
 		private FlatStyle flat_style;
 		private bool use_mnemonic;
 		private int image_index = -1;
+#if NET_2_0
+		private string image_key = string.Empty;
+#endif
 		private ImageList image_list;
 		internal ContentAlignment image_align;
 		internal StringFormat string_format;
@@ -82,6 +84,15 @@ namespace System.Windows.Forms
 			remove { base.BackgroundImageChanged -= value; }
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public new event EventHandler BackgroundImageLayoutChanged {
+			add { base.BackgroundImageLayoutChanged += value; }
+			remove { base.BackgroundImageLayoutChanged -= value; }
+		}
+#endif
+
 		[Browsable(false)]
 		[EditorBrowsable(EditorBrowsableState.Never)]
 		public new event EventHandler ImeModeChanged {
@@ -129,6 +140,7 @@ namespace System.Windows.Forms
 			autosize = false;
 			TabStop = false;
 			string_format = new StringFormat();
+			string_format.FormatFlags = StringFormatFlags.LineLimit;
 			TextAlign = ContentAlignment.TopLeft;
 			image = null;
 			UseMnemonic = true;
@@ -154,12 +166,15 @@ namespace System.Windows.Forms
 
 		#region Public Properties
 
-#if NET_2_0
 		[DefaultValue (false)]
 		[Browsable (true)]
 		[EditorBrowsable (EditorBrowsableState.Always)]
-		public bool AutoEllipsis
-		{
+#if NET_2_0
+		public
+#else
+		internal
+#endif
+		bool AutoEllipsis {
 			get { return this.auto_ellipsis; }
 			set
 			{
@@ -171,31 +186,42 @@ namespace System.Windows.Forms
 					else
 						string_format.Trimming = StringTrimming.Character;
 
+					if (Parent != null)
+						Parent.PerformLayout (this, "AutoEllipsis");
 					this.Invalidate ();
 				}
 			}
 		}
-#endif
 
-		[DefaultValue(false)]
-		[Localizable(true)]
-		[RefreshProperties(RefreshProperties.All)]
 #if NET_2_0
 		[Browsable (true)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Visible)]
 		[EditorBrowsable (EditorBrowsableState.Always)]
 #endif
-		public new virtual bool AutoSize {
+		[DefaultValue(false)]
+		[Localizable(true)]
+		[RefreshProperties(RefreshProperties.All)]
+		public
+#if NET_2_0
+		override
+#else
+		virtual
+#endif
+		bool AutoSize {
 			get { return autosize; }
 			set {
 				if (autosize == value)
 					return;
 
+#if NET_2_0
+				base.SetAutoSizeMode (AutoSizeMode.GrowAndShrink);
+				base.AutoSize = value;
+#endif
 				autosize = value;
 				CalcAutoSize ();
-				Refresh ();
+				Invalidate ();
 
-				OnAutoSizeChanged (EventArgs.Empty);					
+				OnAutoSizeChanged (EventArgs.Empty);
 			}
 		}
 
@@ -206,10 +232,19 @@ namespace System.Windows.Forms
 			get { return base.BackgroundImage; }
 			set {
 				base.BackgroundImage = value;
-				Refresh ();
+				Invalidate ();
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public override ImageLayout BackgroundImageLayout {
+			get { return base.BackgroundImageLayout; }
+			set { base.BackgroundImageLayout = value; }
+		}
+#endif
+	
 		[DefaultValue(BorderStyle.None)]
 		[DispId(-504)]
 		public virtual BorderStyle BorderStyle {
@@ -235,8 +270,14 @@ namespace System.Windows.Forms
 			get { return ImeMode.Disable;}
 		}
 
+#if NET_2_0
+		protected override Padding DefaultMargin {
+			get { return new Padding (3, 0, 3, 0); }
+		}
+#endif
+
 		protected override Size DefaultSize {
-			get {return ThemeEngine.Current.LabelDefaultSize;}
+			get { return ThemeElements.LabelPainter.DefaultSize; }
 		}
 
 		[DefaultValue(FlatStyle.Standard)]
@@ -250,29 +291,45 @@ namespace System.Windows.Forms
 					return;
 
 				flat_style = value;
-				Refresh ();
+				if (Parent != null)
+					Parent.PerformLayout (this, "FlatStyle");
+				Invalidate ();
 			}
 		}
 
 		[Localizable(true)]
 		public Image Image {
 			get {
-				if (image != null) {
-					return image;
-				}
-					
-				if (image_list != null && ImageIndex >= 0) {
-					return image_list.Images[ImageIndex];
-				}
-					
+				if (this.image != null)
+					return this.image;
+
+				if (this.image_index >= 0)
+					if (this.image_list != null)
+						return this.image_list.Images[this.image_index];
+
+#if NET_2_0
+				if (!string.IsNullOrEmpty (this.image_key))
+					if (this.image_list != null)
+						return this.image_list.Images[this.image_key];
+#endif
 				return null;
 			}
 			set {
-				if (image == value)
-					return;
+				if (this.image != value) {
+					this.image = value;
+					this.image_index = -1;
+#if NET_2_0
+					this.image_key = string.Empty;
+#endif
+					this.image_list = null;
 
-				image = value;
-				Refresh ();
+#if NET_2_0
+					if (this.AutoSize && this.Parent != null)
+						this.Parent.PerformLayout (this, "Image");
+#endif
+
+					Invalidate ();
+				}
 			}
 		}
 
@@ -288,7 +345,7 @@ namespace System.Windows.Forms
 					return;
 
 				image_align = value;
-				Refresh ();
+				Invalidate ();
 			}
 		}
 
@@ -316,18 +373,36 @@ namespace System.Windows.Forms
 				if (value < -1)
 					throw new ArgumentException ();
 
-				if (image_index == value)
-					return;
-
-				image_index = value;
-
-				if (ImageList != null && image_index !=-1)
-					Image = null;
-
-				Refresh ();
+				if (this.image_index != value) {
+					this.image_index = value;
+					this.image = null;
+#if NET_2_0
+					this.image_key = string.Empty;
+#endif
+					Invalidate ();
+				}
 			}
 		}
 
+#if NET_2_0
+		[Localizable (true)]
+		[DefaultValue ("")]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[TypeConverter (typeof (ImageKeyConverter))]
+		public string ImageKey {
+			get { return this.image_key; }
+			set {
+				if (this.image_key != value) {
+					this.image = null;
+					this.image_index = -1;
+					this.image_key = value;
+					this.Invalidate ();
+				}
+			}
+		}
+#endif
+
 		[DefaultValue(null)]
 #if NET_2_0
 		[RefreshProperties (RefreshProperties.Repaint)]
@@ -343,7 +418,7 @@ namespace System.Windows.Forms
 				if (image_list != null && image_index !=-1)
 					Image = null;
 
-				Refresh ();
+				Invalidate ();
 			}
 		}
 
@@ -354,37 +429,54 @@ namespace System.Windows.Forms
 			set { base.ImeMode = value; }
 		}
 
+		internal virtual Size InternalGetPreferredSize (Size proposed)
+		{
+			Size size;
+
+			if (Text == string.Empty) {
+				size = new Size (0, Font.Height);
+			} else {
+				size = Size.Ceiling (TextRenderer.MeasureString (Text, Font, req_witdthsize, string_format));
+				size.Width += 3;
+			}
+
+#if NET_2_0
+			size.Width += Padding.Horizontal;
+			size.Height += Padding.Vertical;
+			
+			if (!use_compatible_text_rendering)
+				return size;
+#else
+				size.Height = Font.Height;
+#endif
+
+			if (border_style == BorderStyle.None)
+				size.Height += 3;
+			else
+				size.Height += 6;
+			
+			return size;
+		}
+
+#if NET_2_0
+		public override	Size GetPreferredSize (Size proposedSize)
+		{
+			return InternalGetPreferredSize (proposedSize);
+		}
+#endif
+
 		[Browsable(false)]
 		[DesignerSerializationVisibilityAttribute(DesignerSerializationVisibility.Hidden)]
 		[EditorBrowsable(EditorBrowsableState.Advanced)]
 		public virtual int PreferredHeight {
-			get { 
-				int preferred_height = Font.Height;
-
-#if NET_2_0
-				if (!use_compatible_text_rendering)
-					return preferred_height;
-#endif
-
-				if (border_style == BorderStyle.None)
-					return preferred_height + 3;
-					
-				return preferred_height + 6;
-			}
+			get { return InternalGetPreferredSize (Size.Empty).Height; }
 		}
 
 		[Browsable(false)]
 		[DesignerSerializationVisibilityAttribute(DesignerSerializationVisibility.Hidden)]
 		[EditorBrowsable(EditorBrowsableState.Advanced)]
 		public virtual int PreferredWidth {
-			get {
-				if (Text == string.Empty)
-					return 0;
-
-				SizeF size;
-				size = DeviceContext.MeasureString (Text, Font, req_witdthsize, string_format);
-				return (int) size.Width + 3;
-			}
+			get { return InternalGetPreferredSize (Size.Empty).Width; }
 		}
 
 #if NET_2_0
@@ -412,22 +504,19 @@ namespace System.Windows.Forms
 				if (!Enum.IsDefined (typeof (ContentAlignment), value))
 					throw new InvalidEnumArgumentException (string.Format("Enum argument value '{0}' is not valid for ContentAlignment", value));
 
-					if (text_align != value) {
-
-						text_align = value;
-
-						switch (value) {
-
-						case ContentAlignment.BottomLeft:
-							string_format.LineAlignment = StringAlignment.Far;
-							string_format.Alignment = StringAlignment.Near;
-							break;
-						case ContentAlignment.BottomCenter:
-							string_format.LineAlignment = StringAlignment.Far;
-							string_format.Alignment = StringAlignment.Center;
-							break;
-						case ContentAlignment.BottomRight:
-							string_format.LineAlignment = StringAlignment.Far;
+				if (text_align != value) {
+					text_align = value;
+					switch (value) {
+					case ContentAlignment.BottomLeft:
+						string_format.LineAlignment = StringAlignment.Far;
+						string_format.Alignment = StringAlignment.Near;
+						break;
+					case ContentAlignment.BottomCenter:
+						string_format.LineAlignment = StringAlignment.Far;
+						string_format.Alignment = StringAlignment.Center;
+						break;
+					case ContentAlignment.BottomRight:
+						string_format.LineAlignment = StringAlignment.Far;
 						string_format.Alignment = StringAlignment.Far;
 						break;
 					case ContentAlignment.TopLeft:
@@ -446,20 +535,20 @@ namespace System.Windows.Forms
 						string_format.LineAlignment = StringAlignment.Center;
 						string_format.Alignment = StringAlignment.Near;
 						break;
-						case ContentAlignment.MiddleRight:
-							string_format.LineAlignment = StringAlignment.Center;
-							string_format.Alignment = StringAlignment.Far;
-							break;
-						case ContentAlignment.MiddleCenter:
-							string_format.LineAlignment = StringAlignment.Center;
-							string_format.Alignment = StringAlignment.Center;
+					case ContentAlignment.MiddleRight:
+						string_format.LineAlignment = StringAlignment.Center;
+						string_format.Alignment = StringAlignment.Far;
+						break;
+					case ContentAlignment.MiddleCenter:
+						string_format.LineAlignment = StringAlignment.Center;
+						string_format.Alignment = StringAlignment.Center;
 						break;
 					default:
 						break;
 					}
 
 					OnTextAlignChanged (EventArgs.Empty);
-						Refresh();
+					Invalidate ();
 				}
 			}
 		}
@@ -471,7 +560,7 @@ namespace System.Windows.Forms
 				if (use_mnemonic != value) {
 					use_mnemonic = value;
 					SetUseMnemonic (use_mnemonic);
-					Refresh ();
+					Invalidate ();
 				}
 			}
 		}
@@ -480,34 +569,32 @@ namespace System.Windows.Forms
 
 		#region Public Methods
 
-		protected Rectangle CalcImageRenderBounds (Image image, Rectangle area, ContentAlignment img_align)
+		protected Rectangle CalcImageRenderBounds (Image image, Rectangle r, ContentAlignment align)
 		{
-				Rectangle rcImageClip = area;
+			Rectangle rcImageClip = r;
 			rcImageClip.Inflate (-2,-2);
 
-			int X = area.X;
-			int Y = area.Y;
+			int X = r.X;
+			int Y = r.Y;
 
-			if (img_align == ContentAlignment.TopCenter ||
-				img_align == ContentAlignment.MiddleCenter ||
-				img_align == ContentAlignment.BottomCenter) {
-				X += (area.Width - image.Width) / 2;
-			}
-			else if (img_align == ContentAlignment.TopRight ||
-				img_align == ContentAlignment.MiddleRight||
-				img_align == ContentAlignment.BottomRight) {
-				X += (area.Width - image.Width);
+			if (align == ContentAlignment.TopCenter ||
+				align == ContentAlignment.MiddleCenter ||
+				align == ContentAlignment.BottomCenter) {
+				X += (r.Width - image.Width) / 2;
+			} else if (align == ContentAlignment.TopRight ||
+				align == ContentAlignment.MiddleRight||
+				align == ContentAlignment.BottomRight) {
+				X += (r.Width - image.Width);
 			}
 
-			if( img_align == ContentAlignment.BottomCenter ||
-				img_align == ContentAlignment.BottomLeft ||
-				img_align == ContentAlignment.BottomRight) {
-				Y += area.Height - image.Height;
-			}
-			else if(img_align == ContentAlignment.MiddleCenter ||
-					img_align == ContentAlignment.MiddleLeft ||
-					img_align == ContentAlignment.MiddleRight) {
-				Y += (area.Height - image.Height) / 2;
+			if( align == ContentAlignment.BottomCenter ||
+				align == ContentAlignment.BottomLeft ||
+				align == ContentAlignment.BottomRight) {
+				Y += r.Height - image.Height;
+			} else if(align == ContentAlignment.MiddleCenter ||
+					align == ContentAlignment.MiddleLeft ||
+					align == ContentAlignment.MiddleRight) {
+				Y += (r.Height - image.Height) / 2;
 			}
 
 			rcImageClip.X = X;
@@ -531,12 +618,12 @@ namespace System.Windows.Forms
 				string_format.Dispose ();
 		}
 
-		protected void DrawImage (Graphics g, Image image, Rectangle area, ContentAlignment img_align)
+		protected internal void DrawImage (Graphics g, Image image, Rectangle r, ContentAlignment align)
 		{
  			if (image == null || g == null)
 				return;
 
-			Rectangle rcImageClip = CalcImageRenderBounds (image, area, img_align);
+			Rectangle rcImageClip = CalcImageRenderBounds (image, r, align);
 
 			if (Enabled)
 				g.DrawImage (image, rcImageClip.X, rcImageClip.Y, rcImageClip.Width, rcImageClip.Height);
@@ -563,21 +650,34 @@ namespace System.Windows.Forms
 			base.OnFontChanged (e);
 			if (autosize)
 				CalcAutoSize();
-			Refresh ();
+			Invalidate ();
 		}
 
-		protected override void OnPaint (PaintEventArgs pevent)
+#if NET_2_0
+		protected override void OnPaddingChanged (EventArgs e)
 		{
-			ThemeEngine.Current.DrawLabel (pevent.Graphics, ClientRectangle, this);
-			DrawImage (pevent.Graphics, Image, ClientRectangle, image_align);
-			base.OnPaint(pevent);
+			base.OnPaddingChanged (e);
+		}
+#endif
+
+		protected override void OnPaint (PaintEventArgs e)
+		{
+			ThemeElements.LabelPainter.Draw (e.Graphics, ClientRectangle, this);
+			base.OnPaint(e);
 		}
 
 		protected override void OnParentChanged (EventArgs e)
 		{
-				base.OnParentChanged (e);
+			base.OnParentChanged (e);
 		}
 
+#if NET_2_0
+		protected override void OnRightToLeftChanged (EventArgs e)
+		{
+			base.OnRightToLeftChanged (e);
+		}
+#endif
+
 		protected virtual void OnTextAlignChanged (EventArgs e)
 		{
 			EventHandler eh = (EventHandler)(Events [TextAlignChangedEvent]);
@@ -587,10 +687,10 @@ namespace System.Windows.Forms
 
 		protected override void OnTextChanged (EventArgs e)
 		{
-			base.OnTextChanged (e);			
+			base.OnTextChanged (e);
 			if (autosize)
 				CalcAutoSize ();
-			Refresh ();
+			Invalidate ();
 		}
 
 		protected override void OnVisibleChanged (EventArgs e)
@@ -600,11 +700,10 @@ namespace System.Windows.Forms
 
 		protected override bool ProcessMnemonic (char charCode)
 		{
-			if (IsMnemonic(charCode, Text) == true) {
+			if (IsMnemonic (charCode, Text)) {
 				// Select item next in line in tab order
-				if (this.Parent != null) {
+				if (this.Parent != null)
 					Parent.SelectNextControl(this, true, false, false, false);
-				}
 				return true;
 			}
 			
@@ -624,13 +723,12 @@ namespace System.Windows.Forms
 		protected override void WndProc (ref Message m)
 		{
 			switch ((Msg) m.Msg) {
-				case Msg.WM_DRAWITEM: {
-					m.Result = (IntPtr)1;
-				}
-					break;
-				default:
-					base.WndProc (ref m);
-					break;
+			case Msg.WM_DRAWITEM:
+				m.Result = (IntPtr)1;
+				break;
+			default:
+				base.WndProc (ref m);
+				break;
 			}
 		}
 
@@ -643,8 +741,9 @@ namespace System.Windows.Forms
 			if (!AutoSize)
 				return;
 
-		 	Width =  PreferredWidth;
-		 	Height =  PreferredHeight;			 	
+			Size s = InternalGetPreferredSize (Size.Empty);
+			
+			SetBounds (Left, Top, s.Width, s.Height, BoundsSpecified.Size);
 		}
 
 		private void OnHandleCreatedLB (Object o, EventArgs e)
@@ -692,6 +791,5 @@ namespace System.Windows.Forms
 			base.OnHandleDestroyed (e);
 		}
 #endif
-
 	}
 }
