@@ -22,6 +22,7 @@
 // Authors:
 //	Jordi Mas i Hernandez, jordi@ximian.com
 //	Chris Toshok  <toshok@ximian.com>
+//	Everaldo Canuto  <ecanuto@novell.com>
 //
 // Based on work by:
 //	Daniel Carrera, dcarrera@math.toronto.edu (stubbed out)
@@ -33,6 +34,7 @@ using System.Collections;
 using System.Drawing;
 using System.Drawing.Drawing2D;
 using System.Runtime.InteropServices;
+using System.Windows.Forms.Theming;
 
 namespace System.Windows.Forms
 {
@@ -40,6 +42,7 @@ namespace System.Windows.Forms
 #if NET_2_0
 	[ClassInterface (ClassInterfaceType.AutoDispatch)]
 	[ComVisible (true)]
+	[ToolboxItem ("System.Windows.Forms.Design.AutoSizeToolboxItem," + Consts.AssemblySystem_Design)]
 #endif
 	public class LinkLabel : Label, IButtonControl
 	{
@@ -68,6 +71,7 @@ namespace System.Windows.Forms
 		private LinkArea link_area;
 		private LinkBehavior link_behavior;
 		private LinkCollection link_collection;
+		private ArrayList links = new ArrayList();
 		internal Link[] sorted_links;
 		private bool link_visited;
 		internal Piece[] pieces;
@@ -87,6 +91,15 @@ namespace System.Windows.Forms
 			add { Events.AddHandler (LinkClickedEvent, value); }
 			remove { Events.RemoveHandler (LinkClickedEvent, value); }
 		}
+
+#if NET_2_0
+		[Browsable (true)]
+		[EditorBrowsable (EditorBrowsableState.Always)]
+		public new event EventHandler TabStopChanged {
+			add { base.TabStopChanged += value; }
+			remove { base.TabStopChanged -= value; }
+		}
+#endif
 		#endregion // Events
 
 		public LinkLabel ()
@@ -97,7 +110,7 @@ namespace System.Windows.Forms
 			pieces = null;
 			focused_index = -1;
 
-			string_format.FormatFlags = StringFormatFlags.NoClip;
+			string_format.FormatFlags |= StringFormatFlags.NoClip;
 			
 			ActiveLinkColor = Color.Red;
 			DisabledLinkColor = ThemeEngine.Current.ColorGrayText;
@@ -115,7 +128,7 @@ namespace System.Windows.Forms
 				| ControlStyles.DoubleBuffer
 #endif
 				, true);
-
+			CreateLinkPieces ();
 		}
 
 		#region Public Properties
@@ -177,8 +190,7 @@ namespace System.Windows.Forms
 				if (value.Start <0 || value.Length < -1)
 					throw new ArgumentException ();
 
-				if (Links.IsDefault)
-					Links.Clear ();
+				Links.Clear ();
 
 				if (!value.IsEmpty) {
 					Links.Add (value.Start, value.Length);
@@ -209,11 +221,10 @@ namespace System.Windows.Forms
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public LinkLabel.LinkCollection Links {
 			get {
-                                if (link_collection == null)
-                                        link_collection = new LinkCollection (this);
-                                        
-                                return link_collection;
-                        }
+				if (link_collection == null)
+					link_collection = new LinkCollection (this);
+				return link_collection;
+			}
 		}
 
 		[DefaultValue (false)]
@@ -239,16 +250,41 @@ namespace System.Windows.Forms
 
 		[RefreshProperties(RefreshProperties.Repaint)]
 		public override string Text {
-			get { return base.Text;	}
+			get { return base.Text; }
 			set {
 				if (base.Text == value)
 					return;
 
 				base.Text = value;
-				CreateLinkPieces ();				
+				CreateLinkPieces ();
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public new FlatStyle FlatStyle {
+			get { return base.FlatStyle; }
+			set {
+				if (base.FlatStyle == value)
+					return;
+
+				base.FlatStyle = value;
+			}
+		}
+
+		[RefreshProperties (RefreshProperties.Repaint)]
+		public new Padding Padding {
+			get { return base.Padding; }
+			set {
+				if (base.Padding == value)
+					return;
+
+				base.Padding = value;
+			}
+		}
+#endif
+
 		#endregion // Public Properties
 
 		DialogResult IButtonControl.DialogResult {
@@ -274,9 +310,16 @@ namespace System.Windows.Forms
 		protected override void CreateHandle ()
 		{
 			base.CreateHandle ();
-			CreateLinkPieces ();			
+			CreateLinkPieces ();
 		}
 
+#if NET_2_0
+		protected override void OnAutoSizeChanged (EventArgs e)
+		{
+			base.OnAutoSizeChanged (e);
+		}
+#endif
+
 		protected override void OnEnabledChanged (EventArgs e)
 		{
 			base.OnEnabledChanged (e);
@@ -291,7 +334,7 @@ namespace System.Windows.Forms
 
 		protected override void OnGotFocus (EventArgs e)
 		{
-			base.OnGotFocus (e);			
+			base.OnGotFocus (e);
 
 			// Set focus to the first enabled link piece
 			if (focused_index == -1) {
@@ -339,7 +382,7 @@ namespace System.Windows.Forms
 
 		protected override void OnLostFocus (EventArgs e)
 		{
-			base.OnLostFocus (e);			
+			base.OnLostFocus (e);
 			
 			// Clean focus in link pieces
 			if (focused_index != -1)
@@ -372,6 +415,13 @@ namespace System.Windows.Forms
 			UpdateHover (null);
 		}
 
+#if NET_2_0
+		protected override void OnPaddingChanged (EventArgs e)
+		{
+			base.OnPaddingChanged (e);
+		}
+#endif
+
 		private void UpdateHover (Link link)
 		{
 			if (link == hovered_link)
@@ -420,17 +470,20 @@ namespace System.Windows.Forms
 			active_link = null;
 
 			if (clicked_link != null)
+#if NET_2_0
+				OnLinkClicked (new LinkLabelLinkClickedEventArgs (clicked_link, e.Button));
+#else
 				OnLinkClicked (new LinkLabelLinkClickedEventArgs (clicked_link));
+#endif
 		}
 
-		protected override void OnPaint (PaintEventArgs pevent)
+		protected override void OnPaint (PaintEventArgs e)
 		{
 			// We need to invoke paintbackground because control is opaque
 			// and can have transparent colors.
-			base.InvokePaintBackground (this, pevent);
+			base.InvokePaintBackground (this, e);
 			
-			ThemeEngine.Current.DrawLinkLabel (pevent.Graphics, pevent.ClipRectangle, this);
-			DrawImage (pevent.Graphics, Image, ClientRectangle, image_align);
+			ThemeElements.LinkLabelPainter.Draw (e.Graphics, e.ClipRectangle, this);
 			// Do not call base.OnPaint since it's the Label class 
 		}
 
@@ -441,7 +494,7 @@ namespace System.Windows.Forms
 
 		protected override void OnTextAlignChanged (EventArgs e)
 		{
-			CreateLinkPieces ();			
+			CreateLinkPieces ();
 			base.OnTextAlignChanged (e);
 		}
 
@@ -472,8 +525,10 @@ namespace System.Windows.Forms
 		protected override void Select (bool directed, bool forward)
 		{
 			if (directed) {
-				if (focused_index != -1)
+				if (focused_index != -1) {
 					sorted_links[focused_index].Focused = false;
+					focused_index = -1;
+				}
 
 				if (forward) {
 					for (int n = focused_index + 1; n < sorted_links.Length; n++) {
@@ -484,8 +539,7 @@ namespace System.Windows.Forms
 							return;
 						}
 					}
-				}
-				else {
+				} else {
 					if (focused_index == -1)
 						focused_index = sorted_links.Length;
 
@@ -512,9 +566,9 @@ namespace System.Windows.Forms
 			CreateLinkPieces();
 		}
 
-		protected override void WndProc (ref Message m)
+		protected override void WndProc (ref Message msg)
 		{
-			base.WndProc (ref m);
+			base.WndProc (ref msg);
 		}
 
 		#endregion //Public Methods
@@ -555,6 +609,8 @@ namespace System.Windows.Forms
 			if (Text.Length == 0) {
 				SetStyle (ControlStyles.Selectable, false);
 				TabStop = false;
+				link_area.Start = 0;
+				link_area.Length = 0;
 				return;
 			}
 			
@@ -617,7 +673,7 @@ namespace System.Windows.Forms
 
 			string_format.SetMeasurableCharacterRanges (ranges);
 
-			Region[] regions = DeviceContext.MeasureCharacterRanges (Text,
+			Region[] regions = TextRenderer.MeasureCharacterRanges (Text,
 										 ThemeEngine.Current.GetLinkFont (this),
 										 ClientRectangle,
 										 string_format);
@@ -659,7 +715,7 @@ namespace System.Windows.Forms
 		// System.Windows.Forms.LinkLabel.Link
 		//
 #if NET_2_0
-		// XXX [TypeConverter (typeof (LinkConverter))]
+		[TypeConverter (typeof (LinkConverter))]
 #endif
 		public class Link
 		{
@@ -667,12 +723,17 @@ namespace System.Windows.Forms
 			internal int length;
 			private object linkData;
 			private int start;
-			private bool visited;			
+			private bool visited;
 			private LinkLabel owner;
 			private bool hovered;
 			internal ArrayList pieces;
 			private bool focused;
 			private bool active;
+#if NET_2_0
+			private string description;
+			private string name;
+			private object tag;
+#endif
 
 			internal Link (LinkLabel owner)
 			{
@@ -683,9 +744,53 @@ namespace System.Windows.Forms
 				linkData = null;
 				this.owner = owner;
 				pieces = new ArrayList ();
+#if NET_2_0
+				name = string.Empty;
+#endif
 			}
 
 #if NET_2_0
+			public Link ()
+			{
+				this.enabled = true;
+				this.name = string.Empty;
+				this.pieces = new ArrayList ();
+			}
+
+			public Link (int start, int length) : this ()
+			{
+				this.start = start;
+				this.length = length;
+			}
+
+			public Link (int start, int length, Object linkData) : this (start, length)
+			{
+				this.linkData = linkData;
+			}
+#endif
+
+			#region Public Properties
+#if NET_2_0
+			public string Description {
+				get { return this.description; }
+				set { this.description = value; }
+			}
+			
+			[DefaultValue ("")]
+			public string Name {
+				get { return this.name; }
+				set { this.name = value; }
+			}
+			
+			[Bindable (true)]
+			[Localizable (false)]
+			[DefaultValue (null)]
+			[TypeConverter (typeof (StringConverter))]
+			public Object Tag {
+				get { return this.tag; }
+				set { this.tag = value; }
+			}
+				
 			[DefaultValue (true)]
 #endif
 			public bool Enabled {
@@ -777,6 +882,11 @@ namespace System.Windows.Forms
 				}
 			}
 
+			internal LinkLabel Owner {
+				set { owner = value; }
+			}
+			#endregion
+
 			private void Invalidate ()
 			{
 				for (int i = 0; i < pieces.Count; i ++)
@@ -805,45 +915,82 @@ namespace System.Windows.Forms
 		}
 
 		//
-		// System.Windows.Forms.LinkLabel.Link
+		// System.Windows.Forms.LinkLabel.LinkCollection
 		//
 		public class LinkCollection :  IList, ICollection, IEnumerable
 		{
 			private LinkLabel owner;
-			private ArrayList collection = new ArrayList();
+			private bool links_added = false;
 
 			public LinkCollection (LinkLabel owner)
 			{
-				if (owner==null)
-					throw new ArgumentNullException ();
+				if (owner == null)
+					throw new ArgumentNullException ("owner");
 
 				this.owner = owner;
 			}
 
 			[Browsable (false)]
 			public int Count {
-				get { return collection.Count; }
+				get { return owner.links.Count; }
 			}
 
 			public bool IsReadOnly {
 				get { return false; }
 			}
 
-			public virtual LinkLabel.Link this[int index]  {
+			public virtual LinkLabel.Link this[int index] {
 				get {
 					if (index < 0 || index >= Count)
 						throw  new  ArgumentOutOfRangeException();
 
-					return (LinkLabel.Link) collection[index];
+					return (LinkLabel.Link) owner.links[index];
 				}
 				set {
 					if (index < 0 || index >= Count)
 						throw new  ArgumentOutOfRangeException();
 
-					collection[index] = value;
+					owner.links[index] = value;
 				}
 			}
 
+#if NET_2_0
+			public virtual Link this[string key] {
+				get {
+					if (string.IsNullOrEmpty (key))
+						return null;
+						
+					foreach (Link l in owner.links)
+						if (string.Compare (l.Name, key, true) == 0)
+							return l;
+							
+					return null;
+				}
+			}
+			
+			public
+#else
+			internal
+#endif
+			int Add (Link value)
+			{
+				value.Owner = owner;
+				/* remove the default 0,-1 link */
+				if (IsDefault) {
+					/* don't call Clear() here to save the additional CreateLinkPieces */
+					owner.links.Clear ();
+				}
+
+				int idx = owner.links.Add (value);
+				links_added = true;
+
+				owner.sorted_links = null;
+				owner.CheckLinks ();
+				owner.CreateLinkPieces ();
+
+				return idx;
+			}
+
 			public Link Add (int start, int length)
 			{
 				return Add (start, length, null);
@@ -857,63 +1004,81 @@ namespace System.Windows.Forms
 				}
 			}
 
-			public Link Add (int start, int length, object o)
+			public Link Add (int start, int length, object linkData)
 			{
 				Link link = new Link (owner);
-				int idx;
-
-				/* remove the default 0,-1 link */
-				if (IsDefault) {
-					/* don't call Clear() here to save the additional CreateLinkPieces */
-					collection.Clear ();
-				}
-
 				link.Length = length;
 				link.Start = start;
-				link.LinkData = o;
-				idx = collection.Add (link);
+				link.LinkData = linkData;
 
-				owner.sorted_links = null;
-				owner.CheckLinks ();
-				owner.CreateLinkPieces ();
-				return (Link) collection[idx];
+				int idx = Add (link);
+
+				return (Link) owner.links[idx];
 			}
 
 			public virtual void Clear ()
 			{
-				collection.Clear();
+				owner.links.Clear();
 				owner.sorted_links = null;
 				owner.CreateLinkPieces ();
 			}
 
 			public bool Contains (Link link)
 			{
-				return collection.Contains (link);
+				return owner.links.Contains (link);
 			}
 
+#if NET_2_0
+			public virtual bool ContainsKey (string key)
+			{
+				return !(this[key] == null);
+			}
+#endif
+
 			public IEnumerator GetEnumerator ()
 			{
-				return collection.GetEnumerator ();
+				return owner.links.GetEnumerator ();
 			}
 
 			public int IndexOf (Link link)
 			{
-				return collection.IndexOf (link);
+				return owner.links.IndexOf (link);
 			}
 
+#if NET_2_0
+			public virtual int IndexOfKey (string key)
+			{
+				if (string.IsNullOrEmpty (key))
+					return -1;
+					
+				return IndexOf (this[key]);
+			}
+			
+			public bool LinksAdded {
+				get { return this.links_added; }
+			}
+#endif
+
 			public void Remove (LinkLabel.Link value)
 			{
-				collection.Remove (value);
+				owner.links.Remove (value);
 				owner.sorted_links = null;
 				owner.CreateLinkPieces ();
 			}
 
+#if NET_2_0
+			public virtual void RemoveByKey (string key)
+			{	
+				Remove (this[key]);
+			}
+#endif
+
 			public void RemoveAt (int index)
 			{
 				if (index >= Count)
 					throw new ArgumentOutOfRangeException ("Invalid value for array index");
 
-				collection.Remove (collection[index]);
+				owner.links.Remove (owner.links[index]);
 				owner.sorted_links = null;
 				owner.CreateLinkPieces ();
 			}
@@ -923,8 +1088,8 @@ namespace System.Windows.Forms
 			}
 
 			object IList.this[int index] {
-				get { return collection[index]; }
-				set { collection[index] = value; }
+				get { return owner.links[index]; }
+				set { owner.links[index] = value; }
 			}
 
 			object ICollection.SyncRoot {
@@ -937,39 +1102,39 @@ namespace System.Windows.Forms
 
 			void ICollection.CopyTo (Array dest, int index)
 			{
-				collection.CopyTo (dest, index);
+				owner.links.CopyTo (dest, index);
 			}
 
-			int IList.Add (object control)
+			int IList.Add (object value)
 			{
-				int idx = collection.Add (control);
+				int idx = owner.links.Add (value);
 				owner.sorted_links = null;
 				owner.CheckLinks ();
 				owner.CreateLinkPieces ();
 				return idx;
 			}
 
-			bool IList.Contains (object control)
+			bool IList.Contains (object link)
 			{
-				return Contains ((Link)control);
+				return Contains ((Link) link);
 			}
 
-			int IList.IndexOf (object control)
+			int IList.IndexOf (object link)
 			{
-				return collection.IndexOf (control);
+				return owner.links.IndexOf (link);
 			}
 
 			void IList.Insert (int index, object value)
 			{
-				collection.Insert (index, value);
+				owner.links.Insert (index, value);
 				owner.sorted_links = null;
 				owner.CheckLinks ();
 				owner.CreateLinkPieces ();
 			}
 
-			void IList.Remove (object control)
+			void IList.Remove (object value)
 			{
-				Remove ((Link)control);
+				Remove ((Link) value);
 			}
 		}
 #if NET_2_0
@@ -979,7 +1144,6 @@ namespace System.Windows.Forms
 			get {
 				return use_compatible_text_rendering;
 			}
-
 			set {
 				use_compatible_text_rendering = value;
 			}
