@@ -41,21 +41,17 @@ namespace System.Windows.Forms {
 		internal bool popup_active;
 		internal bool popdown_menu;
 		internal bool hotkey_active;
+		private bool mouse_down = false;
 		public Menu CurrentMenu;
 		public Menu TopMenu;
-		Form grab_control;
-
+		public Control GrabControl;
+		Point last_motion = Point.Empty;
+		
 	    public MenuTracker (Menu top_menu)
 		{
 			TopMenu = CurrentMenu = top_menu;
 			foreach (MenuItem item in TopMenu.MenuItems)
 				AddShortcuts (item);
-
-			if (top_menu is ContextMenu) {
-				grab_control = (top_menu as ContextMenu).SourceControl.FindForm ();
-				grab_control.ActiveTracker = this;
-			} else
-				grab_control = top_menu.Wnd.FindForm ();
 		}
 
 		enum KeyNavState {
@@ -81,7 +77,7 @@ namespace System.Windows.Forms {
 
 		private void UpdateCursor ()
 		{
-			Control child_control = grab_control.GetRealChildAtPoint (Cursor.Position);
+			Control child_control = GrabControl.GetRealChildAtPoint (Cursor.Position);
 			if (child_control != null) {
 				if (active)
 					XplatUI.SetCursor (child_control.Handle, Cursors.Default.handle);
@@ -92,19 +88,26 @@ namespace System.Windows.Forms {
 
 		void Deactivate ()
 		{
+			bool redrawbar = (keynav_state != KeyNavState.Idle) && (TopMenu is MainMenu); 
+
 			active = false;
 			popup_active = false;
 			hotkey_active = false;
-			grab_control.ActiveTracker = null;
+			if (GrabControl != null)
+				GrabControl.ActiveTracker = null;
 			keynav_state = KeyNavState.Idle;
 			if (TopMenu is ContextMenu) {
 				PopUpWindow puw = TopMenu.Wnd as PopUpWindow;
 				DeselectItem (TopMenu.SelectedItem);
-				puw.HideWindow ();
+				if (puw != null)
+					puw.HideWindow ();
 			} else {
 				DeselectItem (TopMenu.SelectedItem);
 			}
 			CurrentMenu = TopMenu;
+
+			if (redrawbar)
+				(TopMenu as MainMenu).Draw ();			
 		}
 
 		MenuItem FindItemByCoords (Menu menu, Point pt)
@@ -137,6 +140,8 @@ namespace System.Windows.Forms {
 		{
 			MenuItem item = GetItemAtXY (args.X, args.Y);
 
+			mouse_down = true;
+
 			if (item == null) {
 				Deactivate ();
 				return false;
@@ -158,12 +163,19 @@ namespace System.Windows.Forms {
 			if ((CurrentMenu == TopMenu) && !popdown_menu)
 				SelectItem (item.Parent, item, item.IsPopup);
 			
-			grab_control.ActiveTracker = this;
+			GrabControl.ActiveTracker = this;
 			return true;
 		}
 
 		public void OnMotion (MouseEventArgs args)
 		{
+			// Windows helpfully sends us MOUSEMOVE messages when any key is pressed.
+			// So if the mouse hasn't actually moved since the last MOUSEMOVE, ignore it.
+			if (args.Location == last_motion)
+				return;
+				
+			last_motion = args.Location;
+			
 			MenuItem item = GetItemAtXY (args.X, args.Y);
 
 			UpdateCursor ();
@@ -171,7 +183,7 @@ namespace System.Windows.Forms {
 			if (CurrentMenu.SelectedItem == item)
 				return;
 
-			grab_control.ActiveTracker = (active || item != null) ? this : null;
+			GrabControl.ActiveTracker = (active || item != null) ? this : null;
 
 			if (item == null) {
 				MenuItem old_item = CurrentMenu.SelectedItem;
@@ -184,8 +196,6 @@ namespace System.Windows.Forms {
 				if (keynav_state == KeyNavState.Navigating)
 					return;
 				
-				keynav_state = KeyNavState.Navigating;
-				
 				// Select parent menu when move outside of menu item
 				if (old_item.Parent is MenuItem) {
 					MenuItem new_item = (old_item.Parent as MenuItem);
@@ -206,6 +216,13 @@ namespace System.Windows.Forms {
 
 		public void OnMouseUp (MouseEventArgs args)
 		{
+			/* mouse down dont comes from menu */
+			if (!mouse_down)
+				return;
+
+			mouse_down = false;
+
+			/* is not left button */
 			if ((args.Button & MouseButtons.Left) == 0)
 				return;
 			
@@ -229,38 +246,70 @@ namespace System.Windows.Forms {
 			/* Perform click when is not a popup */
 			if (!item.IsPopup) {
 				DeselectItem (item);
+				
+				// Raise the form's MenuComplete event
+				if (TopMenu != null && TopMenu.Wnd != null) {
+					Form f = TopMenu.Wnd.FindForm ();
+					
+					if (f != null)
+						f.OnMenuComplete (EventArgs.Empty);	
+				}
+				
 				item.PerformClick ();
 			}
 		}
 
 		static public bool TrackPopupMenu (Menu menu, Point pnt)
 		{
-			Object	queue_id;
-
 			if (menu.MenuItems.Count <= 0)	// No submenus to track
 				return true;				
 
-			MenuTracker tracker = new MenuTracker (menu);
+			MenuTracker tracker = menu.tracker;
 			tracker.active = true;
 			tracker.popup_active = true;
-			menu.tracker = tracker;
-
-			menu.Wnd = new PopUpWindow (tracker.grab_control, menu);
+			
+			// Set GrabControl
+			Control src_ctrl = (tracker.TopMenu as ContextMenu).SourceControl;
+			tracker.GrabControl = src_ctrl.FindForm ();
+			if (tracker.GrabControl == null)
+				tracker.GrabControl = src_ctrl.FindRootParent ();
+			tracker.GrabControl.ActiveTracker = tracker;
+			
+			menu.Wnd = new PopUpWindow (tracker.GrabControl, menu);
 			menu.Wnd.Location =  menu.Wnd.PointToClient (pnt);
-
 			((PopUpWindow)menu.Wnd).ShowWindow ();
 
 			bool no_quit = true;
 
-			queue_id = XplatUI.StartLoop(Thread.CurrentThread);
+			Object queue_id = XplatUI.StartLoop(Thread.CurrentThread);
 
 			while ((menu.Wnd != null) && menu.Wnd.Visible && no_quit) {
 				MSG msg = new MSG ();
 				no_quit = XplatUI.GetMessage(queue_id, ref msg, IntPtr.Zero, 0, 0);
-				XplatUI.TranslateMessage(ref msg);
-				XplatUI.DispatchMessage(ref msg);				
+
+				switch((Msg)msg.message) {
+				case Msg.WM_KEYDOWN:
+				case Msg.WM_SYSKEYDOWN:
+				case Msg.WM_CHAR:
+				case Msg.WM_SYSCHAR:
+				case Msg.WM_KEYUP:
+				case Msg.WM_SYSKEYUP:
+					Control c = Control.FromHandle(msg.hwnd);
+					if (c != null) {
+						Message m = Message.Create(msg.hwnd, (int)msg.message, msg.wParam, msg.lParam);
+						c.PreProcessControlMessageInternal (ref m);
+					}
+					break;
+				default:
+					XplatUI.TranslateMessage (ref msg);
+					XplatUI.DispatchMessage (ref msg);
+					break;
+				}
 			}
 
+			if (tracker.GrabControl.IsDisposed)
+				return true;
+
 			if (!no_quit)
 				XplatUI.PostQuitMessage(0);
 
@@ -281,7 +330,7 @@ namespace System.Windows.Forms {
 
 			/* When popup item then close all sub popups and unselect all sub items */
 			if (item.IsPopup) {
-				HideSubPopups (item);
+				HideSubPopups (item, TopMenu);
 				
 				/* Unselect all selected sub itens */
 				foreach (MenuItem subitem in item.MenuItems)
@@ -352,7 +401,7 @@ namespace System.Windows.Forms {
 			}
 
 			popup_active = true;
-			PopUpWindow puw = new PopUpWindow (grab_control, item);
+			PopUpWindow puw = new PopUpWindow (GrabControl, item);
 			
 			Point pnt;
 			if (menu is MainMenu)
@@ -366,18 +415,26 @@ namespace System.Windows.Forms {
 			puw.ShowWindow ();
 		}
 
-		static public void HideSubPopups (Menu menu)
+		static public void HideSubPopups (Menu menu, Menu topmenu)
 		{
 			foreach (MenuItem item in menu.MenuItems)
 				if (item.IsPopup)
-					HideSubPopups (item);
+					HideSubPopups (item, null);
 
 			if (menu.Wnd == null)
 				return;
 
 			PopUpWindow puw = menu.Wnd as PopUpWindow;
-			puw.Hide ();
+			if (puw != null) {
+				puw.Hide ();
+				puw.Dispose ();
+			}
 			menu.Wnd = null;
+
+#if NET_2_0
+			if ((topmenu != null) && (topmenu is MainMenu))
+				((MainMenu) topmenu).OnCollapse (EventArgs.Empty);
+#endif
 		}
 
 		MenuItem FindSubItemByCoord (Menu menu, Point pnt)
@@ -407,17 +464,19 @@ namespace System.Windows.Forms {
 
 		static MenuItem FindItemByKey (Menu menu, IntPtr key)
 		{
-			char key_char = (char ) (key.ToInt32() & 0xff);
-			key_char = Char.ToUpper (key_char);
-
+			char key_char = Char.ToUpper ((char) (key.ToInt32() & 0xff));
 			foreach (MenuItem item in menu.MenuItems) {
-				if (item.Mnemonic == '\0')
-					continue;
-
 				if (item.Mnemonic == key_char)
 					return item;
 			}
 
+			string key_str = key_char.ToString (); 
+			foreach (MenuItem item in menu.MenuItems) {
+				//if (item.Mnemonic == key_char)
+				if (item.Text.StartsWith (key_str))
+					return item;
+			}
+
 			return null;
 		}
 
@@ -528,13 +587,13 @@ namespace System.Windows.Forms {
 				case KeyNavState.Idle:
 					keynav_state = KeyNavState.Startup;
 					hotkey_active = true;
+					GrabControl.ActiveTracker = this;
 					CurrentMenu = TopMenu;
 					main_menu.Draw ();
 					break;
 				case KeyNavState.Startup:
 					break;
 				default:
-					keynav_state = KeyNavState.Idle;
 					Deactivate ();
 					main_menu.Draw ();
 					break;
@@ -551,7 +610,6 @@ namespace System.Windows.Forms {
 					SelectItem (TopMenu, TopMenu.MenuItems [0], false);
 					break;
 				default:
-					keynav_state = KeyNavState.Idle;
 					Deactivate ();
 					main_menu.Draw ();
 					break;
@@ -564,11 +622,11 @@ namespace System.Windows.Forms {
 		{
 			keynav_state = KeyNavState.Navigating;
 			MenuItem item = FindItemByKey (CurrentMenu, msg.WParam);
-			if (item == null)
+			if ((item == null) || (GrabControl == null))
 				return false;
 
 			active = true;
-			grab_control.ActiveTracker = this;
+			GrabControl.ActiveTracker = this;
 			
 			SelectItem (CurrentMenu, item, true);
 			if (item.IsPopup) {
@@ -610,13 +668,21 @@ namespace System.Windows.Forms {
 			if (item == null)
 				return false;
 
-			Deactivate ();
+			if (active)
+				Deactivate ();
 			item.PerformClick ();
 			return true;
 		}
 
 		public bool ProcessKeys (ref Message msg, Keys keyData)
 		{
+			// If we get Alt-F4, Windows will ignore it because we have a capture,
+			// release the capture and the program will exit.  (X11 doesn't care.)
+			if ((keyData & Keys.Alt) == Keys.Alt && (keyData & Keys.F4) == Keys.F4) {
+				GrabControl.ActiveTracker = null;
+				return false;
+			}
+			
 			if ((Msg)msg.Msg != Msg.WM_SYSKEYUP && ProcessShortcut (keyData))
 				return true;
 			else if ((keyData & Keys.KeyCode) == Keys.Menu && TopMenu is MainMenu) {
@@ -654,7 +720,7 @@ namespace System.Windows.Forms {
 						SelectItem (item, item.MenuItems [0], false);
 						CurrentMenu = item;
 						active = true;
-						grab_control.ActiveTracker = this;
+						GrabControl.ActiveTracker = this;
 					}
 					return true;
 				}
@@ -677,12 +743,21 @@ namespace System.Windows.Forms {
 					ShowSubPopup (CurrentMenu, item);
 					SelectItem (item, item.MenuItems [0], false);
 					CurrentMenu = item;
-				} else if (CurrentMenu.parent_menu is MainMenu) {
-					item = GetNextItem (CurrentMenu.parent_menu, ItemNavigation.Next);
-					SelectItem (CurrentMenu.parent_menu, item, item.IsPopup);
-					if (item.IsPopup) {
-						SelectItem (item, item.MenuItems [0], false);
-						CurrentMenu = item;
+				} else {
+					//Search up for a main menu
+					Menu Prnt = CurrentMenu.parent_menu;
+					while (Prnt != null && !(Prnt is MainMenu)) {
+						Prnt = Prnt.parent_menu;
+ 					}
+					if (Prnt is MainMenu)
+					{
+						item = GetNextItem(Prnt, ItemNavigation.Next);
+						SelectItem(Prnt, item, item.IsPopup);
+						if (item.IsPopup)
+						{
+							SelectItem(item, item.MenuItems[0], false);
+							CurrentMenu = item;
+						}
 					}
 				}
 				break;
@@ -704,7 +779,7 @@ namespace System.Windows.Forms {
 						CurrentMenu = item;
 					}
 				} else {
-					HideSubPopups (CurrentMenu);
+					HideSubPopups (CurrentMenu, TopMenu);
 					CurrentMenu = CurrentMenu.parent_menu;
 				}
 				break;
@@ -718,7 +793,7 @@ namespace System.Windows.Forms {
 						SelectItem (item, item.MenuItems [0], false);
 						CurrentMenu = item;
 						active = true;
-						grab_control.ActiveTracker = this;
+						GrabControl.ActiveTracker = this;
 					}
 					return true;
 				}
@@ -735,16 +810,16 @@ namespace System.Windows.Forms {
 				break;
 			}
 
-			return false;
+			return active;
 		}
 	}
 
 	internal class PopUpWindow : Control
 	{
 		private Menu menu;
-		private Form form;
+		private Control form;
 
-		public PopUpWindow (Form form, Menu menu): base ()
+		public PopUpWindow (Control form, Menu menu): base ()
 		{
 			this.menu = menu;
 			this.form = form;
@@ -779,17 +854,10 @@ namespace System.Windows.Forms {
 		public void HideWindow ()
 		{
 			XplatUI.SetCursor (form.Handle, form.Cursor.handle);
-			MenuTracker.HideSubPopups (menu);
+			MenuTracker.HideSubPopups (menu, null);
     		Hide ();
 		}
 
-#if false
-		protected override bool ProcessCmdKey (ref Message msg, Keys keyData)
-		{	
-			return MenuTracker.ProcessKeys (menu, ref msg, keyData, tracker);
-		}
-#endif
-
 		protected override void CreateHandle ()
 		{
 			base.CreateHandle ();
@@ -799,15 +867,30 @@ namespace System.Windows.Forms {
 		// Called when the number of items has changed
 		internal void RefreshItems ()
 		{
+			Point pt = new Point (Location.X, Location.Y);
+			
 			ThemeEngine.Current.CalcPopupMenuSize (DeviceContext, menu);
 
-			if ((Location.X + menu.Rect.Width) > SystemInformation.WorkingArea.Width) {
-				Location = new Point (Location.X - menu.Rect.Width, Location.Y);
+			if ((pt.X + menu.Rect.Width) > SystemInformation.VirtualScreen.Width) {
+				if (((pt.X - menu.Rect.Width) > 0) && !(menu.parent_menu is MainMenu))
+					pt.X = pt.X - menu.Rect.Width;
+				else
+					pt.X = SystemInformation.VirtualScreen.Width - menu.Rect.Width;
+
+				if (pt.X < 0)
+					pt.X = 0;
 			}
-			if ((Location.Y + menu.Rect.Height) > SystemInformation.WorkingArea.Height) {
-				Location = new Point (Location.X, Location.Y - menu.Rect.Height);
+			if ((pt.Y + menu.Rect.Height) > SystemInformation.VirtualScreen.Height) {
+				if ((pt.Y - menu.Rect.Height) > 0)
+					pt.Y = pt.Y - menu.Rect.Height;
+				else
+					pt.Y = SystemInformation.VirtualScreen.Height - menu.Rect.Height;
+
+				if (pt.Y < 0)
+					pt.Y = 0;
 			}
 
+			Location = pt;
 			Width = menu.Rect.Width;
 			Height = menu.Rect.Height;			
 		}
