@@ -35,6 +35,7 @@ namespace System.Windows.Forms
 {
 	[ClassInterface (ClassInterfaceType.AutoDispatch)]
 	[ComVisible (true)]
+	[Designer ("System.Windows.Forms.Design.ToolStripDropDownDesigner, " + Consts.AssemblySystem_Design, "System.ComponentModel.Design.IDesigner")]
 	public class ToolStripDropDown : ToolStrip
 	{
 		private bool allow_transparency;
@@ -54,11 +55,18 @@ namespace System.Windows.Forms
 			is_visible = false;
 			this.DefaultDropDownDirection = ToolStripDropDownDirection.Right;
 			this.GripStyle = ToolStripGripStyle.Hidden;
-			this.TopLevel = true;
+			this.is_toplevel = true;
 		}
 		#endregion
 
 		#region Public Properties
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public new bool AllowItemReorder {
+			get { return base.AllowItemReorder; }
+			set { base.AllowItemReorder = value; }
+		}
+		
 		[Browsable (false)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		public bool AllowTransparency {
@@ -100,6 +108,9 @@ namespace System.Windows.Forms
 			set { base.AutoSize = value; }
 		}
 
+		[Browsable (false)]
+		[DefaultValue (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
 		public new bool CanOverflow {
 			get { return this.can_overflow; }
 			set { this.can_overflow = value; }
@@ -176,6 +187,7 @@ namespace System.Windows.Forms
 			set { base.GripStyle = value; }
 		}
 
+		[Browsable (false)]
 		public bool IsAutoGenerated {
 			get { return this is ToolStripOverflow; }
 		}
@@ -208,6 +220,7 @@ namespace System.Windows.Forms
 			}
 		}
 
+		[Browsable (false)]
 		[EditorBrowsable (EditorBrowsableState.Never)]
 		public new ToolStripOverflowButton OverflowButton {
 			get { return base.OverflowButton; }
@@ -220,7 +233,7 @@ namespace System.Windows.Forms
 			set { this.owner_item = value; 
 				
 				if (this.owner_item != null)
-					if (this.owner_item.Owner != null)
+					if (this.owner_item.Owner != null && this.owner_item.Owner.RenderMode != ToolStripRenderMode.ManagerRenderMode)
 						this.Renderer = this.owner_item.Owner.Renderer;
 			}
 		}
@@ -254,6 +267,13 @@ namespace System.Windows.Forms
 			set { }
 		}
 
+		[Browsable (false)]
+		[DefaultValue (ToolStripTextDirection.Horizontal)]
+		public override ToolStripTextDirection TextDirection {
+			get { return base.TextDirection; }
+			set { base.TextDirection = value; }
+		}
+
 		[Browsable (false)]
 		[DesignerSerializationVisibility (DesignerSerializationVisibility.Hidden)]
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
@@ -302,9 +322,9 @@ namespace System.Windows.Forms
 			get { return true; }
 		}
 
-		//protected internal override Size MaxItemSize {
-		//        get {  return new Size (Screen.PrimaryScreen.Bounds.Width - 2, Screen.PrimaryScreen.Bounds.Height - 34); }
-		//}
+		protected internal override Size MaxItemSize {
+			get { return new Size (Screen.PrimaryScreen.Bounds.Width - 2, Screen.PrimaryScreen.Bounds.Height - 34); }
+		}
 
 		protected virtual bool TopMost {
 			get { return true; }
@@ -319,6 +339,9 @@ namespace System.Windows.Forms
 
 		public void Close (ToolStripDropDownCloseReason reason)
 		{
+			if (!this.Visible)
+				return;
+				
 			// Give users a chance to cancel the close
 			ToolStripDropDownClosingEventArgs e = new ToolStripDropDownClosingEventArgs (reason);
 			this.OnClosing (e);
@@ -343,8 +366,7 @@ namespace System.Windows.Forms
 
 			// Recursive hide all child dropdowns
 			foreach (ToolStripItem tsi in this.Items)
-				if (tsi is ToolStripMenuItem)
-					(tsi as ToolStripMenuItem).HideDropDown (reason);
+				tsi.Dismiss (reason);
 			
 			this.OnClosed (new ToolStripDropDownClosedEventArgs (reason));
 		}
@@ -353,40 +375,26 @@ namespace System.Windows.Forms
 		[EditorBrowsable (EditorBrowsableState.Never)]
 		public new void Show ()
 		{
-			CancelEventArgs e = new CancelEventArgs ();
-			this.OnOpening (e);
-			
-			if (e.Cancel)
-				return;
-
-			// The tracker lets us know when the form is clicked or loses focus
-			ToolStripManager.AppClicked += new EventHandler (ToolStripMenuTracker_AppClicked);
-			ToolStripManager.AppFocusChange += new EventHandler (ToolStripMenuTracker_AppFocusChange);
-
-			base.Show ();
-			
-			this.OnOpened (EventArgs.Empty);
+			Show (Location, DefaultDropDownDirection);
 		}
 		
 		public void Show (Point screenLocation)
 		{
-			this.Location = screenLocation;
-			Show ();
+			Show (screenLocation, DefaultDropDownDirection);
 		}
 		
 		public void Show (Control control, Point position)
 		{
 			if (control == null)
 				throw new ArgumentNullException ("control");
-				
-			this.Location = control.PointToScreen (position);
-			Show ();
+			
+			XplatUI.SetOwner (Handle, control.Handle);
+			Show (control.PointToScreen (position), DefaultDropDownDirection);
 		}
 		
 		public void Show (int x, int y)
 		{
-			this.Location = new Point (x, y);
-			Show ();
+			Show (new Point (x, y), DefaultDropDownDirection);
 		}
 		
 		public void Show (Point position, ToolStripDropDownDirection direction)
@@ -394,7 +402,69 @@ namespace System.Windows.Forms
 			this.PerformLayout ();
 			
 			Point show_point = position;
+			Point max_screen = new Point (SystemInformation.VirtualScreen.Width, SystemInformation.VirtualScreen.Height);
+			
+			if (this is ContextMenuStrip) {
+				// If we are going to go offscreen, adjust our direction so we don't...
+				// X direction
+				switch (direction) {
+					case ToolStripDropDownDirection.AboveLeft:
+						if (show_point.X - this.Width < 0)
+							direction = ToolStripDropDownDirection.AboveRight;
+						break;
+					case ToolStripDropDownDirection.BelowLeft:
+						if (show_point.X - this.Width < 0)
+							direction = ToolStripDropDownDirection.BelowRight;
+						break;
+					case ToolStripDropDownDirection.Left:
+						if (show_point.X - this.Width < 0)
+							direction = ToolStripDropDownDirection.Right;
+						break;
+					case ToolStripDropDownDirection.AboveRight:
+						if (show_point.X + this.Width > max_screen.X)
+							direction = ToolStripDropDownDirection.AboveLeft;
+						break;
+					case ToolStripDropDownDirection.BelowRight:
+					case ToolStripDropDownDirection.Default:
+						if (show_point.X + this.Width > max_screen.X)
+							direction = ToolStripDropDownDirection.BelowLeft;
+						break;
+					case ToolStripDropDownDirection.Right:
+						if (show_point.X + this.Width > max_screen.X)
+							direction = ToolStripDropDownDirection.Left;
+						break;
+				}
 
+				// Y direction
+				switch (direction) {
+					case ToolStripDropDownDirection.AboveLeft:
+						if (show_point.Y - this.Height < 0)
+							direction = ToolStripDropDownDirection.BelowLeft;
+						break;
+					case ToolStripDropDownDirection.AboveRight:
+						if (show_point.Y - this.Height < 0)
+							direction = ToolStripDropDownDirection.BelowRight;
+						break;
+					case ToolStripDropDownDirection.BelowLeft:
+						if (show_point.Y + this.Height > max_screen.Y)
+							direction = ToolStripDropDownDirection.AboveLeft;
+						break;
+					case ToolStripDropDownDirection.BelowRight:
+					case ToolStripDropDownDirection.Default:
+						if (show_point.Y + this.Height > max_screen.Y)
+							direction = ToolStripDropDownDirection.AboveRight;
+						break;
+					case ToolStripDropDownDirection.Left:
+						if (show_point.Y + this.Height > max_screen.Y)
+							direction = ToolStripDropDownDirection.AboveLeft;
+						break;
+					case ToolStripDropDownDirection.Right:
+						if (show_point.Y + this.Height > max_screen.Y)
+							direction = ToolStripDropDownDirection.AboveRight;
+						break;
+				}
+			}
+		
 			switch (direction) {
 				case ToolStripDropDownDirection.AboveLeft:
 					show_point.Y -= this.Height;
@@ -412,11 +482,31 @@ namespace System.Windows.Forms
 				case ToolStripDropDownDirection.Right:
 					break;
 			}
+
+			// Fix offscreen positions
+			if ((show_point.X + this.Width) > max_screen.X)
+				show_point.X = max_screen.X - this.Width;
+			if (show_point.X < 0)
+				show_point.X = 0;
 			
 			if (this.Location != show_point)
 				this.Location = show_point;
-				
-			Show ();
+
+			CancelEventArgs e = new CancelEventArgs ();
+			this.OnOpening (e);
+
+			if (e.Cancel)
+				return;
+
+			// The tracker lets us know when the form is clicked or loses focus
+			ToolStripManager.AppClicked += new EventHandler (ToolStripMenuTracker_AppClicked);
+			ToolStripManager.AppFocusChange += new EventHandler (ToolStripMenuTracker_AppFocusChange);
+
+			base.Show ();
+
+			ToolStripManager.SetActiveToolStrip (this, ToolStripManager.ActivatedByKeyboard);
+
+			this.OnOpened (EventArgs.Empty);
 		}
 		
 		public void Show (Control control, int x, int y)
@@ -432,11 +522,17 @@ namespace System.Windows.Forms
 			if (control == null)
 				throw new ArgumentNullException ("control");
 
+			XplatUI.SetOwner (Handle, control.Handle);
 			Show (control.PointToScreen (position), direction);
 		}
 		#endregion
 
 		#region Protected Methods
+		protected override AccessibleObject CreateAccessibilityInstance ()
+		{
+			return new ToolStripDropDownAccessibleObject (this);
+		}
+		
 		protected override void CreateHandle ()
 		{
 			base.CreateHandle ();
@@ -469,6 +565,9 @@ namespace System.Windows.Forms
 		protected override void OnHandleCreated (EventArgs e)
 		{
 			base.OnHandleCreated (e);
+
+			if (Application.MWFThread.Current.Context != null && Application.MWFThread.Current.Context.MainForm != null)
+				XplatUI.SetOwner (this.Handle, Application.MWFThread.Current.Context.MainForm.Handle);
 		}
 
 		protected override void OnItemClicked (ToolStripItemClickedEventArgs e)
@@ -558,6 +657,13 @@ namespace System.Windows.Forms
 
 		protected override bool ProcessDialogKey (Keys keyData)
 		{
+			// We don't want to let our base change the active ToolStrip
+			switch (keyData) {
+				case Keys.Control | Keys.Tab:
+				case Keys.Control | Keys.Shift | Keys.Tab:
+					return true;
+			}
+			
 			return base.ProcessDialogKey (keyData);
 		}
 
@@ -566,6 +672,11 @@ namespace System.Windows.Forms
 			return base.ProcessMnemonic (charCode);
 		}
 
+		protected override void ScaleControl (SizeF factor, BoundsSpecified specified)
+		{
+			base.ScaleControl (factor, specified);
+		}
+		
 		[EditorBrowsable (EditorBrowsableState.Never)]
 		protected override void ScaleCore (float dx, float dy)
 		{
@@ -577,9 +688,9 @@ namespace System.Windows.Forms
 			base.SetBoundsCore (x, y, width, height, specified);
 		}
 
-		protected override void SetVisibleCore (bool value)
+		protected override void SetVisibleCore (bool visible)
 		{
-			base.SetVisibleCore (value);
+			base.SetVisibleCore (visible);
 		}
 
 		protected override void WndProc (ref Message m)
@@ -797,19 +908,118 @@ namespace System.Windows.Forms
 		#endregion
 
 		#region Private Methods
+		internal override void Dismiss (ToolStripDropDownCloseReason reason)
+		{
+			this.Close (reason);
+			base.Dismiss (reason);
+		}
+
+		internal override ToolStrip GetTopLevelToolStrip ()
+		{
+			if (this.OwnerItem == null)
+				return this;
+				
+			return this.OwnerItem.GetTopLevelToolStrip ();
+		}
+
+		internal override bool ProcessArrowKey (Keys keyData)
+		{
+			switch (keyData) {
+				case Keys.Down:
+				case Keys.Tab:
+					this.SelectNextToolStripItem (this.GetCurrentlySelectedItem (), true);
+					return true;
+				case Keys.Up:
+				case Keys.Shift | Keys.Tab:
+					this.SelectNextToolStripItem (this.GetCurrentlySelectedItem (), false);
+					return true;
+				case Keys.Right:
+					this.GetTopLevelToolStrip ().SelectNextToolStripItem (this.TopLevelOwnerItem, true);
+					return true;
+				case Keys.Left:
+				case Keys.Escape:
+					this.Dismiss (ToolStripDropDownCloseReason.Keyboard);
+					
+					ToolStrip parent_strip = this.OwnerItem.Parent;
+					ToolStripManager.SetActiveToolStrip (parent_strip, true);
+					
+					if (parent_strip is MenuStrip && keyData == Keys.Left) {
+						parent_strip.SelectNextToolStripItem (this.TopLevelOwnerItem, false);
+						this.TopLevelOwnerItem.Invalidate ();
+					} else if (parent_strip is MenuStrip && keyData == Keys.Escape) {
+						(parent_strip as MenuStrip).MenuDroppedDown = false;
+						this.TopLevelOwnerItem.Select ();
+					}				
+					return true;
+			}
+			
+			return false;
+		}
+
+		internal override ToolStripItem SelectNextToolStripItem (ToolStripItem start, bool forward)
+		{
+			ToolStripItem next_item = this.GetNextItem (start, forward ? ArrowDirection.Down : ArrowDirection.Up);
+
+			if (next_item != null)
+				this.ChangeSelection (next_item);
+				
+			return (next_item);
+		}
+		
 		private void ToolStripMenuTracker_AppFocusChange (object sender, EventArgs e)
 		{
-			this.Close (ToolStripDropDownCloseReason.AppFocusChange);
+			this.GetTopLevelToolStrip ().Dismiss (ToolStripDropDownCloseReason.AppFocusChange);
 		}
 
 		private void ToolStripMenuTracker_AppClicked (object sender, EventArgs e)
 		{
-			this.Close (ToolStripDropDownCloseReason.AppClicked);
+			this.GetTopLevelToolStrip ().Dismiss (ToolStripDropDownCloseReason.AppClicked);
 		}
 		#endregion
 
 		#region Internal Properties
 		internal override bool ActivateOnShow { get { return false; } }
+		
+		internal ToolStripItem TopLevelOwnerItem {
+			get {
+				ToolStripItem owner_item = this.OwnerItem;
+				ToolStrip ts = null;
+
+				while (owner_item != null) {
+					ts = owner_item.Owner;
+
+					if (ts != null && (ts is ToolStripDropDown))
+						owner_item = (ts as ToolStripDropDown).OwnerItem;
+					else
+						return owner_item;
+				}
+
+				return null;
+			}
+		}
+		#endregion
+
+		#region ToolStripDropDownAccessibleObject
+		[ComVisible (true)]
+		public class ToolStripDropDownAccessibleObject : ToolStripAccessibleObject
+		{
+			#region Public Constructor
+			public ToolStripDropDownAccessibleObject (ToolStripDropDown owner) : base (owner)
+			{
+			}
+			#endregion
+			
+			#region Public Properties
+			public override string Name {
+				get { return base.Name; }
+				set { base.Name = value; }
+			}
+
+			public override AccessibleRole Role {
+				get { return AccessibleRole.MenuPopup; }
+			}
+			#endregion
+		}
 		#endregion
 	}
 }
