@@ -99,7 +99,7 @@ namespace MonoTests.System.Windows.Forms {
 		[Test]
 		public void ControlSelectNextFlatTest ()
 		{
-			if (RunningOnUnix) {
+			if (TestHelper.RunningOnUnix) {
 				Assert.Ignore ("Relies on form.Show() synchronously generating WM_ACTIVATE");
 			}
 
@@ -827,7 +827,7 @@ namespace MonoTests.System.Windows.Forms {
 		[Test]
 		public void ActiveControl ()
 		{
-			if (RunningOnUnix) {
+			if (TestHelper.RunningOnUnix) {
 				Assert.Ignore ("Relies on form.Show() synchronously generating WM_ACTIVATE");
 			}
 
@@ -980,7 +980,7 @@ namespace MonoTests.System.Windows.Forms {
 		[Test]
 		public void EnterLeaveFocusEventTest ()
 		{
-			if (RunningOnUnix) {
+			if (TestHelper.RunningOnUnix) {
 				Assert.Ignore ("Relies on form.Show() synchronously generating WM_ACTIVATE");
 			}
 
@@ -1197,13 +1197,102 @@ OnGotFocus: ContainerControl 1 System.Windows.Forms.ContainerControl
 			form.Dispose ();
 		}
 
-		private bool RunningOnUnix {
-			get {
-				// check for Unix platforms - see FAQ for more details
-				// http://www.mono-project.com/FAQ:_Technical#How_to_detect_the_execution_platform_.3F
-				int platform = (int) Environment.OSVersion.Platform;
-				return ((platform == 4) || (platform == 128));
+		/// <summary>
+		/// Check that the active control is set before OnEnter is fired, and
+		/// that the events are fired without a Form
+		/// </summary>
+		[Test]
+		public void ActiveControl4 () {
+			sb = new StringBuilder ();
+
+			ContainerControl cc = new ContainerControl ();
+			WrappedEnterLeave1 c1 = new WrappedEnterLeave1 (sb);
+			WrappedEnterLeave2 c2 = new WrappedEnterLeave2 (sb);
+
+			cc.Controls.Add (c1);
+			cc.Controls.Add (c2);
+
+			sb.Append (":Selecting WrappedEnterLeave1:");
+			c1.Select ();
+			sb.Append (":Selecting WrappedEnterLeave2:");
+			c2.Select ();
+
+			Assert.AreEqual (
+				":Selecting WrappedEnterLeave1:"+
+				":Before OnEnter:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave1:"+
+				":After OnEnter:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave1:"+
+				":Selecting WrappedEnterLeave2:"+
+				":Before OnLeave:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":After OnLeave:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":Before OnValidating:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":After OnValidating:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":Before OnValidated:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":After OnValidated:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":Before OnEnter:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"+
+				":After OnEnter:System.Windows.Forms.ContainerControl:MonoTests.System.Windows.Forms.FocusTest+WrappedEnterLeave2:"
+				, 
+				sb.ToString(), "#A1");
+		}
+
+		class WrappedEnterLeave : Control {
+			StringBuilder sb;
+			public WrappedEnterLeave (StringBuilder sb) {
+				this.sb = sb;
 			}
+
+			protected override void OnEnter (EventArgs e) {
+				IContainerControl c = GetContainerControl ();
+				sb.AppendFormat (":Before OnEnter:{0}:{1}:", c, c.ActiveControl);
+				base.OnEnter (e);
+				sb.AppendFormat (":After OnEnter:{0}:{1}:", c, c.ActiveControl);
+			}
+
+			protected override void OnLeave (EventArgs e) {
+				IContainerControl c = GetContainerControl ();
+				sb.AppendFormat (":Before OnLeave:{0}:{1}:", c, c.ActiveControl);
+				base.OnLeave(e);
+				sb.AppendFormat (":After OnLeave:{0}:{1}:", c, c.ActiveControl);
+			}
+
+			protected override void OnValidated(EventArgs e) {
+				IContainerControl c = GetContainerControl ();
+				sb.AppendFormat (":Before OnValidated:{0}:{1}:", c, c.ActiveControl);
+				base.OnValidated (e);
+				sb.AppendFormat (":After OnValidated:{0}:{1}:", c, c.ActiveControl);
+			}
+
+			protected override void OnValidating(CancelEventArgs e) {
+				IContainerControl c = GetContainerControl ();
+				sb.AppendFormat (":Before OnValidating:{0}:{1}:", c, c.ActiveControl);
+				base.OnValidating (e);
+				sb.AppendFormat (":After OnValidating:{0}:{1}:", c, c.ActiveControl);
+			}
+		}
+
+		class WrappedEnterLeave1: WrappedEnterLeave {
+			public WrappedEnterLeave1 (StringBuilder sb): base(sb){}
+		}
+		class WrappedEnterLeave2: WrappedEnterLeave {
+			public WrappedEnterLeave2 (StringBuilder sb): base(sb){}
+		}
+		
+		// #372616
+		[Test]
+		public void UserControlFocus ()
+		{
+			Form form = new Form ();
+			UserControl c = new UserControl ();
+			TextBox t1 = new TextBox ();
+			TextBox t2 = new TextBox ();
+			form.Controls.Add (t1);
+			c.Controls.Add (t2);
+			form.Controls.Add (c);
+			form.Show ();
+			
+			c.Focus ();
+			
+			Assert.IsTrue (t2.Focused, "#A1");
+			Assert.AreSame (t2, c.ActiveControl, "#A2");
 		}
 	}
 }
