@@ -43,14 +43,14 @@ namespace System.Data.Odbc
 	/// </summary>
 
 #if NET_2_0
-        public sealed class OdbcCommandBuilder : DbCommandBuilder
+	public sealed class OdbcCommandBuilder : DbCommandBuilder
 #else // 1_1
 	public sealed class OdbcCommandBuilder : Component
 #endif // NET_2_0
 	{
 		#region Fields
 
-		private OdbcDataAdapter 	_adapter;
+		private OdbcDataAdapter _adapter;
 		private string 			_quotePrefix;
 		private string 			_quoteSuffix;
 
@@ -62,18 +62,19 @@ namespace System.Data.Odbc
 
 		bool _disposed;
 
+		private OdbcRowUpdatingEventHandler rowUpdatingHandler;
+		
 		#endregion // Fields
 
 		#region Constructors
 		
 		public OdbcCommandBuilder ()
 		{
-			_adapter = null;
-			_quotePrefix = String.Empty;
-			_quoteSuffix = String.Empty;
+			_quotePrefix = string.Empty;
+			_quoteSuffix = string.Empty;
 		}
 
-		public OdbcCommandBuilder (OdbcDataAdapter adapter) 
+		public OdbcCommandBuilder (OdbcDataAdapter adapter)
 			: this ()
 		{
 			DataAdapter = adapter;
@@ -87,27 +88,28 @@ namespace System.Data.Odbc
 		[DefaultValue (null)]
 		public
 #if NET_2_0
-                new
+		new
 #endif // NET_2_0
-                OdbcDataAdapter DataAdapter {
+		OdbcDataAdapter DataAdapter {
 			get {
 				return _adapter;
 			}
 			set {
 				if (_adapter == value)
 					return;
+
+				if (rowUpdatingHandler != null)
+					rowUpdatingHandler = new OdbcRowUpdatingEventHandler (OnRowUpdating);
 				
 				if (_adapter != null)
-					_adapter.RowUpdating -= new OdbcRowUpdatingEventHandler (OnRowUpdating);
+					_adapter.RowUpdating -= rowUpdatingHandler;
 				_adapter = value;
 				if (_adapter != null)
-					_adapter.RowUpdating += new OdbcRowUpdatingEventHandler (OnRowUpdating);
-
+					_adapter.RowUpdating += rowUpdatingHandler;
 			}
 		}
 
-		private OdbcCommand SelectCommand
-		{
+		private OdbcCommand SelectCommand {
 			get {
 				if (DataAdapter == null)
 					return null;
@@ -115,8 +117,7 @@ namespace System.Data.Odbc
 			}
 		}
 
-		private DataTable Schema 
-		{
+		private DataTable Schema {
 			get {
 				if (_schema == null)
 					RefreshSchema ();
@@ -124,10 +125,9 @@ namespace System.Data.Odbc
 			}
 		}
 		
-		private string TableName 
-		{
+		private string TableName {
 			get {
-				if (_tableName != String.Empty)
+				if (_tableName != string.Empty)
 					return _tableName;
 
 				DataRow [] schemaRows = Schema.Select ("BaseTableName is not null and BaseTableName <> ''");
@@ -147,12 +147,13 @@ namespace System.Data.Odbc
 
 		[BrowsableAttribute (false)]
 		[OdbcDescriptionAttribute ("The prefix string wrapped around sql objects")]
-                [DesignerSerializationVisibilityAttribute (DesignerSerializationVisibility.Hidden)]
+		[DesignerSerializationVisibilityAttribute (DesignerSerializationVisibility.Hidden)]
+#if ONLY_1_1
 		public
-#if NET_2_0
-                override
-#endif // NET_2_0
-                string QuotePrefix {
+#else
+		new
+#endif
+		string QuotePrefix {
 			get {
 				return _quotePrefix;
 			}
@@ -162,13 +163,14 @@ namespace System.Data.Odbc
 		}
 
 		[BrowsableAttribute (false)]
-                [OdbcDescriptionAttribute ("The suffix string wrapped around sql objects")]
-                [DesignerSerializationVisibilityAttribute (DesignerSerializationVisibility.Hidden)]
+		[OdbcDescriptionAttribute ("The suffix string wrapped around sql objects")]
+		[DesignerSerializationVisibilityAttribute (DesignerSerializationVisibility.Hidden)]
+#if ONLY_1_1
 		public
-#if NET_2_0
-                override
+#else
+		new
 #endif // NET_2_0
-                string QuoteSuffix {
+		string QuoteSuffix {
 			get {
 				return _quoteSuffix;
 			}
@@ -182,22 +184,31 @@ namespace System.Data.Odbc
 		#region Methods
 
 		[MonoTODO]
-		public static void DeriveParameters (OdbcCommand command) 
+		public static void DeriveParameters (OdbcCommand command)
 		{
 			throw new NotImplementedException ();
 		}
 
-		protected override void Dispose (bool disposing) 
+#if ONLY_1_1
+		protected override
+#else
+		new
+#endif
+		void Dispose (bool disposing)
 		{
 			if (_disposed)
 				return;
 			
 			if (disposing) {
 				// dispose managed resource
-				if (_insertCommand != null) _insertCommand.Dispose ();
-				if (_updateCommand != null) _updateCommand.Dispose ();
-				if (_deleteCommand != null) _deleteCommand.Dispose ();
-				if (_schema != null) _insertCommand.Dispose ();
+				if (_insertCommand != null)
+					_insertCommand.Dispose ();
+				if (_updateCommand != null)
+					_updateCommand.Dispose ();
+				if (_deleteCommand != null)
+					_deleteCommand.Dispose ();
+				if (_schema != null)
+					_insertCommand.Dispose ();
 
 				_insertCommand = null;
 				_updateCommand = null;
@@ -226,7 +237,7 @@ namespace System.Data.Odbc
 				columnName = schemaRow.IsNull ("ColumnName") ? String.Empty : (string) schemaRow ["ColumnName"];
 			return columnName;
 		}
-
+		
 		private OdbcParameter AddParameter (OdbcCommand cmd, string paramName, OdbcType odbcType,
 						    int length, string sourceColumnName, DataRowVersion rowVersion)
 		{
@@ -242,25 +253,29 @@ namespace System.Data.Odbc
 		/*
 		 * creates where clause for optimistic concurrency
 		 */
-		private string CreateOptWhereClause (OdbcCommand command)
+		private string CreateOptWhereClause (OdbcCommand command, bool option)
 		{
 			string [] whereClause = new string [Schema.Rows.Count];
 
 			int count = 0;
 
 			foreach (DataRow schemaRow in Schema.Rows) {
-				
 				// exclude non updatable columns
 				if (! IsUpdatable (schemaRow))
 					continue;
 
-				string columnName = GetColumnName (schemaRow);
+				string columnName = null;
+				if (option)
+					columnName = GetColumnName (schemaRow);
+				else
+					columnName = String.Format ("@p{0}", count);
+				
 				if (columnName == String.Empty)
 					throw new InvalidOperationException ("Cannot form delete command. Column name is missing!");
 
-				bool 	allowNull 	= schemaRow.IsNull ("AllowDBNull") || (bool) schemaRow ["AllowDBNull"];
-				OdbcType sqlDbType 	= schemaRow.IsNull ("ProviderType") ? OdbcType.VarChar : (OdbcType) schemaRow ["ProviderType"];
-				int 	length 		= schemaRow.IsNull ("ColumnSize") ? -1 : (int) schemaRow ["ColumnSize"];
+				bool 	allowNull  = schemaRow.IsNull ("AllowDBNull") || (bool) schemaRow ["AllowDBNull"];
+				OdbcType sqlDbType = schemaRow.IsNull ("ProviderType") ? OdbcType.VarChar : (OdbcType) schemaRow ["ProviderType"];
+				int 	length 	   = schemaRow.IsNull ("ColumnSize") ? -1 : (int) schemaRow ["ColumnSize"];
 
 				if (allowNull) {
 					whereClause [count] = String.Format ("((? = 1 AND {0} IS NULL) OR ({0} = ?))",
@@ -278,25 +293,24 @@ namespace System.Data.Odbc
 			return String.Join (" AND ", whereClause, 0, count);
 		}
 
-		public
-#if NET_2_0
-                new
-#endif // NET_2_0
-                OdbcCommand GetInsertCommand ()
+		private void CreateNewCommand (ref OdbcCommand command)
 		{
-			// FIXME: check validity of adapter
-			if (_insertCommand != null)
-				return _insertCommand;
-
-			if (_schema == null)
-				RefreshSchema ();
+			OdbcCommand sourceCommand = SelectCommand;
+			if (command == null) {
+				command = new OdbcCommand ();
+				command.Connection = sourceCommand.Connection;
+				command.CommandTimeout = sourceCommand.CommandTimeout;
+				command.Transaction = sourceCommand.Transaction;
+			}
+			command.CommandType = CommandType.Text;
+			command.UpdatedRowSource = UpdateRowSource.None;
+			command.Parameters.Clear ();
+		}
+		
+		private OdbcCommand CreateInsertCommand (bool option)
+		{
+			CreateNewCommand (ref _insertCommand);
 			
-			_insertCommand = new OdbcCommand ();
-			_insertCommand.Connection = DataAdapter.SelectCommand.Connection;
-			_insertCommand.Transaction = DataAdapter.SelectCommand.Transaction;
-			_insertCommand.CommandType = CommandType.Text;
-			_insertCommand.UpdatedRowSource = UpdateRowSource.None;
-
 			string query = String.Format ("INSERT INTO {0}", QuoteIdentifier (TableName));
 			string [] columns = new string [Schema.Rows.Count];
 			string [] values  = new string [Schema.Rows.Count];
@@ -304,12 +318,17 @@ namespace System.Data.Odbc
 			int count = 0;
 
 			foreach (DataRow schemaRow in Schema.Rows) {
-				
 				// exclude non updatable columns
 				if (! IsUpdatable (schemaRow))
 					continue;
 
-				string columnName = GetColumnName (schemaRow);
+				string columnName = null;
+				
+				if (option)
+					columnName = GetColumnName (schemaRow);
+				else
+					columnName = String.Format ("@p{0}", count); 
+				
 				if (columnName == String.Empty)
 					throw new InvalidOperationException ("Cannot form insert command. Column name is missing!");
 
@@ -332,9 +351,11 @@ namespace System.Data.Odbc
 			return _insertCommand;
 		}
 
+		public
 #if NET_2_0
-		[MonoTODO]
-		public new OdbcCommand GetInsertCommand (bool option)
+		new
+#endif // NET_2_0
+		OdbcCommand GetInsertCommand ()
 		{
 			// FIXME: check validity of adapter
 			if (_insertCommand != null)
@@ -343,32 +364,26 @@ namespace System.Data.Odbc
 			if (_schema == null)
 				RefreshSchema ();
 
-			if (option == false) {
-				return GetInsertCommand ();
-			} else {
-				throw new NotImplementedException ();
-			}
+			return CreateInsertCommand (false);
 		}
-#endif // NET_2_0
-			
-		public
+
 #if NET_2_0
-                new
-#endif // NET_2_0
-                OdbcCommand GetUpdateCommand ()
+		public new OdbcCommand GetInsertCommand (bool useColumnsForParameterNames)
 		{
 			// FIXME: check validity of adapter
-			if (_updateCommand != null)
-				return _updateCommand;
+			if (_insertCommand != null)
+				return _insertCommand;
 
 			if (_schema == null)
 				RefreshSchema ();
-			
-			_updateCommand = new OdbcCommand ();
-			_updateCommand.Connection = DataAdapter.SelectCommand.Connection;
-			_updateCommand.Transaction = DataAdapter.SelectCommand.Transaction;
-			_updateCommand.CommandType = CommandType.Text;
-			_updateCommand.UpdatedRowSource = UpdateRowSource.None;
+
+			return CreateInsertCommand (useColumnsForParameterNames);
+		}
+#endif // NET_2_0
+
+		private OdbcCommand CreateUpdateCommand (bool option)
+		{
+			CreateNewCommand (ref _updateCommand);
 
 			string query = String.Format ("UPDATE {0} SET", QuoteIdentifier (TableName));
 			string [] setClause = new string [Schema.Rows.Count];
@@ -376,12 +391,16 @@ namespace System.Data.Odbc
 			int count = 0;
 
 			foreach (DataRow schemaRow in Schema.Rows) {
-				
 				// exclude non updatable columns
 				if (! IsUpdatable (schemaRow))
 					continue;
 
-				string columnName = GetColumnName (schemaRow);
+				string columnName = null; 
+				if (option)
+					columnName = GetColumnName (schemaRow);
+				else
+					columnName = String.Format ("@p{0}", count);
+				
 				if (columnName == String.Empty)
 					throw new InvalidOperationException ("Cannot form update command. Column name is missing!");
 
@@ -396,7 +415,7 @@ namespace System.Data.Odbc
 
 			// create where clause. odbc uses positional parameters. so where class
 			// is created seperate from the above loop.
-			string whereClause = CreateOptWhereClause (_updateCommand);
+			string whereClause = CreateOptWhereClause (_updateCommand, option);
 			
 			query = String.Format ("{0} {1} WHERE ({2})", 
 					       query, 
@@ -405,10 +424,12 @@ namespace System.Data.Odbc
 			_updateCommand.CommandText = query;
 			return _updateCommand;
 		}
-
+		
+		public
 #if NET_2_0
-		[MonoTODO]
-                public new OdbcCommand GetUpdateCommand (bool option)
+		new
+#endif // NET_2_0
+		OdbcCommand GetUpdateCommand ()
 		{
 			// FIXME: check validity of adapter
 			if (_updateCommand != null)
@@ -417,44 +438,53 @@ namespace System.Data.Odbc
 			if (_schema == null)
 				RefreshSchema ();
 
-			if (option == false) {
-				return GetUpdateCommand ();
-			} else {
-				throw new NotImplementedException ();
-			}
+			return CreateUpdateCommand (false);
 		}
-#endif // NET_2_0
-			
-		public
+
 #if NET_2_0
-                new
-#endif // NET_2_0
-                OdbcCommand GetDeleteCommand ()
+		public new OdbcCommand GetUpdateCommand (bool useColumnsForParameterNames)
 		{
 			// FIXME: check validity of adapter
-			if (_deleteCommand != null)
-				return _deleteCommand;
+			if (_updateCommand != null)
+				return _updateCommand;
 
 			if (_schema == null)
 				RefreshSchema ();
-			
-			_deleteCommand = new OdbcCommand ();
-			_deleteCommand.Connection = DataAdapter.SelectCommand.Connection;
-			_deleteCommand.Transaction = DataAdapter.SelectCommand.Transaction;
-			_deleteCommand.CommandType = CommandType.Text;
-			_deleteCommand.UpdatedRowSource = UpdateRowSource.None;
+
+			return CreateUpdateCommand (useColumnsForParameterNames);
+		}
+#endif // NET_2_0
+
+		private OdbcCommand CreateDeleteCommand (bool option)
+		{
+			CreateNewCommand (ref _deleteCommand);
 
 			string query = String.Format ("DELETE FROM {0}", QuoteIdentifier (TableName));
-			string whereClause = CreateOptWhereClause (_deleteCommand);
+			string whereClause = CreateOptWhereClause (_deleteCommand, option);
 			
 			query = String.Format ("{0} WHERE ({1})", query, whereClause);
 			_deleteCommand.CommandText = query;
 			return _deleteCommand;
 		}
 
+		public
 #if NET_2_0
-		[MonoTODO]
-                public new OdbcCommand GetDeleteCommand (bool option)
+		new
+#endif // NET_2_0
+		OdbcCommand GetDeleteCommand ()
+		{
+			// FIXME: check validity of adapter
+			if (_deleteCommand != null)
+				return _deleteCommand;
+
+			if (_schema == null)
+				RefreshSchema ();
+			
+			return CreateDeleteCommand (false);
+		}
+
+#if NET_2_0
+		public new OdbcCommand GetDeleteCommand (bool useColumnsForParameterNames)
 		{
 			// FIXME: check validity of adapter
 			if (_deleteCommand != null)
@@ -463,19 +493,16 @@ namespace System.Data.Odbc
 			if (_schema == null)
 				RefreshSchema ();
 
-			if (option == false) {
-				return GetDeleteCommand ();
-			} else {
-				throw new NotImplementedException ();
-			}
+			return CreateDeleteCommand (useColumnsForParameterNames);
 		}
 #endif // NET_2_0
 
+#if ONLY_1_1
 		public
-#if NET_2_0
-                override
+#else
+		new
 #endif // NET_2_0
-                void RefreshSchema ()
+		void RefreshSchema ()
 		{
 			// creates metadata
 			if (SelectCommand == null)
@@ -499,73 +526,86 @@ namespace System.Data.Odbc
 			_deleteCommand 	= null;
 			_tableName	= String.Empty;
 		}
-                
+
 #if NET_2_0
-                [MonoTODO]
-                protected override void ApplyParameterInfo (DbParameter dbParameter,
-							    DataRow row,
-							    StatementType statementType,
-							    bool whereClause)
-                {
-                        throw new NotImplementedException ();
-                }
+		protected override void ApplyParameterInfo (DbParameter parameter,
+		                                            DataRow row,
+		                                            StatementType statementType,
+		                                            bool whereClause)
+		{
+			OdbcParameter odbcParam = (OdbcParameter) parameter;
+			odbcParam.Size = int.Parse (row ["ColumnSize"].ToString ());
+			if (row ["NumericPrecision"] != DBNull.Value)
+				odbcParam.Precision = byte.Parse (row ["NumericPrecision"].ToString ());
+			if (row ["NumericScale"] != DBNull.Value)
+				odbcParam.Scale = byte.Parse (row ["NumericScale"].ToString ());
+			odbcParam.DbType = (DbType) row ["ProviderType"];
+		}
 
-                [MonoTODO]
-                protected override string GetParameterName (int position)
-                {
-                        throw new NotImplementedException ();                        
-                }
+		protected override string GetParameterName (int parameterOrdinal)
+		{
+			return String.Format("@p{0}", parameterOrdinal);
+		}
 
-                [MonoTODO]
-                protected override string GetParameterName (string parameterName)
-                {
-                        throw new NotImplementedException ();                        
-                }
-                
-                [MonoTODO]
-                protected override string GetParameterPlaceholder (int position)
-                {
-                        throw new NotImplementedException ();                        
-                }
-                
-                [MonoTODO]
-                protected override void SetRowUpdatingHandler (DbDataAdapter adapter)
-                {
-                        throw new NotImplementedException ();
-                }
+		protected override string GetParameterName (string parameterName)
+		{
+			return String.Format("@{0}", parameterName);
+		}
 
+		protected override string GetParameterPlaceholder (int parameterOrdinal)
+		{
+			return GetParameterName (parameterOrdinal);
+		}
+
+		// FIXME: According to MSDN - "if this method is called again with
+		// the same DbDataAdapter, the DbCommandBuilder is unregistered for 
+		// that DbDataAdapter's RowUpdating event" - this behaviour is yet
+		// to be verified
+		protected override void SetRowUpdatingHandler (DbDataAdapter adapter)
+		{
+			if (!(adapter is OdbcDataAdapter))
+				throw new InvalidOperationException ("Adapter needs to be a SqlDataAdapter");
+			if (rowUpdatingHandler == null)
+				rowUpdatingHandler = new OdbcRowUpdatingEventHandler (OnRowUpdating);
+
+			((OdbcDataAdapter) adapter).RowUpdating += rowUpdatingHandler;
+		}
 #endif // NET_2_0
-		
+
 #if NET_2_0
-		[MonoTODO]
 		public override
 #else
 		private
-#endif		
+#endif
 		string QuoteIdentifier (string unquotedIdentifier)
 		{
-#if NET_2_0
-			throw new NotImplementedException ();
-#else
-			if (unquotedIdentifier == null || unquotedIdentifier == String.Empty)
+			if (unquotedIdentifier == null || unquotedIdentifier.Length == 0)
 				return unquotedIdentifier;
 			return String.Format ("{0}{1}{2}", QuotePrefix, 
-					      unquotedIdentifier, QuoteSuffix);
-#endif			
+				unquotedIdentifier, QuoteSuffix);
 		}
 
 #if NET_2_0
-		[MonoTODO]
+		// FIXME:  Not sure what the extra "connection" param does!
+		public string QuoteIdentifier (string unquotedIdentifier, OdbcConnection connection)
+		{
+			return QuoteIdentifier (unquotedIdentifier);
+		}
+
+		public string UnquoteIdentifier (string quotedIdentifier, OdbcConnection connection)
+		{
+			return UnquoteIdentifier (quotedIdentifier);
+		}
+#endif
+
+#if NET_2_0
 		public override
 #else
 		private
-#endif		
+#endif
 		string UnquoteIdentifier (string quotedIdentifier)
 		{
-#if NET_2_0
-			throw new NotImplementedException ();
-#else
-			if (quotedIdentifier == null || quotedIdentifier == String.Empty)
+			if (quotedIdentifier == null || quotedIdentifier.Length == 0)
 				return quotedIdentifier;
 			
 			StringBuilder sb = new StringBuilder (quotedIdentifier.Length);
@@ -575,7 +615,6 @@ namespace System.Data.Odbc
 			if (quotedIdentifier.EndsWith (QuoteSuffix))
 				sb.Remove (sb.Length - QuoteSuffix.Length, QuoteSuffix.Length );
 			return sb.ToString ();
-#endif			
 		}
 
 		private void OnRowUpdating (object sender, OdbcRowUpdatingEventArgs args)
@@ -599,7 +638,6 @@ namespace System.Data.Odbc
 				args.Status = UpdateStatus.ErrorsOccurred;
 			}
 		}
-		
 
 		#endregion // Methods
 	}
