@@ -64,10 +64,13 @@ namespace System.Data {
 	[DesignTimeVisible (false)]
 	[EditorAttribute ("Microsoft.VSDesigner.Data.Design.DataTableEditor, "+ Consts.AssemblyMicrosoft_VSDesigner, "System.Drawing.Design.UITypeEditor, "+ Consts.AssemblySystem_Drawing )]
 	[Serializable]
+#if NET_2_0
+	[XmlSchemaProviderAttribute ("GetDataTableSchema")]
+#endif
 	public class DataTable : MarshalByValueComponent, IListSource, ISupportInitialize,
 				 ISerializable
 #if NET_2_0
-				 , ISupportInitializeNotification
+				 , ISupportInitializeNotification, IXmlSerializable
 #endif
 	{
 		#region Fields
@@ -518,6 +521,7 @@ namespace System.Data {
 			DataTableNewRowEventArgs e = new DataTableNewRowEventArgs (dr);
 			OnTableNewRow (e);
 		}
+
 		private void DataTableInitialized ()
 		{
 			EventArgs e = new EventArgs ();
@@ -679,6 +683,10 @@ namespace System.Data {
 			}
 		}
 
+		internal bool LocaleSpecified {
+			get { return _locale != null; }
+		}
+
 		/// <summary>
 		/// Gets or sets the initial starting size for this table.
 		/// </summary>
@@ -1182,7 +1190,7 @@ namespace System.Data {
 				for (int i=0; i < ExtendedProperties.Count; i++)
 					Copy.ExtendedProperties.Add (tgtArray.GetValue (i), ExtendedProperties[tgtArray.GetValue (i)]);
 			}
-			Copy.Locale = Locale;
+			Copy._locale = _locale;
 			Copy.MinimumCapacity = MinimumCapacity;
 			Copy.Namespace = Namespace;
 			// Copy.ParentRelations
@@ -1691,6 +1699,42 @@ namespace System.Data {
 		}
 
 #if NET_2_0
+
+		#region IXmlSerializable Members
+
+		[MonoNotSupported("")]
+		XmlSchema IXmlSerializable.GetSchema () 
+		{
+			return GetSchema ();
+		}
+
+		void IXmlSerializable.ReadXml (XmlReader reader) 
+		{
+			ReadXml_internal (reader, true);
+		}
+
+		void IXmlSerializable.WriteXml (XmlWriter writer) 
+		{
+			DataSet dset = dataSet;
+			bool isPartOfDataSet = true;
+
+			if (dataSet == null) {
+				dset = new DataSet ();
+				dset.Tables.Add (this);
+				isPartOfDataSet = false;
+			}
+
+			XmlSchemaWriter.WriteXmlSchema (writer, new DataTable [] { this }, 
+											null, TableName, dset.DataSetName, LocaleSpecified ? Locale : dset.LocaleSpecified ? dset.Locale : null);
+			dset.WriteIndividualTableContent (writer, this, XmlWriteMode.DiffGram);
+			writer.Flush ();
+
+			if (!isPartOfDataSet)
+				dataSet.Tables.Remove(this);
+		}
+
+		#endregion
+
                 /// <summary>
                 ///     Loads the table with the values from the reader
                 /// </summary>
@@ -1945,7 +1989,8 @@ namespace System.Data {
 			// initialize default values row for the first time
 			if ( _defaultValuesRowIndex == -1 ) {
 				_defaultValuesRowIndex = RecordCache.NewRecord();
-				foreach(DataColumn column in Columns) {
+				for (int i = 0; i < Columns.Count; ++i) {
+					DataColumn column = Columns [i];
 					column.DataContainer[_defaultValuesRowIndex] = column.DefaultValue;
 				}
 			}
@@ -1969,7 +2014,12 @@ namespace System.Data {
 		/// </summary>
 		protected internal DataRow[] NewRowArray (int size) 
 		{
-			return (DataRow[]) Array.CreateInstance (GetRowType (), size);
+			Type t = GetRowType ();
+			/* Avoid reflection if possible */
+			if (t == typeof (DataRow))
+				return new DataRow [size];
+			else
+				return (DataRow[]) Array.CreateInstance (GetRowType (), size);
 		}
 
 		/// <summary>
@@ -1994,10 +2044,9 @@ namespace System.Data {
 			throw new NotImplementedException ();
 		}
 
-		[MonoTODO]
 		public static XmlSchemaComplexType GetDataTableSchema (XmlSchemaSet schemaSet)
 		{
-			throw new NotImplementedException ();
+			return new XmlSchemaComplexType ();
 		}
 
 		public XmlReadMode ReadXml (Stream stream)
@@ -2021,6 +2070,11 @@ namespace System.Data {
 		}
 
 		public XmlReadMode ReadXml (XmlReader reader)
+		{
+			return ReadXml_internal (reader, false);
+		}
+
+		public XmlReadMode ReadXml_internal (XmlReader reader, bool serializable) 
 		{
 			// The documentation from MS for this method is rather
 			// poor.  The following cases have been observed
@@ -2047,7 +2101,7 @@ namespace System.Data {
 			DataSet ds = new DataSet ();
 
 			reader.MoveToContent ();
-			if (Columns.Count > 0 && reader.LocalName != "diffgram") 
+			if (Columns.Count > 0 && reader.LocalName != "diffgram" || serializable)
 				mode = ds.ReadXml (reader);
 			else if (Columns.Count > 0 && reader.LocalName == "diffgram") {
 				  try {
@@ -2164,7 +2218,7 @@ namespace System.Data {
 				return;
 
 			DataSet ds = new DataSet ();
-			new XmlSchemaDataImporter (ds, reader).Process ();
+			new XmlSchemaDataImporter (ds, reader, false).Process ();
 			DataTable target = null;
 			if (TableName == String.Empty) {
 				if (ds.Tables.Count > 0)
@@ -2178,6 +2232,12 @@ namespace System.Data {
 			if (target != null)
 				target.CopyProperties (this);
 		}
+
+		[MonoNotSupported("")]
+		protected virtual void ReadXmlSerializable (XmlReader reader)
+		{
+			throw new NotImplementedException ();
+		}
 #endif
 
 		/// <summary>
@@ -2266,13 +2326,13 @@ namespace System.Data {
 				columns = ParseSortString (this, sort, out sorts, false);
 
 			if (Rows.Count == 0)
-				return new DataRow [0];
+				return NewRowArray (0);
 
 			//if sort order is not given, sort it in Ascending order of the
 			//columns involved in the filter
 			if (columns.Length ==0 && filter != null) {
 				ArrayList list = new ArrayList ();
-				for (int i=0, j=0; i < Columns.Count; ++i) {
+				for (int i=0; i < Columns.Count; ++i) {
 					if (!filter.DependsOn (Columns [i]))
 						continue;
 					list.Add (Columns [i]);
@@ -2402,10 +2462,9 @@ namespace System.Data {
 		}
 
 		internal void AddRowToIndexes (DataRow row) {
-			if (_indexes != null) {
-				foreach (Index indx in _indexes) {
-					indx.Add (row);
-				}
+				if (_indexes != null) {
+				for (int i = 0; i < _indexes.Count; ++i)
+					((Index)_indexes [i]).Add (row);
 			}
 		}
 
@@ -2609,6 +2668,8 @@ namespace System.Data {
 
 		public void WriteXmlSchema (XmlWriter writer)
 		{
+			WriteXmlSchema (writer, false);
+/*
 			if (TableName == "") {
 				throw new InvalidOperationException ("Cannot serialize the DataTable. DataTable name is not set.");
 			}
@@ -2634,6 +2695,7 @@ namespace System.Data {
 				if (tmp != null)
 					ds.Tables.Remove (this);
 			}
+*/
 		}
 
 		public void WriteXmlSchema (string fileName)
@@ -2682,10 +2744,10 @@ namespace System.Data {
 			if (TableName == "") {
 				throw new InvalidOperationException ("Cannot serialize the DataTable. DataTable name is not set.");
 			}
-			if (writeHierarchy == false) {
-				WriteXmlSchema (writer);
-			}
-			else {
+//			if (writeHierarchy == false) {
+//				WriteXmlSchema (writer);
+//			}
+//			else {
 				DataSet ds = DataSet;
 				DataSet tmp = null;
 				try {
@@ -2702,7 +2764,7 @@ namespace System.Data {
 					//	relations[i] = ds.Relations[i];
 					//}
 					DataRelation [] relations = null;
-					if (ChildRelations.Count > 0) {
+					if (writeHierarchy && ChildRelations.Count > 0) {
 						relations = new DataRelation [ChildRelations.Count];
 						for (int i = 0; i < ChildRelations.Count; i++) {
 							relations [i] = ChildRelations [i];
@@ -2719,12 +2781,12 @@ namespace System.Data {
 						tableName = TableName;
 					else
 						tableName = ds.Namespace + "_x003A_" + TableName;
-					XmlSchemaWriter.WriteXmlSchema (writer, tables, relations, tableName, ds.DataSetName);
+					XmlSchemaWriter.WriteXmlSchema (writer, tables, relations, tableName, ds.DataSetName, LocaleSpecified ? Locale : ds.LocaleSpecified ? ds.Locale : null);
 				} finally {
 					if (tmp != null)
 						ds.Tables.Remove (this);
 				}
-			}
+//			}
 		}
 
 		public void WriteXmlSchema (string fileName, bool writeHierarchy)
@@ -2960,11 +3022,11 @@ namespace System.Data {
 					string rawColumnName = columnExpression[c].Trim ();
 
 					Match match = SortRegex.Match (rawColumnName);
-					CaptureCollection cc = match.Groups["ColName"].Captures ;
-					if (cc.Count == 0)
+					Group g = match.Groups["ColName"] ;
+					if (!g.Success)
 						throw new IndexOutOfRangeException ("Could not find column: " + rawColumnName);
 
-					string columnName = cc[0].Value;
+					string columnName = g.Value;
 					DataColumn dc = table.Columns[columnName];
 					if (dc == null){
 						try {
@@ -2975,8 +3037,8 @@ namespace System.Data {
 					}
 					columns.Add (dc);
 
-					cc = match.Groups["Order"].Captures;
-					if (cc.Count == 0 || String.Compare (cc [0].Value, "ASC", true, CultureInfo.InvariantCulture) == 0)
+					g = match.Groups["Order"];
+					if (!g.Success || String.Compare (g.Value, "ASC", true, CultureInfo.InvariantCulture) == 0)
 						sorts.Add(ListSortDirection.Ascending);
 					else
 						sorts.Add (ListSortDirection.Descending);
