@@ -36,6 +36,7 @@ using System.IO;
 using System.Text;
 using System.Web.Caching;
 using System.Web.Configuration;
+using System.Web.Hosting;
 using System.Web.UI;
 using System.Web.UI.HtmlControls;
 using System.Web.Util;
@@ -189,7 +190,8 @@ namespace System.Web.Compilation
 		static Hashtable emptyHash = new Hashtable ();
 		bool inForm;
 		bool useOtherTags;
-
+		TagType lastTag;
+		
 		public AspGenerator (TemplateParser tparser)
 		{
 			this.tparser = tparser;
@@ -231,62 +233,152 @@ namespace System.Web.Compilation
 
 			throw new Exception ("Got type: " + type);
 		}
-		
-		void InitParser (string filename)
+
+		void InitParser (TextReader reader, string filename)
 		{
-			StreamReader reader = new StreamReader (filename, WebEncoding.FileEncoding);
 			AspParser parser = new AspParser (filename, reader);
-			reader.Close ();
 			parser.Error += new ParseErrorHandler (ParseError);
 			parser.TagParsed += new TagParsedHandler (TagParsed);
 			parser.TextParsed += new TextParsedHandler (TextParsed);
 			if (!pstack.Push (parser))
 				throw new ParseException (Location, "Infinite recursion detected including file: " + filename);
-			tparser.AddDependency (filename);
-		}
 
+			if (filename != "@@inner_string@@") {
+				string arvp = Path.Combine (tparser.BaseVirtualDir, Path.GetFileName (filename));
+				if (VirtualPathUtility.IsAbsolute (arvp))
+					arvp = VirtualPathUtility.ToAppRelative (arvp);
+				
+				tparser.AddDependency (arvp);
+			}
+#if NET_2_0
+			tparser.MD5Checksum = parser.MD5Checksum;
+#endif
+		}
+		
+#if NET_2_0
+		void InitParser (string filename)
+		{
+			StreamReader reader = new StreamReader (filename, WebEncoding.FileEncoding);
+			InitParser (reader, filename);
+		}
+#endif
+		
 		public void Parse (string file)
 		{
-			InitParser (file);
+#if ONLY_1_1
+			Parse (file, true);
+#else
+			Parse (file, false);
+#endif
+		}
+		
+		public void Parse (TextReader reader, string filename, bool doInitParser)
+		{
+			try {
+				isApplication = tparser.DefaultDirectiveName == "application";
 
-			pstack.Parser.Parse ();
-			if (text.Length > 0)
-				FlushText ();
+				if (doInitParser)
+					InitParser (reader, filename);
 
-			pstack.Pop ();
+				pstack.Parser.Parse ();
+				if (text.Length > 0)
+					FlushText ();
+
+				pstack.Pop ();
 
 #if DEBUG
-			PrintTree (rootBuilder, 0);
+				PrintTree (rootBuilder, 0);
 #endif
 
-			if (stack.Count > 1 && pstack.Count == 0)
-				throw new ParseException (stack.Builder.location,
-						"Expecting </" + stack.Builder.TagName + "> " + stack.Builder);
+				if (stack.Count > 1 && pstack.Count == 0)
+					throw new ParseException (stack.Builder.location,
+								  "Expecting </" + stack.Builder.TagName + "> " + stack.Builder);
+			} finally {
+				if (reader != null)
+					reader.Close ();
+			}
+		}
+
+		public void Parse (Stream stream, string filename, bool doInitParser)
+		{
+			Parse (new StreamReader (stream, WebEncoding.FileEncoding), filename, doInitParser);
+		}
+		
+		public void Parse (string filename, bool doInitParser)
+		{
+			StreamReader reader = new StreamReader (filename, WebEncoding.FileEncoding);
+			Parse (reader, filename, doInitParser);
 		}
 
 		public void Parse ()
 		{
+#if NET_2_0
+			string inputFile = tparser.InputFile;
+			TextReader inputReader = tparser.Reader;
+
+			try {			
+				if (String.IsNullOrEmpty (inputFile)) {
+					StreamReader sr = inputReader as StreamReader;
+					if (sr != null) {
+						FileStream fr = sr.BaseStream as FileStream;
+						if (fr != null)
+							inputFile = fr.Name;
+					}
+
+					if (String.IsNullOrEmpty (inputFile))
+						inputFile = "@@inner_string@@";
+				}
+
+				if (inputReader != null) {
+					Parse (inputReader, inputFile, true);
+				} else {
+					if (String.IsNullOrEmpty (inputFile))
+						throw new HttpException ("Parser input file is empty, cannot continue.");
+					inputFile = Path.GetFullPath (inputFile);
+					InitParser (inputFile);
+					Parse (inputFile);
+				}
+			} finally {
+				if (inputReader != null)
+					inputReader.Close ();
+			}
+#else
 			Parse (Path.GetFullPath (tparser.InputFile));
+#endif
 		}
 
 		internal static void AddTypeToCache (ArrayList dependencies, string inputFile, Type type)
 		{
-			CacheDependency cd = new CacheDependency ((string[])dependencies.ToArray (typeof (string)));
-			HttpRuntime.Cache.InsertPrivate ("@@Type" + inputFile, type, cd);
+			if (type == null || inputFile == null || inputFile.Length == 0)
+				return;
+
+			if (dependencies != null && dependencies.Count > 0) {
+				string [] deps = (string []) dependencies.ToArray (typeof (string));
+				HttpContext ctx = HttpContext.Current;
+				HttpRequest req = ctx != null ? ctx.Request : null;
+				
+				if (req == null)
+					throw new HttpException ("No current context, cannot compile.");
+
+				for (int i = 0; i < deps.Length; i++)
+					deps [i] = req.MapPath (deps [i]);			
+
+				HttpRuntime.InternalCache.Insert ("@@Type" + inputFile, type, new CacheDependency (deps));
+			} else
+				HttpRuntime.InternalCache.Insert ("@@Type" + inputFile, type);
 		}
 		
 		public Type GetCompiledType ()
 		{
-			Type type = (Type) HttpRuntime.Cache.Get ("@@Type" + tparser.InputFile);
+			Type type = (Type) HttpRuntime.InternalCache.Get ("@@Type" + tparser.InputFile);
 			if (type != null) {
 				return type;
 			}
 
-			isApplication = tparser.DefaultDirectiveName == "application";
 			Parse ();
 
 			BaseCompiler compiler = GetCompilerFromType ();
-
+			
 			type = compiler.GetCompiledType ();
 			AddTypeToCache (tparser.Dependencies, tparser.InputFile, type);
 			return type;
@@ -334,7 +426,7 @@ namespace System.Web.Compilation
 				tparser.Location = location;
 
 			if (text.Length != 0)
-				FlushText ();
+				FlushText (lastTag == TagType.CodeRender);
 
 			if (0 == String.Compare (tagid, "script", true, CultureInfo.InvariantCulture)) {
 				bool in_script = (inScript || ignore_text);
@@ -344,6 +436,7 @@ namespace System.Web.Compilation
 				}
 			}
 
+			lastTag = tagtype;
 			switch (tagtype) {
 			case TagType.Directive:
 				if (tagid == "")
@@ -384,6 +477,7 @@ namespace System.Web.Compilation
 			case TagType.CodeRenderExpression:
 				goto case TagType.CodeRender;
 			case TagType.CodeRender:
+				lastTag = TagType.CodeRender;
 				if (isApplication)
 					throw new ParseException (location, "Invalid content for application file.");
 			
@@ -399,12 +493,31 @@ namespace System.Web.Compilation
 					file = attributes ["file"] as string;
 
 				if (isvirtual) {
-					file = tparser.MapPath (file);
-				} else {
-					file = GetIncludeFilePath (tparser.BaseDir, file);
-				}
+					bool parsed = false;
+#if NET_2_0
+					VirtualPathProvider vpp = HostingEnvironment.VirtualPathProvider;
 
-				Parse (file);
+					if (vpp.FileExists (file)) {
+						VirtualFile vf = vpp.GetFile (file);
+						if (vf != null) {
+							Parse (vf.Open (), file, true);
+							parsed = true;
+						}
+					}
+#endif
+					
+					if (!parsed)
+						Parse (tparser.MapPath (file), true);
+				} else {
+					string includeFilePath = GetIncludeFilePath (tparser.ParserDir, file);
+					tparser.PushIncludeDir (Path.GetDirectoryName (includeFilePath));
+					try {
+						Parse (includeFilePath, true);
+					} finally {
+						tparser.PopIncludeDir ();
+					}
+				}
+				
 				break;
 			default:
 				break;
@@ -444,23 +557,31 @@ namespace System.Web.Compilation
 
 			if (text.IndexOf ("<%") != -1 && !inScript) {
 				if (this.text.Length > 0)
-					FlushText ();
+					FlushText (true);
 				CodeRenderParser r = new CodeRenderParser (text, stack.Builder);
 				r.AddChildren ();
 				return;
 			}
-
+			
 			this.text.Append (text);
 			//PrintLocation (location);
 		}
 
 		void FlushText ()
+		{
+			FlushText (false);
+		}
+		
+		void FlushText (bool ignoreEmptyString)
 		{
 			string t = text.ToString ();
 			text.Length = 0;
+
+			if (ignoreEmptyString && t.Trim ().Length == 0)
+				return;
+			
 			if (inScript) {
-				// TODO: store location
-				tparser.Scripts.Add (t);
+				tparser.Scripts.Add (new ServerSideScript (t, new System.Web.Compilation.Location (tparser.Location)));
 				return;
 			}
 
@@ -474,6 +595,56 @@ namespace System.Web.Compilation
 			}
 		}
 
+#if NET_2_0
+		bool BuilderHasOtherThan (Type type, ControlBuilder cb)
+		{
+			ArrayList al = cb.OtherTags;
+			if (al != null && al.Count > 0)
+				return true;
+			
+			al = cb.Children;
+			if (al != null) {
+				ControlBuilder tmp;
+				
+				foreach (object o in al) {
+					if (o == null)
+						continue;
+					
+					tmp = o as ControlBuilder;
+					if (tmp == null) {
+						string s = o as string;
+						if (s != null && String.IsNullOrEmpty (s.Trim ()))
+							continue;
+						
+						return true;
+					}
+					
+					if (tmp is System.Web.UI.WebControls.ContentBuilderInternal)
+						continue;
+					
+					if (tmp.ControlType != typeof (System.Web.UI.WebControls.Content))
+						return true;
+				}
+			}
+
+			return false;
+		}
+		
+		bool OtherControlsAllowed (ControlBuilder cb)
+		{
+			if (cb == null)
+				return true;
+			
+			if (!typeof (System.Web.UI.WebControls.Content).IsAssignableFrom (cb.ControlType))
+				return true;
+
+			if (BuilderHasOtherThan (typeof (System.Web.UI.WebControls.Content), rootBuilder))
+				return false;
+			
+			return true;
+		}
+#endif
+		
 		bool ProcessTag (string tagid, TagAttributes atts, TagType tagtype)
 		{
 			if (isApplication) {
@@ -511,6 +682,11 @@ namespace System.Web.Compilation
 			if (builder == null)
 				return false;
 
+#if NET_2_0
+			if (!OtherControlsAllowed (builder))
+				throw new ParseException (Location, "Only Content controls are allowed directly in a content page that contains Content controls.");
+#endif
+			
 			builder.location = location;
 			builder.ID = htable ["id"] as string;
 			if (typeof (HtmlForm).IsAssignableFrom (builder.ControlType)) {
@@ -557,7 +733,10 @@ namespace System.Web.Compilation
 		{
 			if (tagtype != TagType.Close) {
 				if (attributes != null && attributes.IsRunAtServer ()) {
-					CheckLanguage ((string) attributes ["language"]);
+					string language = (string) attributes ["language"];
+					if (language != null && language.Length > 0 && tparser.ImplicitLanguage)
+						tparser.SetLanguage (language);
+					CheckLanguage (language);
 					string src = (string) attributes ["src"];
 					if (src != null) {
 						if (src == "")
@@ -608,7 +787,7 @@ namespace System.Web.Compilation
 		bool CloseControl (string tagid)
 		{
 			ControlBuilder current = stack.Builder;
-			string btag = current.TagName;
+			string btag = current.OriginalTagName;
 			if (String.Compare (btag, "tbody", true, CultureInfo.InvariantCulture) != 0 &&
 			    String.Compare (tagid, "tbody", true, CultureInfo.InvariantCulture) == 0) {
 				if (!current.ChildrenAsProperties) {
