@@ -42,11 +42,6 @@ namespace System.Web {
 	// CAS - no InheritanceDemand here as the class is sealed
 	[AspNetHostingPermission (SecurityAction.LinkDemand, Level = AspNetHostingPermissionLevel.Minimal)]
 	public sealed class HttpUtility {
-#if NET_2_0
-		internal const char QueryParamSeparator = ';';
-#else
-		internal const char QueryParamSeparator = '&';
-#endif
 		#region Fields
 	
 		static Hashtable entities;
@@ -344,12 +339,21 @@ namespace System.Web {
 			if (null == s) 
 				return null;
 	
-			if (s.IndexOf ('&') == -1 && s.IndexOf ('"') == -1 && s.IndexOf ('<') == -1)
+			bool needEncode = false;
+			for (int i = 0; i < s.Length; i++) {
+				if (s [i] == '&' || s [i] == '"' || s [i] == '<') {
+					needEncode = true;
+					break;
+				}
+			}
+
+			if (!needEncode)
 				return s;
 
 			StringBuilder output = new StringBuilder ();
-			foreach (char c in s) 
-				switch (c) {
+			int len = s.Length;
+			for (int i = 0; i < len; i++)
+				switch (s [i]) {
 				case '&' : 
 					output.Append ("&amp;");
 					break;
@@ -360,7 +364,7 @@ namespace System.Web {
 					output.Append ("&lt;");
 					break;
 				default:
-					output.Append (c);
+					output.Append (s [i]);
 					break;
 				}
 	
@@ -623,8 +627,26 @@ namespace System.Web {
 			if (s == "")
 				return "";
 
-			byte [] bytes = Enc.GetBytes (s);
-			return Encoding.ASCII.GetString (UrlEncodeToBytes (bytes, 0, bytes.Length));
+			bool needEncode = false;
+			int len = s.Length;
+			for (int i = 0; i < len; i++) {
+				char c = s [i];
+				if ((c < '0') || (c < 'A' && c > '9') || (c > 'Z' && c < 'a') || (c > 'z')) {
+					if (NotEncoded (c))
+						continue;
+
+					needEncode = true;
+					break;
+				}
+			}
+
+			if (!needEncode)
+				return s;
+
+			// avoided GetByteCount call
+			byte [] bytes = new byte[Enc.GetMaxByteCount(s.Length)];
+			int realLen = Enc.GetBytes (s, 0, s.Length, bytes, 0);
+			return Encoding.ASCII.GetString (UrlEncodeToBytes (bytes, 0, realLen));
 		}
 	  
 		public static string UrlEncode (byte [] bytes)
@@ -678,7 +700,11 @@ namespace System.Web {
 		}
 
 		static char [] hexChars = "0123456789abcdef".ToCharArray ();
-		const string notEncoded = "!'()*-._";
+
+		static bool NotEncoded (char c)
+		{
+			return (c == '!' || c == '\'' || c == '(' || c == ')' || c == '*' || c == '-' || c == '.' || c == '_');
+		}
 
 		static void UrlEncodeChar (char c, Stream result, bool isUnicode) {
 			if (c > 255) {
@@ -701,7 +727,7 @@ namespace System.Web {
 				return;
 			}
 			
-			if (c>' ' && notEncoded.IndexOf (c)!=-1) {
+			if (c > ' ' && NotEncoded (c)) {
 				result.WriteByte ((byte)c);
 				return;
 			}
@@ -907,10 +933,23 @@ namespace System.Web {
 			if (s == null)
 				return null;
 
+			bool needEncode = false;
+			for (int i = 0; i < s.Length; i++) {
+				char c = s [i];
+				if (c == '&' || c == '"' || c == '<' || c == '>' || c > 159) {
+					needEncode = true;
+					break;
+				}
+			}
+
+			if (!needEncode)
+				return s;
+
 			StringBuilder output = new StringBuilder ();
 			
-			foreach (char c in s) 
-				switch (c) {
+			int len = s.Length;
+			for (int i = 0; i < len; i++) 
+				switch (s [i]) {
 				case '&' :
 					output.Append ("&amp;");
 					break;
@@ -927,12 +966,16 @@ namespace System.Web {
 					// MS starts encoding with &# from 160 and stops at 255.
 					// We don't do that. One reason is the 65308/65310 unicode
 					// characters that look like '<' and '>'.
-					if (c > 159) {
+#if TARGET_JVM
+					if (s [i] > 159 && s [i] < 256) {
+#else
+					if (s [i] > 159) {
+#endif
 						output.Append ("&#");
-						output.Append (((int) c).ToString (CultureInfo.InvariantCulture));
+						output.Append (((int) s [i]).ToString (CultureInfo.InvariantCulture));
 						output.Append (";");
 					} else {
-						output.Append (c);
+						output.Append (s [i]);
 					}
 					break;
 				}
@@ -965,7 +1008,11 @@ namespace System.Web {
 		}
 		
 		static void UrlPathEncodeChar (char c, Stream result) {
+#if NET_2_0
+			if (c < 33 || c > 126) {
+#else
 			if (c > 127) {
+#endif
 				byte [] bIn = Encoding.UTF8.GetBytes (c.ToString ());
 				for (int i = 0; i < bIn.Length; i++) {
 					result.WriteByte ((byte) '%');
@@ -1013,35 +1060,45 @@ namespace System.Web {
 			if (query.Length == 0)
 				return;
 
+			string decoded = HtmlDecode (query);
+			int decodedLength = decoded.Length;
 			int namePos = 0;
-			while (namePos <= query.Length) {
+			bool first = true;
+			while (namePos <= decodedLength) {
 				int valuePos = -1, valueEnd = -1;
-				for (int q = namePos; q < query.Length; q++) {
-					if (valuePos == -1 && query[q] == '=') {
+				for (int q = namePos; q < decodedLength; q++) {
+					if (valuePos == -1 && decoded [q] == '=') {
 						valuePos = q + 1;
-					} else if (query[q] == ';' || query[q] == '&') {
+					} else if (decoded [q] == '&') {
 						valueEnd = q;
 						break;
 					}
 				}
 
+				if (first) {
+					first = false;
+					if (decoded [namePos] == '?')
+						namePos++;
+				}
+				
 				string name, value;
 				if (valuePos == -1) {
 					name = null;
 					valuePos = namePos;
 				} else {
-					name = UrlDecode (query.Substring (namePos, valuePos - namePos - 1), encoding);
+					name = UrlDecode (decoded.Substring (namePos, valuePos - namePos - 1), encoding);
 				}
 				if (valueEnd < 0) {
 					namePos = -1;
-					valueEnd = query.Length;
+					valueEnd = decoded.Length;
 				} else {
 					namePos = valueEnd + 1;
 				}
-				value = UrlDecode (query.Substring (valuePos, valueEnd - valuePos), encoding);
+				value = UrlDecode (decoded.Substring (valuePos, valueEnd - valuePos), encoding);
 
 				result.Add (name, value);
-				if (namePos == -1) break;
+				if (namePos == -1)
+					break;
 			}
 		}
 		#endregion // Methods
