@@ -751,11 +751,11 @@ namespace System.Xml.XPath
 			if (typeR == XPathResultType.Any)
 				typeR = GetReturnType (_right.Evaluate (iter));
 
-			// Regard RTF as nodeset
+			// Regard RTF as string
 			if (typeL == XPathResultType.Navigator)
-				typeL = XPathResultType.NodeSet;
+				typeL = XPathResultType.String;
 			if (typeR == XPathResultType.Navigator)
-				typeR = XPathResultType.NodeSet;
+				typeR = XPathResultType.String;
 
 			if (typeL == XPathResultType.NodeSet || typeR == XPathResultType.NodeSet)
 			{
@@ -851,11 +851,11 @@ namespace System.Xml.XPath
 			if (typeR == XPathResultType.Any)
 				typeR = GetReturnType (_right.Evaluate (iter));
 
-			// Regard RTF as nodeset
+			// Regard RTF as string
 			if (typeL == XPathResultType.Navigator)
-				typeL = XPathResultType.NodeSet;
+				typeL = XPathResultType.String;
 			if (typeR == XPathResultType.Navigator)
-				typeR = XPathResultType.NodeSet;
+				typeR = XPathResultType.String;
 
 			if (typeL == XPathResultType.NodeSet || typeR == XPathResultType.NodeSet)
 			{
@@ -1862,6 +1862,8 @@ namespace System.Xml.XPath
 				else
 					var = context.ResolveVariable (new XmlQualifiedName (_name.Name, _name.Namespace));
 			}
+			else
+				throw new XPathException (String.Format ("XSLT context is required to resolve variable. Current namespace manager in current node-set '{0}' is '{1}'", iter.GetType (), iter.NamespaceManager != null ? iter.NamespaceManager.GetType () : null));
 			
 			if (var == null)
 				throw new XPathException ("variable "+_name.ToString ()+" not found");
