@@ -118,6 +118,46 @@ namespace System.Xml {
 		  "---ddzzz",
 		  "---ddZ",
 		};
+
+#if NET_2_0
+		static readonly string [] defaultDateTimeFormats = new string [] {
+			"yyyy-MM-ddTHH:mm:ss", // dateTime(1)
+			"yyyy-MM-ddTHH:mm:ss.FFFFFFF", // dateTime(2)
+			"yyyy-MM-dd", // date
+			"HH:mm:ss", // time
+			"yyyy-MM", // gYearMonth
+			"yyyy", // gYear
+			"--MM-dd", // gMonthDay
+			"---dd", // gDay
+			};
+
+		static readonly string [] roundtripDateTimeFormats;
+		static readonly string [] localDateTimeFormats;
+		static readonly string [] utcDateTimeFormats;
+		static readonly string [] unspecifiedDateTimeFormats;
+
+		static XmlConvert ()
+		{
+			int l = defaultDateTimeFormats.Length;
+			roundtripDateTimeFormats = new string [l];
+			localDateTimeFormats = new string [l];
+			utcDateTimeFormats = new string [l * 3];
+			unspecifiedDateTimeFormats = new string [l * 4];
+			for (int i = 0; i < l; i++) {
+				string s = defaultDateTimeFormats [i];
+				localDateTimeFormats [i] = s + "zzz";
+				roundtripDateTimeFormats [i] = s + 'K';
+				utcDateTimeFormats [i * 3] = s;
+				utcDateTimeFormats [i * 3 + 1] = s + 'Z';
+				utcDateTimeFormats [i * 3 + 2] = s + "zzz";
+				unspecifiedDateTimeFormats [i * 4] = s;
+				unspecifiedDateTimeFormats [i * 4 + 1] = localDateTimeFormats [i];
+				unspecifiedDateTimeFormats [i * 4 + 2] = roundtripDateTimeFormats [i];
+				unspecifiedDateTimeFormats [i * 4 + 3] = utcDateTimeFormats [i];
+			}
+		}
+#endif
+		static DateTimeStyles _defaultStyle = DateTimeStyles.AllowLeadingWhite | DateTimeStyles.AllowTrailingWhite;
 		
 		public XmlConvert()
 		{}
@@ -239,10 +279,20 @@ namespace System.Xml {
 		{
 			if (buffer == null)
 				throw new ArgumentNullException ("buffer");
-			if (index < 0)
-				throw new ArgumentOutOfRangeException ("index", index, "index must be non negative integer.");
-			if (count < 0)
-				throw new ArgumentOutOfRangeException ("count", count, "count must be non negative integer.");
+			if (index < 0) {
+				throw new ArgumentOutOfRangeException (
+#if !NET_2_1
+					"index", index,
+#endif
+					"index must be non negative integer.");
+			}
+			if (count < 0) {
+				throw new ArgumentOutOfRangeException (
+#if !NET_2_1
+					"count", count,
+#endif
+					"count must be non negative integer.");
+			}
 			if (buffer.Length < index + count)
 				throw new ArgumentOutOfRangeException ("index and count must be smaller than the length of the buffer.");
 
@@ -285,29 +335,40 @@ namespace System.Xml {
 		public static DateTime ToDateTime (string value, XmlDateTimeSerializationMode mode)
 		{
 			string modestr = null;
+			DateTime dt;
 			switch (mode) {
 			case XmlDateTimeSerializationMode.Local:
+				dt = ToDateTime (value, localDateTimeFormats);
+				return dt == DateTime.MinValue || dt == DateTime.MaxValue ? dt : dt.ToLocalTime ();
 			case XmlDateTimeSerializationMode.RoundtripKind:
-			default:
-				return ToDateTime (value, "yyyy-MM-ddTHH:mm:ss.fffffffzzz");
+				return ToDateTime (value, roundtripDateTimeFormats, _defaultStyle | DateTimeStyles.RoundtripKind);
 			case XmlDateTimeSerializationMode.Utc:
-				return ToDateTime (value, "yyyy-MM-ddTHH:mm:ss.fffffffZ").ToUniversalTime ();
+				dt = ToDateTime (value, utcDateTimeFormats);
+				return dt == DateTime.MinValue || dt == DateTime.MaxValue ? dt : dt.ToUniversalTime ();
 			case XmlDateTimeSerializationMode.Unspecified:
-				return ToDateTime (value, "yyyy-MM-ddTHH:mm:ss.fffffff");
+				return ToDateTime (value, unspecifiedDateTimeFormats);
+			default:
+				return ToDateTime (value, defaultDateTimeFormats);
 			}
 		}
 #endif
 		public static DateTime ToDateTime(string s, string format)
 		{
-			DateTimeFormatInfo d = new DateTimeFormatInfo();
-			d.FullDateTimePattern = format;
-			return DateTime.Parse(s, d);
+			//DateTimeFormatInfo d = new DateTimeFormatInfo();
+			//d.FullDateTimePattern = format;
+			//return DateTime.Parse(s, d);
+			DateTimeStyles style = DateTimeStyles.AllowLeadingWhite |
+					       DateTimeStyles.AllowTrailingWhite;			
+			return DateTime.ParseExact (s, format, DateTimeFormatInfo.InvariantInfo, style);
 		}
 
 		public static DateTime ToDateTime(string s, string[] formats)
 		{
-			DateTimeStyles style = DateTimeStyles.AllowLeadingWhite |
-					       DateTimeStyles.AllowTrailingWhite;
+			return ToDateTime (s, formats, _defaultStyle);			
+		}
+
+		private static DateTime ToDateTime (string s, string [] formats, DateTimeStyles style) 
+		{
 			return DateTime.ParseExact (s, formats, DateTimeFormatInfo.InvariantInfo, style);
 		}
 		
@@ -336,17 +397,17 @@ namespace System.Xml {
 
 		public static short ToInt16(string s)
 		{
-			return Int16.Parse(s, CultureInfo.InvariantCulture);
+			return Int16.Parse (s, NumberStyles.Integer, CultureInfo.InvariantCulture);
 		}
 
 		public static int ToInt32(string s)
 		{
-			return Int32.Parse(s, CultureInfo.InvariantCulture);
+			return Int32.Parse (s, NumberStyles.Integer, CultureInfo.InvariantCulture);
 		}
 
 		public static long ToInt64(string s)
 		{
-			return Int64.Parse(s, CultureInfo.InvariantCulture);
+			return Int64.Parse (s, NumberStyles.Integer, CultureInfo.InvariantCulture);
 		}
 
 		[CLSCompliant (false)]
@@ -459,7 +520,7 @@ namespace System.Xml {
 			if (Double.IsNegativeInfinity(value)) return "-INF";
 			if (Double.IsPositiveInfinity(value)) return "INF";
 			if (Double.IsNaN(value)) return "NaN";
-			return value.ToString(CultureInfo.InvariantCulture);
+			return value.ToString("R", CultureInfo.InvariantCulture);
 		}
 
 		public static string ToString(float value)
@@ -467,7 +528,7 @@ namespace System.Xml {
 			if (Single.IsNegativeInfinity(value)) return "-INF";
 			if (Single.IsPositiveInfinity(value)) return "INF";
 			if (Single.IsNaN(value)) return "NaN";
-			return value.ToString(CultureInfo.InvariantCulture);
+			return value.ToString("R", CultureInfo.InvariantCulture);
 		}
 
 		[CLSCompliant (false)]
@@ -493,23 +554,33 @@ namespace System.Xml {
 #if NET_2_0
 		public static string ToString (DateTime value, XmlDateTimeSerializationMode mode)
 		{
+			// Unlike usual DateTime formatting, it preserves
+			// MaxValue/MinValue as is.
 			string modestr = null;
 			switch (mode) {
 			case XmlDateTimeSerializationMode.Local:
+				return (value == DateTime.MinValue ? DateTime.MinValue : value == DateTime.MaxValue ? value : value.ToLocalTime ()).ToString (
+					"yyyy-MM-ddTHH:mm:ss.FFFFFFFzzz",
+					CultureInfo.InvariantCulture);
+				break;
 			case XmlDateTimeSerializationMode.RoundtripKind:
+				return value.ToString (
+					"yyyy-MM-ddTHH:mm:ss.FFFFFFFK",
+					CultureInfo.InvariantCulture);
+				break;
 			default:
 				return value.ToString (
-					"yyyy-MM-ddTHH:mm:ss.fffffffzzz",
+					"yyyy-MM-ddTHH:mm:ss.FFFFFFFzzz",
 					CultureInfo.InvariantCulture);
 				break;
 			case XmlDateTimeSerializationMode.Utc:
-				return value.ToUniversalTime ().ToString (
-					"yyyy-MM-ddTHH:mm:ss.fffffffZ",
+				return (value == DateTime.MinValue ? DateTime.MinValue : value == DateTime.MaxValue ? value : value.ToUniversalTime ()).ToString (
+					"yyyy-MM-ddTHH:mm:ss.FFFFFFFZ",
 					CultureInfo.InvariantCulture);
 				break;
 			case XmlDateTimeSerializationMode.Unspecified:
 				return value.ToString (
-					"yyyy-MM-ddTHH:mm:ss.fffffff",
+					"yyyy-MM-ddTHH:mm:ss.FFFFFFF",
 					CultureInfo.InvariantCulture);
 				break;
 			}
@@ -743,5 +814,35 @@ namespace System.Xml {
 
 			return bufIndex - offset;
 		}
+
+#if NET_2_0 // actually NET_3_5
+#if !TARGET_JVM
+
+		public static DateTimeOffset ToDateTimeOffset (string s)
+		{
+			return ToDateTimeOffset (s, datetimeFormats);
+		}
+
+		public static DateTimeOffset ToDateTimeOffset (string s, string format)
+		{
+			return ToDateTimeOffset (s, format);
+		}
+
+		public static DateTimeOffset ToDateTimeOffset (string s, string [] formats)
+		{
+			return ToDateTimeOffset (s, formats);
+		}
+
+		public static string ToString (DateTimeOffset value)
+		{
+			return ToString (value, "yyyy-MM-ddTHH:mm:ss.fffffffzzz");
+		}
+
+		public static string ToString (DateTimeOffset value, string format)
+		{
+			return value.ToString (format, CultureInfo.InvariantCulture);
+		}
+#endif
+#endif
 	}
 }
