@@ -325,7 +325,7 @@ namespace System.Security.Cryptography.X509Certificates {
 
 				// again multiple candidates, let's find the AKI that match the SKI (if we have one)
 				string aki = GetAuthorityKeyIdentifier (child);
-				if ((aki == null) || (aki.Length == 0)) {
+				if (String.IsNullOrEmpty (aki)) {
 					return time_valid [0]; // FIXME: out of luck, you get the first one
 				}
 				foreach (X509Certificate2 parent in time_valid) {
@@ -363,10 +363,10 @@ namespace System.Security.Cryptography.X509Certificates {
 			// check that Authority Key Identifier == Subject Key Identifier
 			// e.g. it will be different if a self-signed certificate is part (not the end) of the chain
 			string ski = GetSubjectKeyIdentifier (certificate);
-			if ((ski == null) || (ski.Length == 0))
+			if (String.IsNullOrEmpty (ski))
 				return true;
 			string aki = GetAuthorityKeyIdentifier (certificate);
-			if ((aki == null) || (aki.Length == 0))
+			if (String.IsNullOrEmpty (aki))
 				return true;
 			// if both id are available then they must match
 			return (aki == ski);
@@ -495,15 +495,15 @@ namespace System.Security.Cryptography.X509Certificates {
 			}
 
 			// TODO 6.1.3.d - check if certificate policies extension is present
-			if (false) {
+			//if (false) {
 				// TODO - for X509ChainStatusFlags.InvalidPolicyConstraints
 				//	using X509ChainPolicy.ApplicationPolicy and X509ChainPolicy.CertificatePolicy
 
 				// TODO - check for X509ChainStatusFlags.NoIssuanceChainPolicy
 
-			} else {
+			//} else {
 				// TODO 6.1.3.e - set valid_policy_tree to NULL
-			}
+			//}
 
 			// TODO 6.1.3.f - verify explict_policy > 0 if valid_policy_tree != NULL
 		}
@@ -801,12 +801,6 @@ namespace System.Security.Cryptography.X509Certificates {
 			return X509ChainStatusFlags.NoError;
 		}
 
-		private MX.X509Crl OnlineCheck (X509Certificate2 certificate)
-		{
-			// CRL doesn't exists or couldn't be downloaded
-			return null;
-		}
-
 		private MX.X509Crl FindCrl (X509Certificate2 caCertificate)
 		{
 			string subject = caCertificate.SubjectName.Decode (X500DistinguishedNameFlags.None);
