@@ -33,6 +33,9 @@ namespace MonoTests.System.Net
 	public class HttpWebRequestTest
 	{
 		[Test]
+#if TARGET_JVM
+		[Ignore ("Ignore failures in Sys.Net")]
+#endif
 		public void Proxy_Null ()
 		{
 			HttpWebRequest req = (HttpWebRequest) WebRequest.Create ("http://www.google.com");
@@ -55,7 +58,7 @@ namespace MonoTests.System.Net
 		}
 
 		[Test]
-		[Category("InetAccess")] 
+		[Category("InetAccess")]
 #if TARGET_JVM
 		[Ignore ("NMA - wrong cookies number returned")]
 #endif
@@ -65,21 +68,20 @@ namespace MonoTests.System.Net
 			Assertion.AssertNotNull ("req:If Modified Since: ", req.IfModifiedSince);
 
 			req.UserAgent = "MonoClient v1.0";
-			Assertion.AssertEquals ("req Header 1", "User-Agent", req.Headers.GetKey (0));
-			Assertion.AssertEquals ("req Header 2", "MonoClient v1.0", req.Headers.Get (0));
+			Assert.AreEqual ("User-Agent", req.Headers.GetKey (0), "#A1");
+			Assert.AreEqual ("MonoClient v1.0", req.Headers.Get (0), "#A2");
 
 			HttpWebResponse res = (HttpWebResponse) req.GetResponse ();
-			Assertion.AssertEquals ("res:HttpStatusCode: ", "OK", res.StatusCode.ToString ());
-			Assertion.AssertEquals ("res:HttpStatusDescription: ", "OK", res.StatusDescription);
-			
-			Assertion.AssertEquals ("res Header 1", "text/html", res.Headers.Get ("Content-Type"));
-			Assertion.AssertNotNull ("Last Modified: ", res.LastModified);
-			
-			Assertion.AssertEquals ("res:", 0, res.Cookies.Count);
-				
+			Assert.AreEqual ("OK", res.StatusCode.ToString (), "#B1");
+			Assert.AreEqual ("OK", res.StatusDescription, "#B2");
+
+			Assert.AreEqual ("text/html; charset=ISO-8859-1", res.Headers.Get ("Content-Type"), "#C1");
+			Assert.IsNotNull (res.LastModified, "#C2");
+			Assert.AreEqual (0, res.Cookies.Count, "#C3");
+
 			res.Close ();
 		}
-		
+
 		[Test]
 		public void AddRange ()
 		{
@@ -119,6 +121,7 @@ namespace MonoTests.System.Net
 				Assertion.AssertEquals ("#04", true, object.ReferenceEquals (one, two));
 			}
 		}
+
 #if !TARGET_JVM //NotWorking
 		[Test]
 		public void SslClientBlock ()
@@ -275,7 +278,313 @@ namespace MonoTests.System.Net
 			}
 		}
 
-		private static byte [] EchoRequestHandler (Socket socket)
+		[Test]
+#if TARGET_JVM
+		[Category("NotWorking")]
+#endif
+		public void ReadTimeout ()
+		{
+			IPEndPoint localEP = new IPEndPoint (IPAddress.Loopback, 8764);
+			string url = "http://" + localEP.ToString () + "/original/";
+
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (RedirectRequestHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.AllowAutoRedirect = false;
+				req.Timeout = 200;
+				req.ReadWriteTimeout = 100;
+				req.KeepAlive = false;
+				Stream rs = req.GetRequestStream ();
+				rs.Close ();
+				using (HttpWebResponse resp = (HttpWebResponse) req.GetResponse ()) {
+					try {
+						Stream s = resp.GetResponseStream ();
+						s.ReadByte ();
+						Assert.Fail ("#1");
+					} catch (WebException ex) {
+						Assert.AreEqual (typeof (WebException), ex.GetType (), "#2");
+						Assert.IsNull (ex.InnerException, "#3");
+						Assert.IsNull (ex.Response, "#4");
+						Assert.AreEqual (WebExceptionStatus.Timeout, ex.Status, "#5");
+					}
+				}
+				responder.Stop ();
+			}
+		}
+
+		[Test] // bug #81624
+#if TARGET_JVM
+		[Category("NotWorking")]
+#endif
+		public void AllowAutoRedirect ()
+		{
+			IPEndPoint localEP = new IPEndPoint (IPAddress.Loopback, 8764);
+			string url = "http://" + localEP.ToString () + "/original/";
+
+			// allow autoredirect
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (RedirectRequestHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+				req.KeepAlive = false;
+				Stream rs = req.GetRequestStream ();
+				rs.Close ();
+				using (HttpWebResponse resp = (HttpWebResponse) req.GetResponse ()) {
+					StreamReader sr = new StreamReader (resp.GetResponseStream (),
+						Encoding.UTF8);
+					string body = sr.ReadToEnd ();
+
+					Assert.AreEqual (resp.StatusCode, HttpStatusCode.OK, "#A1");
+					Assert.AreEqual (resp.ResponseUri.ToString (), "http://" +
+						localEP.ToString () + "/moved/", "#A2");
+					Assert.AreEqual ("GET", resp.Method, "#A3");
+					Assert.AreEqual ("LOOKS OK", body, "#A4");
+				}
+				responder.Stop ();
+			}
+
+			// do not allow autoredirect
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (RedirectRequestHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.AllowAutoRedirect = false;
+				req.Timeout = 1000;
+				req.ReadWriteTimeout = 1000;
+				req.KeepAlive = false;
+				Stream rs = req.GetRequestStream ();
+				rs.Close ();
+				using (HttpWebResponse resp = (HttpWebResponse) req.GetResponse ()) {
+					Assert.AreEqual (resp.StatusCode, HttpStatusCode.Found, "#B1");
+					Assert.AreEqual (url, resp.ResponseUri.ToString (), "#B2");
+					Assert.AreEqual ("POST", resp.Method, "#B3");
+				}
+				responder.Stop ();
+			}
+		}
+
+		[Test] // bug #81671
+		[Category ("NotWorking")]
+		public void InternalServerError ()
+		{
+			IPEndPoint localEP = new IPEndPoint (IPAddress.Loopback, 8764);
+			string url = "http://" + localEP.ToString () + "/original/";
+
+			// POST
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (InternalErrorHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+				req.KeepAlive = false;
+				Stream rs = req.GetRequestStream ();
+				rs.Close ();
+
+				try {
+					req.GetResponse ();
+					Assert.Fail ("#A1");
+				} catch (WebException ex) {
+					Assert.AreEqual (typeof (WebException), ex.GetType (), "#A2");
+					Assert.IsNull (ex.InnerException, "#A3");
+					Assert.IsNotNull (ex.Message, "#A4");
+					Assert.AreEqual (WebExceptionStatus.ProtocolError, ex.Status, "#A5");
+
+					HttpWebResponse webResponse = ex.Response as HttpWebResponse;
+					Assert.IsNotNull (webResponse, "#A6");
+					Assert.AreEqual ("POST", webResponse.Method, "#A7");
+					webResponse.Close ();
+				}
+
+				responder.Stop ();
+			}
+
+			// GET
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (InternalErrorHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "GET";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+				req.KeepAlive = false;
+
+				try {
+					req.GetResponse ();
+					Assert.Fail ("#B1");
+				} catch (WebException ex) {
+					Assert.AreEqual (typeof (WebException), ex.GetType (), "#B2");
+					Assert.IsNull (ex.InnerException, "#B3");
+					Assert.AreEqual (WebExceptionStatus.ProtocolError, ex.Status, "#B4");
+
+					HttpWebResponse webResponse = ex.Response as HttpWebResponse;
+					Assert.IsNotNull (webResponse, "#B5");
+					Assert.AreEqual ("GET", webResponse.Method, "#B6");
+					webResponse.Close ();
+				}
+
+				responder.Stop ();
+			}
+		}
+
+		[Test]
+		[Category ("NotWorking")] // we report a timeout
+		public void NoContentLength ()
+		{
+			IPEndPoint localEP = new IPEndPoint (IPAddress.Loopback, 8764);
+			string url = "http://" + localEP.ToString () + "/original/";
+
+			// POST
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (NoContentLengthHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+				req.KeepAlive = false;
+				Stream rs = req.GetRequestStream ();
+				rs.Close ();
+
+				try {
+					req.GetResponse ();
+					Assert.Fail ("#A1");
+				} catch (WebException ex) {
+					Assert.AreEqual (typeof (WebException), ex.GetType (), "#A2");
+#if NET_2_0
+					//Assert.IsNotNull (ex.InnerException, "#A3");
+					Assert.AreEqual (WebExceptionStatus.ReceiveFailure, ex.Status, "#A4");
+					Assert.AreEqual (typeof (IOException), ex.InnerException.GetType (), "#A5");
+					
+					// Unable to read data from the transport connection:
+					// A connection attempt failed because the connected party
+					// did not properly respond after a period of time, or
+					// established connection failed because connected host has
+					// failed to respond
+					IOException ioe = (IOException) ex.InnerException;
+					Assert.IsNotNull (ioe.InnerException, "#A6");
+					Assert.IsNotNull (ioe.Message, "#A7");
+					Assert.AreEqual (typeof (SocketException), ioe.InnerException.GetType (), "#A8");
+
+					// A connection attempt failed because the connected party
+					// did not properly respond after a period of time, or
+					// established connection failed because connected host has
+					// failed to respond
+					SocketException soe = (SocketException) ioe.InnerException;
+					Assert.IsNull (soe.InnerException, "#A9");
+					Assert.IsNotNull (soe.Message, "#A10");
+
+					HttpWebResponse webResponse = ex.Response as HttpWebResponse;
+					Assert.IsNull (webResponse, "#A11");
+#else
+					Assert.IsNull (ex.InnerException, "#A3");
+					Assert.AreEqual (WebExceptionStatus.ProtocolError, ex.Status, "#A4");
+
+					HttpWebResponse webResponse = ex.Response as HttpWebResponse;
+					Assert.IsNotNull (webResponse, "#A5");
+					Assert.AreEqual ("POST", webResponse.Method, "#A6");
+					webResponse.Close ();
+#endif
+				}
+
+				responder.Stop ();
+			}
+
+			// GET
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (NoContentLengthHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "GET";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+				req.KeepAlive = false;
+
+				try {
+					req.GetResponse ();
+					Assert.Fail ("#B1");
+				} catch (WebException ex) {
+					Assert.AreEqual (typeof (WebException), ex.GetType (), "#B2");
+					Assert.IsNull (ex.InnerException, "#B3");
+					Assert.AreEqual (WebExceptionStatus.ProtocolError, ex.Status, "#B4");
+
+					HttpWebResponse webResponse = ex.Response as HttpWebResponse;
+					Assert.IsNotNull (webResponse, "#B5");
+					Assert.AreEqual ("GET", webResponse.Method, "#B6");
+					webResponse.Close ();
+				}
+
+				responder.Stop ();
+			}
+		}
+
+#if NET_2_0
+		[Test] // bug #81504
+#if TARGET_JVM
+		[Category ("NotWorking")]
+#endif
+		public void Stream_CanTimeout ()
+		{
+			IPEndPoint localEP = new IPEndPoint (IPAddress.Loopback, 8764);
+			string url = "http://" + localEP.ToString () + "/original/";
+
+			// allow autoredirect
+			using (SocketResponder responder = new SocketResponder (localEP, new SocketRequestHandler (RedirectRequestHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+				req.KeepAlive = false;
+				Stream rs = req.GetRequestStream ();
+				Assert.IsTrue (rs.CanTimeout, "#1");
+				rs.Close ();
+				using (HttpWebResponse resp = (HttpWebResponse) req.GetResponse ()) {
+					Stream os = resp.GetResponseStream ();
+					Assert.IsTrue (os.CanTimeout, "#2");
+					os.Close ();
+				}
+				responder.Stop ();
+			}
+		}
+#endif
+
+		[Test]
+		[Category ("NotWorking")] // #353495
+		public void LastModifiedKind () // bug #353495
+		{
+			const string reqURL = "http://coffeefaq.com/site/node/25";
+			const string dtFmt = "ddd, dd MMM yyyy HH:mm:ss";
+			HttpWebRequest req = (HttpWebRequest) WebRequest.Create (reqURL);
+			HttpWebResponse resp = (HttpWebResponse) req.GetResponse ();
+			DateTime lastMod = resp.LastModified;
+			string rawLastMod = resp.Headers ["Last-Modified"];
+			resp.Close ();
+			//Assert.AreEqual ("Tue, 15 Jan 2008 08:59:59 GMT", rawLastMod, "#1");
+#if NET_2_0
+			Assert.AreEqual (DateTimeKind.Local, lastMod.Kind, "#2");
+#endif
+			req = (HttpWebRequest) WebRequest.Create (reqURL);
+			req.IfModifiedSince = lastMod;
+			try {
+				resp = (HttpWebResponse) req.GetResponse ();
+				resp.Close ();
+				Assert.Fail ("Should result in 304");
+			} catch (WebException ex) {
+				Assert.AreEqual (WebExceptionStatus.ProtocolError, ex.Status, "#3");
+				Assert.AreEqual (((HttpWebResponse) ex.Response).StatusCode, HttpStatusCode.NotModified, "#4");
+			}
+		}
+
+		static byte [] EchoRequestHandler (Socket socket)
 		{
 			MemoryStream ms = new MemoryStream ();
 			byte [] buffer = new byte [4096];
@@ -293,7 +602,6 @@ namespace MonoTests.System.Net
 			StreamReader sr = new StreamReader (ms, Encoding.UTF8);
 			string request = sr.ReadToEnd ();
 
-			MemoryStream outputStream = new MemoryStream ();
 			StringWriter sw = new StringWriter ();
 			sw.WriteLine ("HTTP/1.1 200 OK");
 			sw.WriteLine ("Content-Type: text/xml");
@@ -305,6 +613,69 @@ namespace MonoTests.System.Net
 			return Encoding.UTF8.GetBytes (sw.ToString ());
 		}
 
+		static byte [] RedirectRequestHandler (Socket socket)
+		{
+			MemoryStream ms = new MemoryStream ();
+			byte [] buffer = new byte [4096];
+			int bytesReceived = socket.Receive (buffer);
+			while (bytesReceived > 0) {
+				ms.Write (buffer, 0, bytesReceived);
+				if (socket.Available > 0) {
+					bytesReceived = socket.Receive (buffer);
+				} else {
+					bytesReceived = 0;
+				}
+			}
+			ms.Flush ();
+			ms.Position = 0;
+			string statusLine = null;
+			using (StreamReader sr = new StreamReader (ms, Encoding.UTF8)) {
+				statusLine = sr.ReadLine ();
+			}
+
+			StringWriter sw = new StringWriter ();
+			if (statusLine.StartsWith ("POST /original/")) {
+				sw.WriteLine ("HTTP/1.1 302 Found");
+				sw.WriteLine ("Location: " + "http://" + IPAddress.Loopback.ToString () + ":8764/moved/");
+				sw.WriteLine ();
+				sw.Flush ();
+			} else if (statusLine.StartsWith ("GET /moved/")) {
+				sw.WriteLine ("HTTP/1.1 200 OK");
+				sw.WriteLine ("Content-Type: text/plain");
+				sw.WriteLine ("Content-Length: 8");
+				sw.WriteLine ();
+				sw.Write ("LOOKS OK");
+				sw.Flush ();
+			} else {
+				sw.WriteLine ("HTTP/1.1 500 Too Lazy");
+				sw.WriteLine ();
+				sw.Flush ();
+			}
+
+			return Encoding.UTF8.GetBytes (sw.ToString ());
+		}
+
+		static byte [] InternalErrorHandler (Socket socket)
+		{
+			StringWriter sw = new StringWriter ();
+			sw.WriteLine ("HTTP/1.1 500 Too Lazy");
+			sw.WriteLine ("Content-Length: 0");
+			sw.WriteLine ();
+			sw.Flush ();
+
+			return Encoding.UTF8.GetBytes (sw.ToString ());
+		}
+
+		static byte [] NoContentLengthHandler (Socket socket)
+		{
+			StringWriter sw = new StringWriter ();
+			sw.WriteLine ("HTTP/1.1 500 Too Lazy");
+			sw.WriteLine ();
+			sw.Flush ();
+
+			return Encoding.UTF8.GetBytes (sw.ToString ());
+		}
+
 		class BadChunkedServer : HttpServer {
 			protected override void Run ()
 			{
@@ -509,4 +880,3 @@ namespace MonoTests.System.Net
 #endif
 	}
 }
-
