@@ -32,7 +32,7 @@ namespace MonoTests.System
 		public void Constructors ()
 		{
 			Uri uri = null;
-			
+
 			/*
 			uri = new Uri ("http://www.ximian.com/foo" + ((char) 0xa9) + "/bar/index.cgi?a=1&b=" + ((char) 0xa9) + "left#fragm?ent2");
 			Print (uri);
@@ -66,7 +66,10 @@ namespace MonoTests.System
 			uri = new Uri ("http://www.contoso.com:1234");
 			Print (uri);
 			*/
-			
+
+			uri = new Uri("  \r  \n http://test.com\r\n \r\r  ");
+			AssertEquals ("#k0", "http://test.com/", uri.ToString());
+
 			uri = new Uri ("http://contoso.com?subject=uri");
 			AssertEquals ("#k1", "/", uri.AbsolutePath);
 			AssertEquals ("#k2", "http://contoso.com/?subject=uri", uri.AbsoluteUri);
@@ -104,6 +107,9 @@ namespace MonoTests.System
 			AssertEquals ("#m15", "mailto", uri.Scheme);
 			AssertEquals ("#m16", false, uri.UserEscaped);
 			AssertEquals ("#m17", "user:pwd", uri.UserInfo);
+
+			uri = new Uri("myscheme://127.0.0.1:5");
+			AssertEquals("#c1", "myscheme://127.0.0.1:5/", uri.ToString());
 			
 			uri = new Uri (@"\\myserver\mydir\mysubdir\myfile.ext");
 			AssertEquals ("#n1", "/mydir/mysubdir/myfile.ext", uri.AbsolutePath);
@@ -532,7 +538,7 @@ namespace MonoTests.System
 		public void InvalidFile3 ()
 		{
 			new Uri ("file:/foo");
-		}	
+		}
 
 		[Test]
 		[ExpectedException (typeof (UriFormatException))]
@@ -640,10 +646,12 @@ namespace MonoTests.System
 			AssertEquals ("#7e", "/", uri.AbsolutePath);
 			AssertEquals ("#7f", "/", uri.PathAndQuery);
 			AssertEquals ("#7g", "file://one_file.txt/", uri.GetLeftPart (UriPartial.Path));
+#if !TARGET_JVM
 			if (isWin32)
 				AssertEquals ("#7b", "\\\\one_file.txt\\", uri.LocalPath);
 			else
 				AssertEquals ("#7b", "/", uri.LocalPath);
+#endif
 			AssertEquals ("#7c", "file", uri.Scheme);
 			AssertEquals ("#7d", "one_file.txt", uri.Host);
 		}
@@ -694,6 +702,17 @@ namespace MonoTests.System
 				AssertEquals ("#9d", "file:///foo%25bar", uri.ToString ());
 			}
 		}
+
+		// Novell Bugzilla #320614
+		[Test]
+		public void QueryEscape ()
+		{
+			Uri u1 = new Uri("http://localhost:8080/test.aspx?ReturnUrl=%2fSearchDoc%2fSearcher.aspx");
+			Uri u2 = new Uri("http://localhost:8080/test.aspx?ReturnUrl=%252fSearchDoc%252fSearcher.aspx");
+			
+			AssertEquals ("QE1", u1.ToString (), "http://localhost:8080/test.aspx?ReturnUrl=/SearchDoc/Searcher.aspx");
+			AssertEquals ("QE2", u2.ToString (), "http://localhost:8080/test.aspx?ReturnUrl=%2fSearchDoc%2fSearcher.aspx");
+		}
 		
 		[Test]
 		public void UnixPath () {
@@ -724,8 +743,10 @@ namespace MonoTests.System
 		[Category("NotDotNet")]
 		public void UncFail ()
 		{
-			Uri uri = new Uri ("/home/user/dir/filename.ext");
-			Assert ("#7", !uri.IsUnc);
+			if (!isWin32) {
+				Uri uri = new Uri ("/home/user/dir/filename.ext");
+				Assert ("#7", !uri.IsUnc);
+			}
 		}
 
 		[Test]
@@ -791,6 +812,15 @@ namespace MonoTests.System
 			} catch (ArgumentOutOfRangeException) {}
 		}
 
+        [Test]
+        public void MoreHexEscape()
+        {
+            string url = "http://guyc-2003-sp/wiki/wiki%20document%20library/בד%20יקה.docx";
+            string escapedAbsolutePath = "/wiki/wiki%20document%20library/%D7%91%D7%93%20%D7%99%D7%A7%D7%94.docx";
+            Uri u = new Uri(url);
+            AssertEquals("Escaped non-english combo", escapedAbsolutePath, u.AbsolutePath);
+        }
+
 		[Test]
 		public void HexUnescape () 
 		{
@@ -922,7 +952,7 @@ namespace MonoTests.System
 		}
 
 		[Test]
-		public void FragmentEscape ()
+		public void Fragment_Escape ()
 		{
 			Uri u = new Uri("http://localhost/index.asp#main#start", false);
 			AssertEquals ("#1", u.Fragment, "#main%23start");
@@ -939,7 +969,27 @@ namespace MonoTests.System
 			n = new Uri (b, "blah#main#start", true);
 			AssertEquals ("#4", n.Fragment, "#main#start");
 		}
-			
+
+#if NET_2_0
+		[Test]
+		public void Fragment_RelativeUri ()
+		{
+			Uri uri1 = new Uri ("http://www.contoso.com/index.htm?x=2");
+			Uri uri2 = new Uri ("http://www.contoso.com/foo/bar/index.htm#fragment");
+			Uri relativeUri = uri1.MakeRelativeUri (uri2);
+
+			try {
+				string fragment = relativeUri.Fragment;
+				Fail ("#1: " + fragment);
+			} catch (InvalidOperationException ex) {
+				// This operation is not supported for a relative URI
+				AssertEquals ("#2", typeof (InvalidOperationException), ex.GetType ());
+				AssertNull ("#3", ex.InnerException);
+				AssertNotNull ("#4", ex.Message);
+			}
+		}
+#endif
+
 		[Test]
 		[ExpectedException(typeof(UriFormatException))]
 		public void IncompleteSchemeDelimiter ()
@@ -1149,9 +1199,9 @@ namespace MonoTests.System
 			uri2 = new Uri ("http://www.contoso.com/index.htm?x=1");
 			Assert ("#2", uri1.GetHashCode () != uri2.GetHashCode ());
 			uri2 = new Uri ("http://www.contoso.com:80/index.htm");
-			AssertEquals ("#3", uri1.GetHashCode (), uri2.GetHashCode ());			
+			AssertEquals ("#3", uri1.GetHashCode (), uri2.GetHashCode ());
 			uri2 = new Uri ("http://www.contoso.com:8080/index.htm");
-			Assert ("#4", uri1.GetHashCode () != uri2.GetHashCode ());			
+			Assert ("#4", uri1.GetHashCode () != uri2.GetHashCode ());
 		}
 
 #if NET_2_0
@@ -1174,6 +1224,40 @@ namespace MonoTests.System
 			Assert ("#5b", !uri3.Equals(uri4));
 		}
 
+		[Test]
+		[ExpectedException(typeof(InvalidOperationException))]
+		public void GetLeftPart_Partial1 ()
+		{
+			Uri u = new Uri ("foo", UriKind.Relative);
+			u.GetLeftPart (UriPartial.Scheme);
+		}
+
+		[ExpectedException(typeof(InvalidOperationException))]
+		[Test]
+		public void GetLeftPart_Partial2 ()
+		{
+			Uri u = new Uri ("foo", UriKind.Relative);
+			u.GetLeftPart (UriPartial.Authority);
+		}
+
+		[ExpectedException(typeof(InvalidOperationException))]
+		[Test]
+		public void GetLeftPart_Partial3 ()
+		{
+			Uri u = new Uri ("foo", UriKind.Relative);
+			u.GetLeftPart (UriPartial.Path);
+		}
+		
+		[Test]
+		public void TestPartialToString ()
+		{
+			AssertEquals ("#1", new Uri ("foo", UriKind.Relative).ToString (), "foo");
+			AssertEquals ("#2", new Uri ("foo#aa", UriKind.Relative).ToString (), "foo#aa");
+			AssertEquals ("#3", new Uri ("foo?aa", UriKind.Relative).ToString (), "foo?aa");
+			AssertEquals ("#4", new Uri ("foo#dingus?aa", UriKind.Relative).ToString (), "foo#dingus?aa");
+			AssertEquals ("#4", new Uri ("foo?dingus#aa", UriKind.Relative).ToString (), "foo?dingus#aa");
+		}
+		
 		[Test]
 		public void RelativeGetHashCodeTest()
 		{
@@ -1201,7 +1285,7 @@ namespace MonoTests.System
 			AssertEquals ("#2", "../../index.htm", uri2.MakeRelative (uri1));
 			
 			AssertEquals ("#3", "../../bar/foo/index.htm", uri2.MakeRelative (uri3));
-			AssertEquals ("#4", "../../foo/bar/index.htm", uri3.MakeRelative (uri2));			
+			AssertEquals ("#4", "../../foo/bar/index.htm", uri3.MakeRelative (uri2));
 
 			AssertEquals ("#5", "../foo2/index.htm", uri3.MakeRelative (uri4));
 			AssertEquals ("#6", "../foo/index.htm", uri4.MakeRelative (uri3));
@@ -1266,7 +1350,7 @@ namespace MonoTests.System
 
 		[Test]
 		public void ToStringTest()
-		{			
+		{
 			Uri uri = new Uri ("dummy://xxx");
 			AssertEquals ("#1", "dummy://xxx/", uri.ToString ());
 		}
@@ -1545,13 +1629,22 @@ namespace MonoTests.System
 				sb.Append ((char) i);
 			
 			AssertEquals ("%00%01%02%03%04%05%06%07%08%09%0A%0B%0C%0D%0E%0F%10%11%12%13%14%15%16%17%18%19%1A%1B%1C%1D%1E%1F%20!%22#$%25&'()*+,-./0123456789:;%3C=%3E?@ABCDEFGHIJKLMNOPQRSTUVWXYZ%5B%5C%5D%5E_%60abcdefghijklmnopqrstuvwxyz%7B%7C%7D~%7F",
-				      Uri.EscapeUriString (sb.ToString ()));
-				      
+				Uri.EscapeUriString (sb.ToString ()));
 			AssertEquals ("%C3%A1", Uri.EscapeDataString ("á"));
 		}
 #endif
 
+		//bnc #363320
+		[Test]
+		public void TestUTF8Strings ()
+		{
+			string [] tests = {
+				"file:///tmp/x (%232).jpg",
+				"file:///tmp/ü (%232).jpg" };
 
+			foreach (string test in tests)
+				AssertEquals (test, new Uri (test).ToString ());
+		}
 
 		// This test doesn't work on Linux, and arguably shouldn't work.
 		// new Uri("file:///tmp/foo/bar").AbsolutePath returns "/tmp/foo/bar" 
@@ -1600,7 +1693,7 @@ namespace MonoTests.System
 			Console.WriteLine ("IsLoopback: " + uri.IsLoopback);
 			Console.WriteLine ("IsUnc: " + uri.IsUnc);
 			Console.WriteLine ("LocalPath: " + uri.LocalPath);
-			Console.WriteLine ("PathAndQuery: " + uri.PathAndQuery);
+			Console.WriteLine ("PathAndQuery	: " + uri.PathAndQuery);
 			Console.WriteLine ("Port: " + uri.Port);
 			Console.WriteLine ("Query: " + uri.Query);
 			Console.WriteLine ("Scheme: " + uri.Scheme);
@@ -1613,9 +1706,8 @@ namespace MonoTests.System
 				Console.WriteLine ("\tNo Segments");
 			else 
 				for (int i = 0; i < segments.Length; i++) 
-					Console.WriteLine ("\t" + segments[i]);					
+					Console.WriteLine ("\t" + segments[i]);
 			Console.WriteLine ("");
 		}
 	}
 }
-
