@@ -33,7 +33,12 @@ namespace System.Collections
 	[ComVisible(true)]
 #endif
 	[Serializable]
-	public class ArrayList
+#if INSIDE_CORLIB
+	public
+#else
+	internal
+#endif
+	class ArrayList
 		: IList, ICloneable, ICollection, IEnumerable 
 	{
 		#region Enumerator
@@ -622,7 +627,8 @@ namespace System.Collections
 
 				while (x <= y) 
 				{
-					z = (x + y) / 2;
+					// Be careful with overflows
+					z = x + ((y - x) / 2);
 				
 					r = comparer.Compare(value, m_Adaptee[z]);
 				
@@ -716,7 +722,8 @@ namespace System.Collections
 
 				// Pick the pivot using the median-of-three strategy.
 
-				middle = (left + right) / 2;
+				// Be careful with overflows
+				middle = left + ((right - left) / 2);
 
 				if (comparer.Compare(list[middle], list[left]) < 0) 
 				{
@@ -2508,6 +2515,8 @@ namespace System.Collections
 		/// Total number of state changes.
 		/// </summary>
 		private int _version;
+		
+		private static readonly object [] EmptyArray = new object [0]; 
 
 		#endregion
 		
@@ -2520,7 +2529,7 @@ namespace System.Collections
 		public ArrayList()
 		{
 #if NET_2_0
-			_items = new object[0];
+			_items = EmptyArray;
 #else
 			_items = new object[DefaultInitialCapacity];
 #endif
@@ -2827,6 +2836,7 @@ namespace System.Collections
 				int x = index - count ;
 
 				Array.Copy(_items, x, _items, index, _size - x);
+				Array.Clear(_items, _size + count, - count);
 			}
 		}
 
@@ -2856,9 +2866,9 @@ namespace System.Collections
 			_version++;
 		}
 
-		public virtual bool Contains(object value) 
+		public virtual bool Contains(object item) 
 		{
-			return IndexOf(value, 0, _size) > -1;
+			return IndexOf(item, 0, _size) > -1;
 		}
 
 		internal virtual bool Contains(object value, int startIndex, int count) 
@@ -2983,11 +2993,11 @@ namespace System.Collections
 			_version++;
 		}
 
-		public virtual void Remove(object value) 
+		public virtual void Remove(object obj) 
 		{
 			int x;
 
-			x = IndexOf(value);
+			x = IndexOf(obj);
 
 			if (x > -1) 
 			{
@@ -3038,9 +3048,9 @@ namespace System.Collections
 			Array.Copy(_items, array, _size);
 		}
 
-		public virtual void CopyTo(System.Array array, int index) 
+		public virtual void CopyTo(System.Array array, int arrayIndex) 
 		{			
-			CopyTo(0, array, index, _size);
+			CopyTo(0, array, arrayIndex, _size);
 		}
 
 		public virtual void CopyTo(int index, System.Array array, int arrayIndex, int count) 
@@ -3195,11 +3205,11 @@ namespace System.Collections
 			return retval;
 		}
 
-		public virtual Array ToArray(Type elementType) 
+		public virtual Array ToArray(Type type) 
 		{
 			Array retval;
 			
-			retval = Array.CreateInstance(elementType, _size);
+			retval = Array.CreateInstance(type, _size);
 
 			CopyTo(retval);
 
@@ -3250,29 +3260,31 @@ namespace System.Collections
 				throw new ArgumentNullException("list");
 			}
 
-			if (list.IsSynchronized) 
-			{
-				return ArrayList.Synchronized(new ArrayListAdapter(list));
-			}
+			ArrayList arrayList = list as ArrayList;
+			if (arrayList != null)
+				return arrayList;
+			else
+				arrayList = new ArrayListAdapter(list);
+
+			if (list.IsSynchronized)
+				return ArrayList.Synchronized(arrayList);
 			else 
-			{
-				return new ArrayListAdapter(list);
-			}
+				return arrayList;
 		}
 
-		public static ArrayList Synchronized(ArrayList arrayList) 
+		public static ArrayList Synchronized(ArrayList list) 
 		{
-			if (arrayList == null) 
+			if (list == null) 
 			{
-				throw new ArgumentNullException("arrayList");
+				throw new ArgumentNullException("list");
 			}
 
-			if (arrayList.IsSynchronized)
+			if (list.IsSynchronized)
 			{
-				return arrayList;
+				return list;
 			}
 
-			return new SynchronizedArrayListWrapper(arrayList);
+			return new SynchronizedArrayListWrapper(list);
 		}
 
 		public static IList Synchronized(IList list) 
@@ -3290,19 +3302,19 @@ namespace System.Collections
 			return new SynchronizedListWrapper(list);
 		}
 
-		public static ArrayList ReadOnly(ArrayList arrayList) 
+		public static ArrayList ReadOnly(ArrayList list) 
 		{
-			if (arrayList == null) 
+			if (list == null) 
 			{
-				throw new ArgumentNullException("arrayList");
+				throw new ArgumentNullException("list");
 			}
 
-			if (arrayList.IsReadOnly)
+			if (list.IsReadOnly)
 			{
-				return arrayList;
+				return list;
 			}
 
-			return new ReadOnlyArrayListWrapper(arrayList);
+			return new ReadOnlyArrayListWrapper(list);
 		}
 
 		public static IList ReadOnly(IList list) 
@@ -3320,19 +3332,19 @@ namespace System.Collections
 			return new ReadOnlyListWrapper(list);
 		}
 
-		public static ArrayList FixedSize(ArrayList arrayList) 
+		public static ArrayList FixedSize(ArrayList list) 
 		{
-			if (arrayList == null) 
+			if (list == null) 
 			{
-				throw new ArgumentNullException("arrayList");
+				throw new ArgumentNullException("list");
 			}
 
-			if (arrayList.IsFixedSize)
+			if (list.IsFixedSize)
 			{
-				return arrayList;
+				return list;
 			}
 
-			return new FixedSizeArrayListWrapper(arrayList);
+			return new FixedSizeArrayListWrapper(list);
 		}
 
 		public static IList FixedSize(IList list) 
