@@ -40,17 +40,19 @@ using System.Runtime.CompilerServices;
 using System.Runtime.InteropServices;
 using System.Diagnostics.SymbolStore;
 
-namespace System.Reflection.Emit {
+namespace System.Reflection.Emit
+{
 #if NET_2_0
 	[ComVisible (true)]
 	[ComDefaultInterface (typeof (_MethodBuilder))]
 #endif
 	[ClassInterface (ClassInterfaceType.None)]
-	public sealed class MethodBuilder : MethodInfo, _MethodBuilder {
+	public sealed class MethodBuilder : MethodInfo, _MethodBuilder
+	{
 		private RuntimeMethodHandle mhandle;
 		private Type rtype;
 		private Type[] parameters;
-		private MethodAttributes attrs;
+		private MethodAttributes attrs;	/* It's used directly by MCS */
 		private MethodImplAttributes iattrs;
 		private string name;
 		private int table_idx;
@@ -69,7 +71,7 @@ namespace System.Reflection.Emit {
 		private bool init_locals = true;
 		private IntPtr generic_container;
 #if NET_2_0 || BOOTSTRAP_NET_2_0
-		private	GenericTypeParameterBuilder[] generic_params;
+		private GenericTypeParameterBuilder[] generic_params;
 #else
 		private Object generic_params; /* so offsets are the same */
 #endif
@@ -79,7 +81,8 @@ namespace System.Reflection.Emit {
 		private Type[][] paramModOpt;
 		private RefEmitPermissionSet[] permissions;
 
-		internal MethodBuilder (TypeBuilder tb, string name, MethodAttributes attributes, CallingConventions callingConvention, Type returnType, Type[] returnModReq, Type[] returnModOpt, Type[] parameterTypes, Type[][] paramModReq, Type[][] paramModOpt) {
+		internal MethodBuilder (TypeBuilder tb, string name, MethodAttributes attributes, CallingConventions callingConvention, Type returnType, Type[] returnModReq, Type[] returnModOpt, Type[] parameterTypes, Type[][] paramModReq, Type[][] paramModOpt)
+		{
 			this.name = name;
 			this.attrs = attributes;
 			this.call_conv = callingConvention;
@@ -102,19 +105,27 @@ namespace System.Reflection.Emit {
 			}
 			type = tb;
 			table_idx = get_next_table_index (this, 0x06, true);
-			//Console.WriteLine ("index for "+name+" set to "+table_idx.ToString());
+
+			((ModuleBuilder)tb.Module).RegisterToken (this, GetToken ().Token);
 		}
 
 		internal MethodBuilder (TypeBuilder tb, string name, MethodAttributes attributes, 
 								CallingConventions callingConvention, Type returnType, Type[] returnModReq, Type[] returnModOpt, Type[] parameterTypes, Type[][] paramModReq, Type[][] paramModOpt, 
 			String dllName, String entryName, CallingConvention nativeCConv, CharSet nativeCharset) 
-			: this (tb, name, attributes, callingConvention, returnType, returnModReq, returnModOpt, parameterTypes, paramModReq, paramModOpt) {
+			: this (tb, name, attributes, callingConvention, returnType, returnModReq, returnModOpt, parameterTypes, paramModReq, paramModOpt)
+		{
 			pi_dll = dllName;
 			pi_entry = entryName;
 			native_cc = nativeCConv;
 			charset = nativeCharset;
 		}
 
+#if NET_2_0
+		public override bool ContainsGenericParameters {
+			get { throw new NotSupportedException (); }
+		}
+#endif
+
 		public bool InitLocals {
 			get {return init_locals;}
 			set {init_locals = value;}
@@ -130,16 +141,31 @@ namespace System.Reflection.Emit {
 			}
 		}
 
-		public override Type ReturnType {get {return rtype;}}
-		public override Type ReflectedType {get {return type;}}
-		public override Type DeclaringType {get {return type;}}
-		public override string Name {get {return name;}}
-		public override MethodAttributes Attributes {get {return attrs;}}
-		public override ICustomAttributeProvider ReturnTypeCustomAttributes {
-			get {return null;}
+		public override Type ReturnType {
+			get { return rtype; }
 		}
 
-		public override CallingConventions CallingConvention { 
+		public override Type ReflectedType {
+			get { return type; }
+		}
+
+		public override Type DeclaringType {
+			get { return type; }
+		}
+
+		public override string Name {
+			get { return name; }
+		}
+
+		public override MethodAttributes Attributes {
+			get { return attrs; }
+		}
+
+		public override ICustomAttributeProvider ReturnTypeCustomAttributes {
+			get { return null; }
+		}
+
+		public override CallingConventions CallingConvention {
 			get { return call_conv; }
 		}
 
@@ -178,17 +204,23 @@ namespace System.Reflection.Emit {
 			}
 		}
 
-		public MethodToken GetToken() {
+		public MethodToken GetToken()
+		{
 			return new MethodToken(0x06000000 | table_idx);
 		}
 		
-		public override MethodInfo GetBaseDefinition() {
+		public override MethodInfo GetBaseDefinition()
+		{
 			return this;
 		}
-		public override MethodImplAttributes GetMethodImplementationFlags() {
+
+		public override MethodImplAttributes GetMethodImplementationFlags()
+		{
 			return iattrs;
 		}
-		public override ParameterInfo[] GetParameters() {
+
+		public override ParameterInfo[] GetParameters()
+		{
 			if (!type.is_created)
 				throw NotSupported ();
 			if (parameters == null)
@@ -209,11 +241,13 @@ namespace System.Reflection.Emit {
 			return parameters.Length;
 		}
 
-		public Module GetModule () {
+		public Module GetModule ()
+		{
 			return type.Module;
 		}
 
-		public void CreateMethodBody( byte[] il, int count) {
+		public void CreateMethodBody (byte[] il, int count)
+		{
 			if ((il != null) && ((count < 0) || (count > il.Length)))
 				throw new ArgumentOutOfRangeException ("Index was out of range.  Must be non-negative and less than the size of the collection.");
 
@@ -228,13 +262,18 @@ namespace System.Reflection.Emit {
 			}
 		}
 
-		public override Object Invoke(Object obj, BindingFlags invokeAttr, Binder binder, Object[] parameters, CultureInfo culture) {
+		public override Object Invoke(Object obj, BindingFlags invokeAttr, Binder binder, Object[] parameters, CultureInfo culture)
+		{
 			throw NotSupported ();
 		}
-		public override bool IsDefined (Type attribute_type, bool inherit) {
+
+		public override bool IsDefined (Type attributeType, bool inherit)
+		{
 			throw NotSupported ();
 		}
-		public override object[] GetCustomAttributes( bool inherit) {
+
+		public override object[] GetCustomAttributes (bool inherit)
+		{
 			/*
 			 * On MS.NET, this always returns not_supported, but we can't do this
 			 * since there would be no way to obtain custom attributes of 
@@ -246,18 +285,21 @@ namespace System.Reflection.Emit {
 				throw NotSupported ();
 		}
 
-		public override object[] GetCustomAttributes( Type attributeType, bool inherit) {
+		public override object[] GetCustomAttributes (Type attributeType, bool inherit)
+		{
 			if (type.is_created)
 				return MonoCustomAttrs.GetCustomAttributes (this, attributeType, inherit);
 			else
 				throw NotSupported ();
 		}
 
-		public ILGenerator GetILGenerator () {
+		public ILGenerator GetILGenerator ()
+		{
 			return GetILGenerator (64);
 		}
 
-		public ILGenerator GetILGenerator (int size) {
+		public ILGenerator GetILGenerator (int size)
+		{
 			if (((iattrs & MethodImplAttributes.CodeTypeMask) != 
 				 MethodImplAttributes.IL) ||
 				((iattrs & MethodImplAttributes.ManagedMask) != 
@@ -286,7 +328,14 @@ namespace System.Reflection.Emit {
 			return pb;
 		}
 
-		internal void fixup () {
+		internal void check_override ()
+		{
+			if (override_method != null && override_method.IsVirtual && !IsVirtual)
+				throw new TypeLoadException (String.Format("Method '{0}' override '{1}' but it is not virtual", name, override_method));
+		}
+
+		internal void fixup ()
+		{
 			if (((attrs & (MethodAttributes.Abstract | MethodAttributes.PinvokeImpl)) == 0) && ((iattrs & (MethodImplAttributes.Runtime | MethodImplAttributes.InternalCall)) == 0)) {
 #if NET_2_0
 				// do not allow zero length method body on MS.NET 2.0 (and higher)
@@ -313,7 +362,7 @@ namespace System.Reflection.Emit {
 			}
 		}
 
-		public void SetCustomAttribute( CustomAttributeBuilder customBuilder)
+		public void SetCustomAttribute (CustomAttributeBuilder customBuilder)
 		{
 			if (customBuilder == null)
 				throw new ArgumentNullException ("customBuilder");
@@ -385,6 +434,9 @@ namespace System.Reflection.Emit {
 					attrs |= MethodAttributes.SpecialName;
 					return;
 #endif
+				case "System.Security.SuppressUnmanagedCodeSecurityAttribute":
+					attrs |= MethodAttributes.HasSecurity;
+					break;
 			}
 
 			if (cattrs != null) {
@@ -401,19 +453,23 @@ namespace System.Reflection.Emit {
 #if NET_2_0
 		[ComVisible (true)]
 #endif
-		public void SetCustomAttribute( ConstructorInfo con, byte[] binaryAttribute) {
+		public void SetCustomAttribute (ConstructorInfo con, byte[] binaryAttribute)
+		{
 			if (con == null)
 				throw new ArgumentNullException ("con");
 			if (binaryAttribute == null)
 				throw new ArgumentNullException ("binaryAttribute");
 			SetCustomAttribute (new CustomAttributeBuilder (con, binaryAttribute));
 		}
-		public void SetImplementationFlags( MethodImplAttributes attributes) {
+
+		public void SetImplementationFlags (MethodImplAttributes attributes)
+		{
 			RejectIfCreated ();
 			iattrs = attributes;
 		}
 
-		public void AddDeclarativeSecurity( SecurityAction action, PermissionSet pset) {
+		public void AddDeclarativeSecurity (SecurityAction action, PermissionSet pset)
+		{
 			if (pset == null)
 				throw new ArgumentNullException ("pset");
 			if ((action == SecurityAction.RequestMinimum) ||
@@ -472,26 +528,30 @@ namespace System.Reflection.Emit {
 			return name.GetHashCode ();
 		}
 
-		internal override int get_next_table_index (object obj, int table, bool inc) {
+		internal override int get_next_table_index (object obj, int table, bool inc)
+		{
 			return type.get_next_table_index (obj, table, inc);
 		}
 
-		internal void set_override (MethodInfo mdecl) {
+		internal void set_override (MethodInfo mdecl)
+		{
 			override_method = mdecl;
 		}
 
-		private void RejectIfCreated () {
+		private void RejectIfCreated ()
+		{
 			if (type.is_created)
 				throw new InvalidOperationException ("Type definition of the method is complete.");
 		}
 
-		private Exception NotSupported () {
+		private Exception NotSupported ()
+		{
 			return new NotSupportedException ("The invoked member is not supported in a dynamic module.");
 		}
 
 #if NET_2_0 || BOOTSTRAP_NET_2_0
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		public override extern MethodInfo MakeGenericMethod (params Type [] types);
+		public override extern MethodInfo MakeGenericMethod (params Type [] typeArguments);
 
 		public override bool IsGenericMethodDefinition {
 			get {
@@ -516,7 +576,7 @@ namespace System.Reflection.Emit {
 		public override Type[] GetGenericArguments ()
 		{
 			if (generic_params == null)
-				return new Type [0];
+				return Type.EmptyTypes;
 
 			Type[] result = new Type [generic_params.Length];
 			for (int i = 0; i < generic_params.Length; i++)
@@ -538,20 +598,6 @@ namespace System.Reflection.Emit {
 			return generic_params;
 		}
 
-		public void SetGenericMethodSignature (MethodAttributes attributes, CallingConventions callingConvention, Type return_type, Type[] parameter_types)
-		{
-			RejectIfCreated ();
-
-			this.attrs = attributes;
-			this.call_conv = callingConvention;
-			if ((attributes & MethodAttributes.Static) == 0)
-				this.call_conv |= CallingConventions.HasThis;
-
-			this.rtype = return_type;
-			this.parameters = new Type [parameter_types.Length];
-			System.Array.Copy (parameter_types, this.parameters, parameter_types.Length);
-		}
-
 		public void SetReturnType (Type returnType)
 		{
 			rtype = returnType;
@@ -586,26 +632,24 @@ namespace System.Reflection.Emit {
 		}
 #endif
 
-                void _MethodBuilder.GetIDsOfNames([In] ref Guid riid, IntPtr rgszNames, uint cNames, uint lcid, IntPtr rgDispId)
-                {
-                        throw new NotImplementedException ();
-                }
+		void _MethodBuilder.GetIDsOfNames([In] ref Guid riid, IntPtr rgszNames, uint cNames, uint lcid, IntPtr rgDispId)
+		{
+			throw new NotImplementedException ();
+		}
 
-                void _MethodBuilder.GetTypeInfo (uint iTInfo, uint lcid, IntPtr ppTInfo)
-                {
-                        throw new NotImplementedException ();
-                }
+		void _MethodBuilder.GetTypeInfo (uint iTInfo, uint lcid, IntPtr ppTInfo)
+		{
+			throw new NotImplementedException ();
+		}
 
-                void _MethodBuilder.GetTypeInfoCount (out uint pcTInfo)
-                {
-                        throw new NotImplementedException ();
-                }
-
-                void _MethodBuilder.Invoke (uint dispIdMember, [In] ref Guid riid, uint lcid, short wFlags, IntPtr pDispParams, IntPtr pVarResult, IntPtr pExcepInfo, IntPtr puArgErr)
-                {
-                        throw new NotImplementedException ();
-                }
+		void _MethodBuilder.GetTypeInfoCount (out uint pcTInfo)
+		{
+			throw new NotImplementedException ();
+		}
 
+		void _MethodBuilder.Invoke (uint dispIdMember, [In] ref Guid riid, uint lcid, short wFlags, IntPtr pDispParams, IntPtr pVarResult, IntPtr pExcepInfo, IntPtr puArgErr)
+		{
+			throw new NotImplementedException ();
+		}
 	}
 }
-
