@@ -81,8 +81,8 @@ public abstract class Encoding
 		case 12001: // UTF32
 		case 65000: // UTF7
 		case 65001: // UTF8
-			decoder_fallback = new DecoderReplacementFallback (String.Empty);
-			encoder_fallback = new EncoderReplacementFallback (String.Empty);
+			decoder_fallback = DecoderFallback.StandardSafeFallback;
+			encoder_fallback = EncoderFallback.StandardSafeFallback;
 			break;
 		}
 #endif
@@ -107,14 +107,9 @@ public abstract class Encoding
 		get { return false; }
 	}
 
-	[MonoTODO ("not used yet")]
 	[ComVisible (false)]
 	public DecoderFallback DecoderFallback {
-		get {
-			if (decoder_fallback == null)
-				decoder_fallback = new DecoderReplacementFallback (String.Empty);
-			return decoder_fallback;
-		}
+		get { return decoder_fallback; }
 		set {
 			if (IsReadOnly)
 				throw new InvalidOperationException ("This Encoding is readonly.");
@@ -124,14 +119,9 @@ public abstract class Encoding
 		}
 	}
 
-	[MonoTODO ("not used yet")]
 	[ComVisible (false)]
 	public EncoderFallback EncoderFallback {
-		get {
-			if (encoder_fallback == null)
-				encoder_fallback = new EncoderReplacementFallback (String.Empty);
-			return encoder_fallback;
-		}
+		get { return encoder_fallback; }
 		set {
 			if (IsReadOnly)
 				throw new InvalidOperationException ("This Encoding is readonly.");
@@ -189,9 +179,9 @@ public abstract class Encoding
 	}
 
 	// Determine if two Encoding objects are equal.
-	public override bool Equals (Object obj)
+	public override bool Equals (Object value)
 	{
-		Encoding enc = (obj as Encoding);
+		Encoding enc = (value as Encoding);
 		if (enc != null) {
 #if NET_2_0
 			return codePage == enc.codePage &&
@@ -249,7 +239,7 @@ public abstract class Encoding
 #if NET_2_0
 		if (charIndex < 0 || charIndex > s.Length)
 			throw new ArgumentOutOfRangeException ("charIndex", _("ArgRange_Array"));
-		if (charCount < 0 || charIndex + charCount > s.Length)
+		if (charCount < 0 || charIndex > (s.Length - charCount))
 			throw new ArgumentOutOfRangeException ("charCount", _("ArgRange_Array"));
 		if (byteIndex < 0 || byteIndex > bytes.Length)
 			throw new ArgumentOutOfRangeException ("byteIndex", _("ArgRange_Array"));
@@ -452,14 +442,14 @@ public abstract class Encoding
 #else
 	public
 #endif
-	static Encoding GetEncoding (int codePage)
+	static Encoding GetEncoding (int codepage)
 	{
-		if (codePage < 0 || codePage > 0xffff)
+		if (codepage < 0 || codepage > 0xffff)
 			throw new ArgumentOutOfRangeException ("codepage", 
 				"Valid values are between 0 and 65535, inclusive.");
 
 		// Check for the builtin code pages first.
-		switch (codePage) {
+		switch (codepage) {
 			case 0: return Default;
 
 			case ASCIIEncoding.ASCII_CODE_PAGE:
@@ -492,14 +482,14 @@ public abstract class Encoding
 		}
 
 		// Try to obtain a code page handler from the I18N handler.
-		Encoding enc = (Encoding)(InvokeI18N ("GetEncoding", codePage));
+		Encoding enc = (Encoding)(InvokeI18N ("GetEncoding", codepage));
 		if (enc != null) {
 			enc.is_readonly = true;
 			return enc;
 		}
 
 		// Build a code page class name.
-		String cpName = "System.Text.CP" + codePage.ToString ();
+		String cpName = "System.Text.CP" + codepage.ToString ();
 
 		// Look for a code page converter in this assembly.
 		Assembly assembly = Assembly.GetExecutingAssembly ();
@@ -521,7 +511,7 @@ public abstract class Encoding
 
 		// We have no idea how to handle this code page.
 		throw new NotSupportedException
-			(String.Format ("CodePage {0} not supported", codePage.ToString ()));
+			(String.Format ("CodePage {0} not supported", codepage.ToString ()));
 	}
 
 #if !ECMA_COMPAT
@@ -535,7 +525,7 @@ public abstract class Encoding
 		return e;
 	}
 
-	public static Encoding GetEncoding (int codePage,
+	public static Encoding GetEncoding (int codepage,
 		EncoderFallback encoderFallback, DecoderFallback decoderFallback)
 	{
 		if (encoderFallback == null)
@@ -543,7 +533,7 @@ public abstract class Encoding
 		if (decoderFallback == null)
 			throw new ArgumentNullException ("decoderFallback");
 
-		Encoding e = GetEncoding (codePage).Clone () as Encoding;
+		Encoding e = GetEncoding (codepage).Clone () as Encoding;
 		e.is_readonly = false;
 		e.encoder_fallback = encoderFallback;
 		e.decoder_fallback = decoderFallback;
@@ -602,11 +592,13 @@ public abstract class Encoding
 		return encoding_infos;
 	}
 
+	[ComVisible (false)]
 	public bool IsAlwaysNormalized ()
 	{
 		return IsAlwaysNormalized (NormalizationForm.FormC);
 	}
 
+	[ComVisible (false)]
 	public virtual bool IsAlwaysNormalized (NormalizationForm form)
 	{
 		// umm, ASCIIEncoding should have overriden this method, no?
@@ -848,6 +840,7 @@ public abstract class Encoding
 	static volatile Encoding isoLatin1Encoding;
 	static volatile Encoding unixConsoleEncoding;
 #if NET_2_0
+	static volatile Encoding utf8EncodingUnsafe;
 	static volatile Encoding utf32Encoding;
 	static volatile Encoding bigEndianUTF32Encoding;
 #endif
@@ -1010,6 +1003,30 @@ public abstract class Encoding
 		}
 	}
 	
+	//
+	// Only internal, to be used by the class libraries: Unmarked and non-input-validating
+	//
+	internal static Encoding UTF8UnmarkedUnsafe {
+		get {
+#if NET_2_0
+			if (utf8EncodingUnsafe == null) {
+				lock (lockobj){
+					if (utf8EncodingUnsafe == null){
+						utf8EncodingUnsafe = new UTF8Encoding (false, false);
+						utf8EncodingUnsafe.is_readonly = false;
+						utf8EncodingUnsafe.DecoderFallback = new DecoderReplacementFallback (String.Empty);
+						utf8EncodingUnsafe.is_readonly = true;
+					}
+				}
+			}
+
+			return utf8EncodingUnsafe;
+#else
+			return UTF8Unmarked;
+#endif
+		}
+	}
+	
 	// Get the standard little-endian Unicode encoding object.
 	public static Encoding Unicode
 	{
