@@ -36,6 +36,7 @@
 using System.Collections;
 using System.Globalization;
 using System.Runtime.CompilerServices;
+using System.Runtime.InteropServices;
 
 namespace System
 {
@@ -44,6 +45,7 @@ namespace System
 		internal Type utype;
 		internal Array values;
 		internal string[] names;
+		[ThreadStatic]
 		static Hashtable cache;
 		
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
@@ -51,9 +53,16 @@ namespace System
 
 		static MonoEnumInfo ()
 		{
-		    	cache = new Hashtable ();
 		}
-		
+
+		static Hashtable Cache {
+			get {
+				if (cache == null) {
+					cache = new Hashtable ();
+				}
+				return cache;
+			}
+		}
 		private MonoEnumInfo (MonoEnumInfo other)
 		{
 			utype = other.utype;
@@ -63,19 +72,20 @@ namespace System
 
 		internal static void GetInfo (Type enumType, out MonoEnumInfo info)
 		{
-			lock (cache) {
-				if (cache.ContainsKey (enumType)) {
-					info = (MonoEnumInfo) cache [enumType];
-					return;
-				}
-				get_enum_info (enumType, out info);
-				Array.Sort (info.values, info.names);
-				cache.Add (enumType, new MonoEnumInfo (info));
+			if (Cache.ContainsKey (enumType)) {
+				info = (MonoEnumInfo) cache [enumType];
+				return;
 			}
+			get_enum_info (enumType, out info);
+			Array.Sort (info.values, info.names);
+			cache.Add (enumType, new MonoEnumInfo (info));
 		}
 	};
 
 	[Serializable]
+#if NET_2_0
+	[ComVisible (true)]
+#endif
 	public abstract class Enum : ValueType, IComparable, IConvertible, IFormattable
 	{
 		protected Enum ()
@@ -85,9 +95,7 @@ namespace System
 		// IConvertible methods Start -->
 		public TypeCode GetTypeCode ()
 		{
-			MonoEnumInfo info;
-			MonoEnumInfo.GetInfo (this.GetType (), out info);
-			return Type.GetTypeCode (info.utype);
+			return Type.GetTypeCode (GetUnderlyingType (this.GetType ()));
 		}
 
 		bool IConvertible.ToBoolean (IFormatProvider provider)
@@ -145,9 +153,9 @@ namespace System
 			return Convert.ToSingle (get_value (), provider);
 		}
 
-		object IConvertible.ToType (Type conversionType, IFormatProvider provider)
+		object IConvertible.ToType (Type type, IFormatProvider provider)
 		{
-			return Convert.ToType (get_value (), conversionType, provider);
+			return Convert.ToType (get_value (), type, provider);
 		}
 
 		ushort IConvertible.ToUInt16 (IFormatProvider provider)
@@ -169,19 +177,25 @@ namespace System
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern object get_value ();
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static Array GetValues (Type enumType)
 		{
 			if (enumType == null)
 				throw new ArgumentNullException ("enumType");
 
 			if (!enumType.IsEnum)
-				throw new ArgumentException ("enumType is not an Enum type.");
+				throw new ArgumentException ("enumType is not an Enum type.", "enumType");
 
 			MonoEnumInfo info;
 			MonoEnumInfo.GetInfo (enumType, out info);
 			return (Array) info.values.Clone ();
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static string[] GetNames (Type enumType)
 		{
 			if (enumType == null)
@@ -195,6 +209,9 @@ namespace System
 			return (string []) info.names.Clone ();
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static string GetName (Type enumType, object value)
 		{
 			if (enumType == null)
@@ -203,19 +220,18 @@ namespace System
 				throw new ArgumentNullException ("value");
 
 			if (!enumType.IsEnum)
-				throw new ArgumentException ("enumType is not an Enum type.");
+				throw new ArgumentException ("enumType is not an Enum type.", "enumType");
 
 			MonoEnumInfo info;
-			int i;
 			value = ToObject (enumType, value);
 			MonoEnumInfo.GetInfo (enumType, out info);
-			for (i = 0; i < info.values.Length; ++i) {
-				if (value.Equals (info.values.GetValue (i)))
-					return info.names [i];
-			}
-			return null;
+			int i = Array.BinarySearch (info.values, value);
+			return (i >= 0) ? info.names [i] : null;
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static bool IsDefined (Type enumType, object value)
 		{
 			if (enumType == null)
@@ -224,7 +240,7 @@ namespace System
 				throw new ArgumentNullException ("value");
 
 			if (!enumType.IsEnum)
-				throw new ArgumentException ("enumType is not an Enum type.");
+				throw new ArgumentException ("enumType is not an Enum type.", "enumType");
 
 			MonoEnumInfo info;
 			MonoEnumInfo.GetInfo (enumType, out info);
@@ -233,14 +249,9 @@ namespace System
 			if (vType == typeof(String)) {
 				return ((IList)(info.names)).Contains (value);
 			} else if ((vType == info.utype) || (vType == enumType)) {
-				int i;
 				value = ToObject (enumType, value);
 				MonoEnumInfo.GetInfo (enumType, out info);
-				for (i = 0; i < info.values.Length; ++i) {
-					if (value.Equals (info.values.GetValue (i)))
-						return true;
-				}
-				return false;
+				return (Array.BinarySearch (info.values, value) >= 0);
 			} else {
 				throw new ArgumentException("The value parameter is not the correct type."
 					+ "It must be type String or the same type as the underlying type"
@@ -248,19 +259,26 @@ namespace System
 			}
 		}
 
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private static extern Type get_underlying_type (Type enumType);
+
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static Type GetUnderlyingType (Type enumType)
 		{
 			if (enumType == null)
 				throw new ArgumentNullException ("enumType");
 
 			if (!enumType.IsEnum)
-				throw new ArgumentException ("enumType is not an Enum type.");
+				throw new ArgumentException ("enumType is not an Enum type.", "enumType");
 
-			MonoEnumInfo info;
-			MonoEnumInfo.GetInfo (enumType, out info);
-			return info.utype;
+			return get_underlying_type (enumType);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static object Parse (Type enumType, string value)
 		{
 			// Note: Parameters are checked in the other overload
@@ -269,9 +287,16 @@ namespace System
 
 		private static int FindName (string [] names, string name,  bool ignoreCase)
 		{
-			for (int i = 0; i < names.Length; ++i) {
-				if (String.Compare (name, names [i], ignoreCase, CultureInfo.InvariantCulture) == 0)
-					return i;
+			if (!ignoreCase) {
+				for (int i = 0; i < names.Length; ++i) {
+					if (name == names [i])
+						return i;
+				}
+			} else {
+				for (int i = 0; i < names.Length; ++i) {
+					if (String.Compare (name, names [i], ignoreCase, CultureInfo.InvariantCulture) == 0)
+						return i;
+				}
 			}
 			return -1;
 		}
@@ -302,6 +327,9 @@ namespace System
 
 		private static char [] split_char = { ',' };
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static object Parse (Type enumType, string value, bool ignoreCase)
 		{
 			if (enumType == null)
@@ -311,11 +339,11 @@ namespace System
 				throw new ArgumentNullException ("value");
 
 			if (!enumType.IsEnum)
-				throw new ArgumentException ("not an Enum type", "enumType");
+				throw new ArgumentException ("enumType is not an Enum type.", "enumType");
 
 			value = value.Trim ();
-			if (String.Empty == value)
-				throw new ArgumentException ("cannot be an empty string", "value");
+			if (value.Length == 0)
+				throw new ArgumentException ("An empty string is not considered a valid value.");
 
 			MonoEnumInfo info;
 			MonoEnumInfo.GetInfo (enumType, out info);
@@ -343,8 +371,8 @@ namespace System
 			// is 'value' a number?
 			try {
 				return ToObject (enumType, Convert.ChangeType (value, typeCode));
-			} catch (Exception e) {
-				throw new ArgumentException (String.Format ("The requested value `{0}' was not found", value), e);
+			} catch (FormatException) {
+				throw new ArgumentException (String.Format ("The requested value '{0}' was not found.", value));
 			}
 		}
 
@@ -353,27 +381,24 @@ namespace System
 		/// </summary>
 		///
 		/// <remarks/>
-		public int CompareTo (object obj)
+		public int CompareTo (object target)
 		{
 			Type thisType;
 
-			if (obj == null)
+			if (target == null)
 				return 1;
 
 			thisType = this.GetType ();
-			if (obj.GetType() != thisType) {
-				throw new ArgumentException (
-					"Object must be the same type as the "
-					+ "enum. The type passed in was " 
-					+ obj.GetType().ToString ()
-					+ "; the enum type was " 
-					+ thisType.ToString () + ".");
+			if (target.GetType() != thisType) {
+				throw new ArgumentException (String.Format (
+					"Object must be the same type as the enum. The type passed in was {0}; the enum type was {1}.", 
+					target.GetType(), thisType));
 			}
 
 			object value1, value2;
 
 			value1 = this.get_value ();
-			value2 = ((Enum)obj).get_value ();
+			value2 = ((Enum)target).get_value ();
 
 			return ((IComparable)value1).CompareTo (value2);
 		}
@@ -384,7 +409,7 @@ namespace System
 		}
 
 #if NET_2_0
-	[Obsolete("Provider is ignored, just use ToString")]
+		[Obsolete("Provider is ignored, just use ToString")]
 #endif
 		public string ToString (IFormatProvider provider)
 		{
@@ -412,47 +437,74 @@ namespace System
 			return Format (this.GetType(), this.get_value (), format);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static object ToObject (Type enumType, byte value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static object ToObject (Type enumType, short value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static object ToObject (Type enumType, int value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static object ToObject (Type enumType, long value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public static extern object ToObject (Type enumType, object value);
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, sbyte value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, ushort value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, uint value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, ulong value)
 		{
@@ -645,6 +697,9 @@ namespace System
 			return retVal;
 		}
 
+#if NET_2_0
+		[ComVisible (true)]
+#endif
 		public static string Format (Type enumType, object value, string format)
 		{
 			if (enumType == null)
@@ -655,7 +710,7 @@ namespace System
 				throw new ArgumentNullException ("format");
 
 			if (!enumType.IsEnum)
-				throw new ArgumentException ("Type provided must be an Enum.");
+				throw new ArgumentException ("enumType is not an Enum type.", "enumType");
 			
 			Type vType = value.GetType();
 			Type underlyingType = Enum.GetUnderlyingType (enumType);
@@ -679,7 +734,7 @@ namespace System
 			char formatChar = format [0];
 			string retVal;
 			if ((formatChar == 'G' || formatChar == 'g')) {
-				if (!Attribute.IsDefined (enumType, typeof(FlagsAttribute))) {
+				if (!enumType.IsDefined (typeof(FlagsAttribute), false)) {
 					retVal = GetName (enumType, value);
 					if (retVal == null)
 						retVal = value.ToString();
