@@ -38,6 +38,7 @@ using System.Reflection;
 #if !TARGET_JVM
 using System.Reflection.Emit;
 #endif
+using System.Threading;
 using System.Runtime.Serialization;
 using System.Security;
 
@@ -46,6 +47,28 @@ namespace MonoTests.System.Reflection
 	[TestFixture]
 	public class AssemblyTest
 	{
+		static string TempFolder = Path.Combine (Path.GetTempPath (),
+			"MonoTests.System.Reflection.AssemblyTest");
+
+		[SetUp]
+		public void SetUp ()
+		{
+			while (Directory.Exists (TempFolder))
+				TempFolder = Path.Combine (TempFolder, "2");
+			Directory.CreateDirectory (TempFolder);
+		}
+
+		[TearDown]
+		public void TearDown ()
+		{
+			try {
+				// This throws an exception under MS.NET, since the directory contains loaded
+				// assemblies.
+				Directory.Delete (TempFolder, true);
+			} catch (Exception) {
+			}
+		}
+
 		[Test] 
 		public void CreateInstance() 
 		{
@@ -63,19 +86,51 @@ namespace MonoTests.System.Reflection
 			Assert.IsNull (obj, "#03");
 		}
 
-		[Test]
+		[Test] // bug #49114
 #if NET_2_0
 		[Category ("NotWorking")]
 		[ExpectedException (typeof (ArgumentException))]
 #else
 		[ExpectedException (typeof (TypeLoadException))]
 #endif
-		public void TestGetType () 
+		public void GetType_TypeName_Invalid () 
 		{
-			// Bug #49114
 			typeof (int).Assembly.GetType ("&blabla", true, true);
 		}
 
+		[Test] // bug #334203
+		public void GetType_TypeName_AssemblyName ()
+		{
+			Assembly a = typeof (int).Assembly;
+			string typeName = typeof (string).AssemblyQualifiedName;
+#if NET_2_0
+			try {
+				a.GetType (typeName, true, false);
+				Assert.Fail ("#A1");
+			} catch (ArgumentException ex) {
+				Assert.AreEqual (typeof (ArgumentException), ex.GetType (), "#A2");
+				Assert.IsNull (ex.InnerException, "#A3");
+				Assert.IsNotNull (ex.Message, "#A4");
+				Assert.IsNull (ex.ParamName, "#A5");
+			}
+#else
+			try {
+				a.GetType (typeName, true, false);
+				Assert.Fail ("#A1");
+			} catch (TypeLoadException ex) {
+				Assert.AreEqual (typeof (TypeLoadException), ex.GetType (), "#A2");
+				Assert.IsNull (ex.InnerException, "#A3");
+				Assert.IsNotNull (ex.Message, "#A4");
+				Assert.IsTrue (ex.Message.IndexOf (typeName) != -1, "#A5");
+			}
+#endif
+
+			Type type = a.GetType (typeName, false);
+			Assert.IsNull (type, "#B1");
+			type = a.GetType (typeName, false, true);
+			Assert.IsNull (type, "#B2");
+		}
+
 		[Test]
 		public void GetEntryAssembly ()
 		{
@@ -96,6 +151,31 @@ namespace MonoTests.System.Reflection
 			}
 		}
 
+#if !TARGET_JVM // Reflection.Emit is not supported.
+		[Test]
+		public void GetModules_MissingFile ()
+		{
+			AssemblyName newName = new AssemblyName ();
+			newName.Name = "AssemblyTest";
+
+			AssemblyBuilder ab = Thread.GetDomain().DefineDynamicAssembly (newName, AssemblyBuilderAccess.RunAndSave, TempFolder);
+
+			ModuleBuilder mb = ab.DefineDynamicModule ("myDynamicModule1", "myDynamicModule.dll", true);
+
+			ab.Save ("test_assembly.dll");
+
+			File.Delete (Path.Combine (TempFolder, "myDynamicModule.dll"));
+
+			Assembly ass = Assembly.LoadFrom (Path.Combine (TempFolder, "test_assembly.dll"));
+			try {
+				ass.GetModules ();
+				Assert.Fail ();
+			} catch (FileNotFoundException ex) {
+				Assert.AreEqual ("myDynamicModule.dll", ex.FileName);
+			}
+		}
+#endif
+
 #if !TARGET_JVM // ManifestModule not supported under TARGET_JVM.
 #if NET_2_0
 		[Category ("NotWorking")]
@@ -156,18 +236,33 @@ namespace MonoTests.System.Reflection
 
 		[Test]
 		[Category("TargetJvmNotWorking")] // Not yet supported for TARGET_JVM
-		[ExpectedException (typeof (ArgumentNullException))]
 		public void GetFile_Null ()
 		{
-			Assembly.GetExecutingAssembly ().GetFile (null);
+			try {
+				Assembly.GetExecutingAssembly ().GetFile (null);
+				Assert.Fail ("#1");
+			} catch (ArgumentNullException ex) {
+				Assert.AreEqual (typeof (ArgumentNullException), ex.GetType (), "#2");
+				Assert.IsNull (ex.InnerException, "#3");
+				Assert.IsNotNull (ex.Message, "#4");
+				Assert.IsNull (ex.ParamName, "#5");
+			}
 		}
 
 		[Test]
 		[Category("TargetJvmNotWorking")] // Not yet supported for TARGET_JVM
-		[ExpectedException (typeof (ArgumentException))]
 		public void GetFile_Empty ()
 		{
-			Assembly.GetExecutingAssembly ().GetFile (String.Empty);
+			try {
+				Assembly.GetExecutingAssembly ().GetFile (
+					String.Empty);
+				Assert.Fail ("#1");
+			} catch (ArgumentException ex) {
+				Assert.AreEqual (typeof (ArgumentException), ex.GetType (), "#2");
+				Assert.IsNull (ex.InnerException, "#3");
+				Assert.IsNotNull (ex.Message, "#4");
+				Assert.IsNull (ex.ParamName, "#5");
+			}
 		}
 
 		[Test]
@@ -196,6 +291,79 @@ namespace MonoTests.System.Reflection
 			Assert.IsTrue (fss.Length <= corlib_test.GetFiles (true).Length, "test.GetFiles (true)");
 		}
 
+		[Test]
+		public void GetManifestResourceStream_Name_Empty ()
+		{
+			Assembly corlib = typeof (int).Assembly;
+
+			try {
+				corlib.GetManifestResourceStream (string.Empty);
+				Assert.Fail ("#A1");
+			} catch (ArgumentException ex) {
+				// String cannot have zero length
+				Assert.AreEqual (typeof (ArgumentException), ex.GetType (), "#A2");
+				Assert.IsNull (ex.InnerException, "#A3");
+				Assert.IsNotNull (ex.Message, "#A4");
+			}
+
+			corlib.GetManifestResourceStream (typeof (int), string.Empty);
+
+			try {
+				corlib.GetManifestResourceStream ((Type) null, string.Empty);
+				Assert.Fail ("#B1");
+			} catch (ArgumentException ex) {
+				// String cannot have zero length
+				Assert.AreEqual (typeof (ArgumentException), ex.GetType (), "#B2");
+				Assert.IsNull (ex.InnerException, "#B3");
+				Assert.IsNotNull (ex.Message, "#B4");
+			}
+		}
+
+		[Test]
+		public void GetManifestResourceStream_Name_Null ()
+		{
+			Assembly corlib = typeof (int).Assembly;
+
+			try {
+				corlib.GetManifestResourceStream ((string) null);
+				Assert.Fail ("#A1");
+			} catch (ArgumentNullException ex) {
+				Assert.AreEqual (typeof (ArgumentNullException), ex.GetType (), "#A2");
+				Assert.IsNull (ex.InnerException, "#A3");
+				Assert.IsNotNull (ex.Message, "#A4");
+			}
+
+			corlib.GetManifestResourceStream (typeof (int), (string) null);
+
+			try {
+				corlib.GetManifestResourceStream ((Type) null, (string) null);
+				Assert.Fail ("#B1");
+			} catch (ArgumentNullException ex) {
+				Assert.AreEqual (typeof (ArgumentNullException), ex.GetType (), "#B2");
+				Assert.IsNull (ex.InnerException, "#B3");
+				Assert.IsNotNull (ex.Message, "#B4");
+				Assert.IsNotNull (ex.ParamName, "#B5");
+				Assert.AreEqual ("type", ex.ParamName, "#B6");
+			}
+		}
+
+		[Test]
+		public void IsDefined_AttributeType_Null ()
+		{
+			Assembly corlib = typeof (int).Assembly;
+
+			try {
+				corlib.IsDefined ((Type) null, false);
+				Assert.Fail ("#1");
+			} catch (ArgumentNullException ex) {
+				Assert.AreEqual (typeof (ArgumentNullException), ex.GetType (), "#2");
+				Assert.IsNull (ex.InnerException, "#3");
+				Assert.IsNotNull (ex.Message, "#4");
+				Assert.IsNotNull (ex.ParamName, "#5");
+				Assert.AreEqual ("attributeType", ex.ParamName, "#6");
+			}
+		}
+
 		[Test] // bug #78517
 #if ONLY_1_1
 		[Category ("NotDotNet")] // MS.NET 1.x throws FileLoadException
@@ -262,11 +430,20 @@ namespace MonoTests.System.Reflection
 
 #if !TARGET_JVM // GetObjectData currently not implemented for Assembly.
 		[Test]
-		[ExpectedException (typeof (ArgumentNullException))]
-		public void GetObjectData_Null ()
+		public void GetObjectData_Info_Null ()
 		{
 			Assembly corlib = typeof (int).Assembly;
-			corlib.GetObjectData (null, new StreamingContext (StreamingContextStates.All));
+			try {
+				corlib.GetObjectData (null, new StreamingContext (
+					StreamingContextStates.All));
+				Assert.Fail ("#1");
+			} catch (ArgumentNullException ex) {
+				Assert.AreEqual (typeof (ArgumentNullException), ex.GetType (), "#2");
+				Assert.IsNull (ex.InnerException, "#3");
+				Assert.IsNotNull (ex.Message, "#4");
+				Assert.IsNotNull (ex.ParamName, "#5");
+				Assert.AreEqual ("info", ex.ParamName, "#6");
+			}
 		}
 #endif // TARGET_JVM
 
@@ -341,7 +518,6 @@ namespace MonoTests.System.Reflection
 		}
 
 		[Test]
-		[Category ("NotWorking")]
 		public void bug78465 ()
 		{
 			string assemblyFileName = Path.Combine (
@@ -648,6 +824,7 @@ namespace MonoTests.System.Reflection
 			}
 		}
 
+
 		[Test] // bug #79715
 		public void Load_PartialVersion ()
 		{
@@ -911,6 +1088,22 @@ namespace MonoTests.System.Reflection
 		}
 #endif // TARGET_JVM
 
+#if NET_2_0
+		[Test]
+		public void ManifestModule ()
+		{
+			Assembly assembly = typeof (int).Assembly;
+			Module module = assembly.ManifestModule;
+			Assert.IsNotNull (module, "#1");
+			Assert.AreEqual (typeof (Module), module.GetType (), "#2");
+			Assert.AreEqual ("mscorlib.dll", module.Name, "#3");
+			Assert.IsFalse (module.IsResource (), "#4");
+			Assert.IsTrue (assembly.GetModules ().Length > 0, "#5");
+			Assert.AreSame (module, assembly.GetModules () [0], "#6");
+			Assert.AreSame (module, assembly.ManifestModule, "#7");
+		}
+#endif
+
 		[Serializable ()]
 		private class AssemblyResolveHandler
 		{
@@ -934,5 +1127,32 @@ namespace MonoTests.System.Reflection
 			private readonly string _assemblyFile;
 			private readonly string _assemblyName;
 		}
+
+		protected internal class Bug328812_NestedFamORAssem { };
+
+		[Test]
+		public void bug328812 ()
+		{
+			Assembly corlib_test = Assembly.GetExecutingAssembly ();
+			Assert.IsNull (corlib_test.GetType ("Bug328812_NestedFamORAssem"));
+			// Just a sanity check, in case the above passed for some other reason
+			Assert.IsNotNull (corlib_test.GetType ("MonoTests.System.Reflection.AssemblyTest+Bug328812_NestedFamORAssem"));
+		}
+		
+		[Test]
+		public void GetCustomAttributes_AttributeType_Null ()
+		{
+			Assembly a = typeof (int).Assembly;
+			try {
+				a.GetCustomAttributes (null, false);
+				Assert.Fail ("#1");
+			} catch (ArgumentNullException ex) {
+				Assert.AreEqual (typeof (ArgumentNullException), ex.GetType (), "#2");
+				Assert.IsNull (ex.InnerException, "#3");
+				Assert.IsNotNull (ex.Message, "#4");
+				Assert.IsNotNull (ex.ParamName, "#5");
+				Assert.AreEqual ("attributeType", ex.ParamName, "#6");
+			}
+		}
 	}
 }
