@@ -9,31 +9,33 @@
 // (C) 2004 Novell (http://www.novell.com)
 //
 
-using NUnit.Framework;
 using System;
+using System.Globalization;
 using System.Security.Principal;
 using System.Threading;
 
-namespace MonoTests.System.Threading {
-	
+using NUnit.Framework;
+
+namespace MonoTests.System.Threading
+{
 	// These tests seem to hang the 2.0 framework. So they are disabled for now
 	// Don't reenable them until you can run a few thousand times on an SMP box.
 	[Category ("NotWorking")]
-	public class ThreadedPrincipalTest : Assertion {
-
+	public class ThreadedPrincipalTest
+	{
 		public static void NoPrincipal () 
 		{
 #if !TARGET_JVM // AppDomain.SetPrincipalPolicy not supported for TARGET_JVM
 			AppDomain.CurrentDomain.SetPrincipalPolicy (PrincipalPolicy.NoPrincipal);
 #endif
 			IPrincipal p = Thread.CurrentPrincipal;
-			AssertNull ("Thread.CurrentPrincipal-1", p);
+			Assert.IsNull (p, "#1");
 
 			Thread.CurrentPrincipal = new GenericPrincipal (new GenericIdentity ("mono"), null);
-			AssertNotNull ("Thread.CurrentPrincipal-2", Thread.CurrentPrincipal);
+			Assert.IsNotNull (Thread.CurrentPrincipal, "#2");
 
 			Thread.CurrentPrincipal = null;
-			AssertNull ("Thread.CurrentPrincipal-3", Thread.CurrentPrincipal);
+			Assert.IsNull (Thread.CurrentPrincipal, "#3");
 			// in this case we can return to null
 		}
 
@@ -42,17 +44,17 @@ namespace MonoTests.System.Threading {
 		{
 			AppDomain.CurrentDomain.SetPrincipalPolicy (PrincipalPolicy.UnauthenticatedPrincipal);
 			IPrincipal p = Thread.CurrentPrincipal;
-			AssertNotNull ("Thread.CurrentPrincipal", p);
-			Assert ("Type", (p is GenericPrincipal));
-			AssertEquals ("Name", String.Empty, p.Identity.Name);
-			AssertEquals ("AuthenticationType", String.Empty, p.Identity.AuthenticationType);
-			Assert ("IsAuthenticated", !p.Identity.IsAuthenticated);
+			Assert.IsNotNull (p, "#1");
+			Assert.IsTrue ((p is GenericPrincipal), "#2");
+			Assert.AreEqual (String.Empty, p.Identity.Name, "#3");
+			Assert.AreEqual (String.Empty, p.Identity.AuthenticationType, "#4");
+			Assert.IsFalse (p.Identity.IsAuthenticated, "#5");
 
 			Thread.CurrentPrincipal = new GenericPrincipal (new GenericIdentity ("mono"), null);
-			AssertNotNull ("Thread.CurrentPrincipal-2", Thread.CurrentPrincipal);
+			Assert.IsNotNull (Thread.CurrentPrincipal, "#6");
 
 			Thread.CurrentPrincipal = null;
-			AssertNotNull ("Thread.CurrentPrincipal-3", Thread.CurrentPrincipal);
+			Assert.IsNotNull (Thread.CurrentPrincipal, "#7");
 			// in this case we can't return to null
 		}
 
@@ -60,32 +62,50 @@ namespace MonoTests.System.Threading {
 		{
 			AppDomain.CurrentDomain.SetPrincipalPolicy (PrincipalPolicy.WindowsPrincipal);
 			IPrincipal p = Thread.CurrentPrincipal;
-			AssertNotNull ("Thread.CurrentPrincipal", p);
-			Assert ("Type", (p is WindowsPrincipal));
-			AssertNotNull ("Name", p.Identity.Name);
-			AssertNotNull ("AuthenticationType", p.Identity.AuthenticationType);
-			Assert ("IsAuthenticated", p.Identity.IsAuthenticated);
+			Assert.IsNotNull (p, "#1");
+			Assert.IsTrue ((p is WindowsPrincipal), "#2");
+			Assert.IsNotNull (p.Identity.Name, "#3");
+			Assert.IsNotNull (p.Identity.AuthenticationType, "#4");
+			Assert.IsTrue (p.Identity.IsAuthenticated, "#5");
 
 			// note: we can switch from a WindowsPrincipal to a GenericPrincipal
 			Thread.CurrentPrincipal = new GenericPrincipal (new GenericIdentity ("mono"), null);
-			AssertNotNull ("Thread.CurrentPrincipal-2", Thread.CurrentPrincipal);
+			Assert.IsNotNull (Thread.CurrentPrincipal, "#6");
 
 			Thread.CurrentPrincipal = null;
-			AssertNotNull ("Thread.CurrentPrincipal-3", Thread.CurrentPrincipal);
+			Assert.IsNotNull (Thread.CurrentPrincipal, "#7");
 			// in this case we can't return to null
 		}
 #endif // TARGET_JVM
 
 		public static void CopyOnNewThread ()
 		{
-			AssertNotNull ("Thread.CurrentPrincipal", Thread.CurrentPrincipal);
-			AssertEquals ("Identity", "good", Thread.CurrentPrincipal.Identity.Name);
+			Assert.IsNotNull (Thread.CurrentPrincipal, "#1");
+			Assert.AreEqual ("good", Thread.CurrentPrincipal.Identity.Name, "#2");
 		}
 	}
 
 	[TestFixture]
-	[Category ("NotWorking")]
-	public class ThreadTest : Assertion {
+	public class ThreadTest
+	{
+		static bool is_win32;
+		static bool is_mono;
+
+		static ThreadTest ()
+		{
+			switch (Environment.OSVersion.Platform) {
+			case PlatformID.Win32NT:
+			case PlatformID.Win32S:
+			case PlatformID.Win32Windows:
+			case PlatformID.WinCE:
+				is_win32 = true;
+				break;
+			}
+
+			// check a class in mscorlib to determine if we're running on Mono
+			if (Type.GetType ("System.MonoType", false) != null)
+				is_mono = true;
+		}
 
 		//Some Classes to test as threads
 		private class C1Test
@@ -204,90 +224,138 @@ namespace MonoTests.System.Threading {
 			}
 		}
 
+		[Test]
 		public void TestCtor1()
-		{			
+		{
 			C1Test test1 = new C1Test();
-			try
-			{
-				Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
-			}
-			catch (Exception e)
-			{
-				Fail ("#01 Unexpected Exception Thrown: " + e.ToString ());
-			}
+			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 		}
 
-		[Category("NotDotNet")]
+		[Test] // bug #325566
+		public void GetHashCodeTest ()
+		{
+			C1Test test1 = new C1Test ();
+			Thread tA = new Thread (new ThreadStart (test1.TestMethod));
+			int hA1 = tA.GetHashCode ();
+#if NET_2_0
+			Assert.IsTrue (hA1 > 0, "#A1");
+#endif
+			tA.Start ();
+			int hA2 = tA.GetHashCode ();
+			Assert.AreEqual (hA1, hA2, "#A2");
+			tA.Join ();
+			int hA3 = tA.GetHashCode ();
+			Assert.AreEqual (hA1, hA3, "#A3");
+#if NET_2_0
+			Assert.AreEqual (hA1, tA.ManagedThreadId, "#A4");
+#endif
+
+			test1 = new C1Test ();
+			Thread tB = new Thread (new ThreadStart (test1.TestMethod));
+			int hB1 = tB.GetHashCode ();
+#if NET_2_0
+			Assert.IsTrue (hB1 > 0, "#B1");
+#endif
+			tB.Start ();
+			int hB2 = tB.GetHashCode ();
+			Assert.AreEqual (hB1, hB2, "#B2");
+			tB.Join ();
+			int hB3 = tB.GetHashCode ();
+			Assert.AreEqual (hB1, hB3, "#B3");
+#if NET_2_0
+			Assert.AreEqual (hB1, tB.ManagedThreadId, "#B4");
+#endif
+			Assert.IsFalse (hA2 == hB2, "#B5");
+		}
+
+#if NET_2_0
+		[Test] // bug #82700
+		public void ManagedThreadId ()
+		{
+			C1Test test1 = new C1Test ();
+			Thread t1 = new Thread (new ThreadStart (test1.TestMethod));
+			int mtA1 = t1.ManagedThreadId;
+			t1.Start ();
+			int mtA2 = t1.ManagedThreadId;
+			t1.Join ();
+			int mtA3 = t1.ManagedThreadId;
+			Assert.AreEqual (mtA1, mtA2, "#A1");
+			Assert.AreEqual (mtA2, mtA3, "#A2");
+
+			test1 = new C1Test ();
+			Thread t2 = new Thread (new ThreadStart (test1.TestMethod));
+			int mtB1 = t2.ManagedThreadId;
+			t2.Start ();
+			int mtB2 = t2.ManagedThreadId;
+			t2.Join ();
+			int mtB3 = t2.ManagedThreadId;
+			Assert.AreEqual (mtB1, mtB2, "#B1");
+			Assert.AreEqual (mtB2, mtB3, "#B2");
+			Assert.IsFalse (mtB1 == mtA1, "#B3");
+		}
+#endif
+
+		[Test]
+		[Category ("NotDotNet")] // it hangs.
 		public void TestStart()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on Win32. The test should be fixed.");
 		{
 			C1Test test1 = new C1Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
-			try
-			{
-				TestThread.Start();
-			}
-			catch (Exception e)
-			{
-				Fail ("#12 Unexpected Exception Thrown: " + e.ToString ());
-			}
+			TestThread.Start();
 			TestThread.Join();
-			AssertEquals("#13 Thread Not started: ", 10,test1.cnt);
+			Assert.AreEqual (10, test1.cnt, "#1");
 		}
 		{
-			bool errorThrown = false;
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			TestThread.Start();
 			TestThread.Abort();
-			try
-			{
+			try {
 				TestThread.Start();
+				Assert.Fail ("#2");
+			} catch (ThreadStateException) {
 			}
-			catch(ThreadStateException)
-			{
-				errorThrown = true;
-			}
-			Assert ("#14 no ThreadStateException trown", errorThrown);
 		}
 		{
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			TestThread.Start();
-			while(!test1.run);
+			while (!test1.run) {
+			}
 			bool started = (TestThread.ThreadState == ThreadState.Running);
-			AssertEquals("#15 Thread Is not in the correct state: ", started , test1.run);	
+			Assert.AreEqual (started, test1.run, "#15 Thread Is not in the correct state: ");
 			TestThread.Abort();
 		}
 		}
 
-		public void TestApartment()
+		[Test]
+		public void TestApartmentState ()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on mono on win32. Our runtime should be fixed.");
+
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
-			ApartmentState before = TestThread.ApartmentState;
+			Assert.AreEqual (ApartmentState.Unknown, TestThread.ApartmentState, "#1");
 			TestThread.Start();
 			TestUtil.WaitForAlive (TestThread, "wait5");
-			ApartmentState after = TestThread.ApartmentState;
+#if NET_2_0
+			Assert.AreEqual (ApartmentState.MTA, TestThread.ApartmentState, "#2");
+#else
+			Assert.AreEqual (ApartmentState.Unknown, TestThread.ApartmentState, "#3");
+#endif
 			TestThread.Abort();
-			AssertEquals("#21 Apartment State Changed when not needed",before,after);
-		}
-
-		[Category("NotDotNet")]
-		public void TestApartmentState()
-		{
-			C2Test test1 = new C2Test();
-			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
-			ApartmentState before = TestThread.ApartmentState;
-			TestThread.Start();
-			TestUtil.WaitForAlive (TestThread, "wait6");
-			ApartmentState after = TestThread.ApartmentState;
-			TestThread.Abort();
-			AssertEquals("#31 Apartment State Changed when not needed: ",before,after);
 		}
 
+		[Test]
 		public void TestPriority1()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on mono on Win32. Our runtime should be fixed.");
+
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			try {
@@ -296,14 +364,14 @@ namespace MonoTests.System.Threading {
 				TestThread.Start();
 				TestUtil.WaitForAlive (TestThread, "wait7");
 				ThreadPriority before = TestThread.Priority;
-				AssertEquals("#41 Unexpected Priority Change: ",before,after);
-			}
-			finally {
+				Assert.AreEqual (before, after, "#41 Unexpected Priority Change: ");
+			} finally {
 				TestThread.Abort();
 			}
 		}
 
 		[Test]
+		[Category ("NotDotNet")] // on MS, Thread is still in AbortRequested state when Start is invoked
 		public void AbortUnstarted ()
 		{
 			C2Test test1 = new C2Test();
@@ -312,24 +380,26 @@ namespace MonoTests.System.Threading {
 			th.Start ();
 		}
 
-		[Category("NotWorking")] // this is a MonoTODO -> no support for Priority
+		[Test]
+		[Category ("NotDotNet")] // on MS, ThreadState is immediately Stopped after Abort
+		[Category ("NotWorking")] // this is a MonoTODO -> no support for Priority
 		public void TestPriority2()
 		{
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			try {
-				AssertEquals("#42 Incorrect Priority in New thread: ",ThreadPriority.Normal, TestThread.Priority);
+				Assert.AreEqual (ThreadPriority.Normal, TestThread.Priority, "#42 Incorrect Priority in New thread: ");
 				TestThread.Start();
 				TestUtil.WaitForAliveOrStop (TestThread, "wait8");
-				AssertEquals("#43 Incorrect Priority in Started thread: ",ThreadPriority.Normal, TestThread.Priority);
-			}
-			finally {
+				Assert.AreEqual (ThreadPriority.Normal, TestThread.Priority, "#43 Incorrect Priority in Started thread: ");
+			} finally {
 				TestThread.Abort();
 			}
-			AssertEquals("#44 Incorrect Priority in Aborted thread: ",ThreadPriority.Normal, TestThread.Priority);
+			Assert.AreEqual (ThreadPriority.Normal, TestThread.Priority, "#44 Incorrect Priority in Aborted thread: ");
 		}
 
-		[Category("NotWorking")] // this is a MonoTODO -> no support for Priority
+		[Test]
+		[Category ("NotWorking")] // this is a MonoTODO -> no support for Priority
 		public void TestPriority3()
 		{
 			C2Test test1 = new C2Test();
@@ -337,147 +407,142 @@ namespace MonoTests.System.Threading {
 			try {
 				TestThread.Start();
 				TestThread.Priority = ThreadPriority.Lowest;
-				AssertEquals("#45A Incorrect Priority:",ThreadPriority.Lowest,TestThread.Priority);
+				Assert.AreEqual (ThreadPriority.Lowest, TestThread.Priority, "#45A Incorrect Priority:");
 				TestThread.Priority = ThreadPriority.BelowNormal;
-				AssertEquals("#45B Incorrect Priority:",ThreadPriority.BelowNormal,TestThread.Priority);
+				Assert.AreEqual (ThreadPriority.BelowNormal, TestThread.Priority, "#45B Incorrect Priority:");
 				TestThread.Priority = ThreadPriority.Normal;
-				AssertEquals("#45C Incorrect Priority:",ThreadPriority.Normal,TestThread.Priority);
+				Assert.AreEqual (ThreadPriority.Normal, TestThread.Priority, "#45C Incorrect Priority:");
 				TestThread.Priority = ThreadPriority.AboveNormal;
-				AssertEquals("#45D Incorrect Priority:",ThreadPriority.AboveNormal,TestThread.Priority);
+				Assert.AreEqual (ThreadPriority.AboveNormal, TestThread.Priority, "#45D Incorrect Priority:");
 				TestThread.Priority = ThreadPriority.Highest;
-				AssertEquals("#45E Incorrect Priority:",ThreadPriority.Highest,TestThread.Priority);
+				Assert.AreEqual (ThreadPriority.Highest, TestThread.Priority, "#45E Incorrect Priority:");
 			}
 			finally {
 				TestThread.Abort();
 			}
 		}
 
-
-		public void TestIsBackground1()
+		[Test]
+		public void TestIsBackground1 ()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on mono on Win32. Our runtime should be fixed.");
+
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			try {
 				TestThread.Start();
 				TestUtil.WaitForAlive (TestThread, "wait9");
 				bool state = TestThread.IsBackground;
-				Assert("#51 IsBackground not set at the default state: ",!(state));
-			}
-			finally {
+				Assert.IsFalse (state, "#51 IsBackground not set at the default state: ");
+			} finally {
 				TestThread.Abort();
 			}
 		}
 
-		public void TestIsBackground2()
+		[Test]
+		[Category ("NotDotNet")] // on MS, ThreadState is immediately Stopped after Abort
+		public void TestIsBackground2 ()
 		{
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			TestThread.IsBackground = true;
 			try {
 				TestThread.Start();
-			}
-			finally {
+			} finally {
 				TestThread.Abort();
 			}
-			Assert("#52 Is Background Changed ot Start ",TestThread.IsBackground);
+			Assert.IsTrue (TestThread.IsBackground, "#52 Is Background Changed to Start ");
 		}
 
-
+		[Test]
 		public void TestName()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on mono on Win32. Our runtime should be fixed.");
+
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
 			try {
 				TestThread.Start();
 				TestUtil.WaitForAlive (TestThread, "wait10");
 				string name = TestThread.Name;
-				AssertEquals("#61 Name set when mustn't be set: ", name, (string)null);
+				Assert.IsNull (name, "#61 Name set when mustn't be set: ");
 				string newname = "Testing....";
 				TestThread.Name = newname;
-				AssertEquals("#62 Name not set when must be set: ",TestThread.Name,newname);
-			}
-			finally {
+				Assert.AreEqual (newname, TestThread.Name, "#62 Name not set when must be set: ");
+			} finally {
 				TestThread.Abort();
 			}
 		}
 
-		[Category("NotDotNet")]
+		[Test]
 		public void TestNestedThreads1()
 		{
-			C3Test  test1 = new C3Test();
+			C3Test test1 = new C3Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod1));
 			try {
 				TestThread.Start();
 				TestUtil.WaitForAlive (TestThread, "wait11");
-			}
-			catch(Exception e) {
-				Fail("#71 Unexpected Exception" + e.Message);
-			}
-			finally {
+			} finally {
 				TestThread.Abort();
 			}
 		}
 
+		[Test]
 		public void TestNestedThreads2()
 		{
 			C4Test test1 = new C4Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod1));
 			try {
 				TestThread.Start();
-			}
-			catch(Exception e) {
-				Fail("#81 Unexpected Exception" + e.ToString());
-			}
-			finally {
+			} finally {
 				TestThread.Abort();
 			}
 		}
 
-		
+		[Test]
 		public void TestJoin1()
 		{
 			C1Test test1 = new C1Test();
 			C1Test test2 = new C1Test();
 			Thread thread1 = new Thread(new ThreadStart(test1.TestMethod));
 			Thread thread2 = new Thread(new ThreadStart(test1.TestMethod2));
-			try
-			{
+			try {
 				thread1.Start();
 				thread2.Start();
 				thread2.Join();
-			}
-			catch(Exception e)
-			{
-				Fail("#91 Unexpected Exception " + e.ToString());
-			}
-			finally
-			{
+			} finally {
 				thread1.Abort();
 				thread2.Abort();
 			}
 		}
-		
-		public void TestThreadState()
+
+		[Test]
+		public void TestThreadState ()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on mono on Win32. Our runtime should be fixed.");
+
 			//TODO: Test The rest of the possible transitions
 			C2Test test1 = new C2Test();
 			Thread TestThread = new Thread(new ThreadStart(test1.TestMethod));
-			AssertEquals("#101 Wrong Thread State",ThreadState.Unstarted,TestThread.ThreadState);
+			Assert.AreEqual (ThreadState.Unstarted, TestThread.ThreadState, "#101 Wrong Thread State");
 			try {
 				TestThread.Start();
 				//while(!TestThread.IsAlive); //In the MS Documentation this is not necessary
 											  //but in the MS SDK it is
-				Assert("#102 Wrong Thread State: " + TestThread.ThreadState.ToString(), TestThread.ThreadState == ThreadState.Running || (TestThread.ThreadState & ThreadState.Unstarted) != 0);
-			}
-			finally {
+				Assert.IsTrue (TestThread.ThreadState == ThreadState.Running || (TestThread.ThreadState & ThreadState.Unstarted) != 0,
+					"#102 Wrong Thread State: " + TestThread.ThreadState.ToString ());
+			} finally {
 				TestThread.Abort();
 			}
 			
 			TestUtil.WaitForNotAlive (TestThread, "wait12");
 			// Docs say state will be Stopped, but Aborted happens sometimes (?)
-			Assert("#103 Wrong Thread State: " + TestThread.ThreadState.ToString(), (ThreadState.Stopped & TestThread.ThreadState) != 0 
-				|| (ThreadState.Aborted & TestThread.ThreadState) != 0);
-		} 
+			Assert.IsTrue ((ThreadState.Stopped & TestThread.ThreadState) != 0 || (ThreadState.Aborted & TestThread.ThreadState) != 0,
+				"#103 Wrong Thread State: " + TestThread.ThreadState.ToString ());
+		}
 
 		[Test]
 		[Ignore ("see comment below.")]
@@ -489,8 +554,7 @@ namespace MonoTests.System.Threading {
 			try {
 				t.Start ();
 				t.Join ();
-			}
-			catch {
+			} catch {
 				t.Abort ();
 			}
 		}
@@ -506,8 +570,7 @@ namespace MonoTests.System.Threading {
 			try {
 				t.Start ();
 				t.Join ();
-			}
-			catch {
+			} catch {
 				t.Abort ();
 			}
 		}
@@ -521,8 +584,7 @@ namespace MonoTests.System.Threading {
 			try {
 				t.Start ();
 				t.Join ();
-			}
-			catch {
+			} catch {
 				t.Abort ();
 			}
 		}
@@ -537,15 +599,14 @@ namespace MonoTests.System.Threading {
 				Thread.CurrentPrincipal = new GenericPrincipal (new GenericIdentity ("good"), null);
 				t.Start ();
 				t.Join ();
-			}
-			catch {
+			} catch {
 				t.Abort ();
 			}
 		}
 
 		int counter = 0;
-		
-		[Category("NotDotNet")]
+
+		[Test]
 		public void TestSuspend ()
 		{
 			Thread t = new Thread (new ThreadStart (DoCount));
@@ -568,11 +629,14 @@ namespace MonoTests.System.Threading {
 			TestUtil.WaitForNotAlive (t, "wait13");
 			CheckIsNotRunning ("t6", t);
 		}
-		
-		[Category("NotDotNet")]
-		[Category("NotWorking")]
+
+		[Test]
+		[Category("NotDotNet")] // On MS, ThreadStateException is thrown on Abort: "Thread is suspended; attempting to abort"
 		public void TestSuspendAbort ()
 		{
+			if (is_win32 && is_mono)
+				Assert.Fail ("This test fails on Win32. The test should be fixed.");
+
 			Thread t = new Thread (new ThreadStart (DoCount));
 			t.IsBackground = true;
 			t.Start ();
@@ -591,24 +655,76 @@ namespace MonoTests.System.Threading {
 				Thread.Sleep (10);
 				n++;
 			}
-			
-			Assert ("Timeout while waiting for abort", n < 200);
+
+			Assert.IsTrue (n < 200, "Timeout while waiting for abort");
 			
 			CheckIsNotRunning ("t6", t);
-		}		
+		}
+
+		[Test]
+		public void Test_Interrupt ()
+		{
+			bool interruptedExceptionThrown = false;
+			ThreadPool.QueueUserWorkItem (Test_Interrupt_Worker, Thread.CurrentThread);
+
+			try {
+				try {
+					Thread.Sleep (3000);
+				} finally {
+					try {
+						Thread.Sleep (0);
+					} catch (ThreadInterruptedException) {
+						Assert.Fail ("ThreadInterruptedException thrown twice");
+					}
+				}
+			} catch (ThreadInterruptedException) {
+				interruptedExceptionThrown = true;
+			}
+
+			Assert.IsTrue (interruptedExceptionThrown, "ThreadInterruptedException expected.");
+		}
+
+		private void Test_Interrupt_Worker (object o)
+		{
+			Thread t = o as Thread;
+			Thread.Sleep (100);
+			t.Interrupt ();
+		}
 		
+		[Test]
+		public void Test_InterruptCurrentThread ()
+		{
+			bool interruptedExceptionThrown = false;
+
+			try {
+				try {
+					Thread.CurrentThread.Interrupt ();
+				} finally {
+					try {
+						Thread.Sleep (0);
+					} catch (ThreadInterruptedException) {
+						Assert.Fail ("ThreadInterruptedException should not be thrown.");
+					}
+				}
+			} catch (ThreadInterruptedException) {
+				interruptedExceptionThrown = true;
+			}
+
+			Assert.IsFalse (interruptedExceptionThrown, "ThreadInterruptedException should not be thrown.");
+		}
+
 		void CheckIsRunning (string s, Thread t)
 		{
 			int c = counter;
 			Thread.Sleep (100);
-			Assert (s, counter > c);
+			Assert.IsTrue (counter > c, s);
 		}
 		
 		void CheckIsNotRunning (string s, Thread t)
 		{
 			int c = counter;
 			Thread.Sleep (100);
-			Assert (s, counter == c);
+			Assert.AreEqual (counter, c, s);
 		}
 		
 		void WaitSuspended (string s, Thread t)
@@ -616,13 +732,13 @@ namespace MonoTests.System.Threading {
 			int n=0;
 			ThreadState state = t.ThreadState;
 			while ((state & ThreadState.Suspended) == 0) {
-				Assert (s + ": expected SuspendRequested state", (state & ThreadState.SuspendRequested) != 0);
+				Assert.IsTrue ((state & ThreadState.SuspendRequested) != 0, s + ": expected SuspendRequested state");
 				Thread.Sleep (10);
 				n++;
-				Assert (s + ": failed to suspend", n < 100);
+				Assert.IsTrue (n < 100, s + ": failed to suspend");
 				state = t.ThreadState;
 			}
-			Assert (s + ": SuspendRequested state not expected", (state & ThreadState.SuspendRequested) == 0);
+			Assert.IsTrue ((state & ThreadState.SuspendRequested) == 0, s + ": SuspendRequested state not expected");
 		}
 		
 		void WaitResumed (string s, Thread t)
@@ -631,7 +747,7 @@ namespace MonoTests.System.Threading {
 			while ((t.ThreadState & ThreadState.Suspended) != 0) {
 				Thread.Sleep (10);
 				n++;
-				Assert (s + ": failed to resume", n < 100);
+				Assert.IsTrue (n < 100, s + ": failed to resume");
 			}
 		}
 		
@@ -643,7 +759,145 @@ namespace MonoTests.System.Threading {
 			}
 		}
 	}
-	
+
+	[TestFixture]
+	public class ThreadStateTest {
+		void Start ()
+		{
+		}
+
+		[Test] // bug #81720
+		public void IsBackGround ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			Assert.AreEqual (ThreadState.Unstarted, t1.ThreadState, "#A1");
+			Assert.IsFalse (t1.IsBackground, "#A2");
+			t1.Start ();
+			t1.Join ();
+			Assert.AreEqual (ThreadState.Stopped, t1.ThreadState, "#A3");
+
+			try {
+				bool isBackGround = t1.IsBackground;
+				Assert.Fail ("#A4: " + isBackGround.ToString ());
+			} catch (ThreadStateException ex) {
+				Assert.AreEqual (typeof (ThreadStateException), ex.GetType (), "#A5");
+				Assert.IsNull (ex.InnerException, "#A6");
+				Assert.IsNotNull (ex.Message, "#A7");
+			}
+
+			Thread t2 = new Thread (new ThreadStart (Start));
+			Assert.AreEqual (ThreadState.Unstarted, t2.ThreadState, "#B1");
+			t2.IsBackground = true;
+			Assert.AreEqual (ThreadState.Unstarted | ThreadState.Background, t2.ThreadState, "#B2");
+			Assert.IsTrue (t2.IsBackground, "#B3");
+			t2.Start ();
+			t2.Join ();
+			Assert.AreEqual (ThreadState.Stopped, t2.ThreadState, "#B4");
+
+			try {
+				bool isBackGround = t2.IsBackground;
+				Assert.Fail ("#B5: " + isBackGround.ToString ());
+			} catch (ThreadStateException ex) {
+				Assert.AreEqual (typeof (ThreadStateException), ex.GetType (), "#B6");
+				Assert.IsNull (ex.InnerException, "#B7");
+				Assert.IsNotNull (ex.Message, "#B8");
+			}
+		}
+	}
+
+	[TestFixture]
+	public class ThreadApartmentTest
+	{
+		void Start ()
+		{
+		}
+
+		[Test] // bug #81658
+		public void ApartmentState_StoppedThread ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			t1.Start ();
+			t1.Join ();
+			try {
+				ApartmentState state = t1.ApartmentState;
+				Assert.Fail ("#A1: " + state.ToString ());
+			} catch (ThreadStateException ex) {
+				Assert.AreEqual (typeof (ThreadStateException), ex.GetType (), "#A2");
+				Assert.IsNull (ex.InnerException, "#A3");
+				Assert.IsNotNull (ex.Message, "#A4");
+			}
+
+			Thread t2 = new Thread (new ThreadStart (Start));
+			t2.IsBackground = true;
+			t2.Start ();
+			t2.Join ();
+			try {
+				ApartmentState state = t2.ApartmentState;
+				Assert.Fail ("#B1: " + state.ToString ());
+			} catch (ThreadStateException ex) {
+				Assert.AreEqual (typeof (ThreadStateException), ex.GetType (), "#B2");
+				Assert.IsNull (ex.InnerException, "#B3");
+				Assert.IsNotNull (ex.Message, "#B4");
+			}
+		}
+
+		[Test]
+		public void ApartmentState_BackGround ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			t1.IsBackground = true;
+			Assert.AreEqual (ApartmentState.Unknown, t1.ApartmentState, "#1");
+			t1.ApartmentState = ApartmentState.STA;
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "#2");
+		}
+
+		[Test]
+		public void TestApartmentState ()
+		{
+			Thread t1 = new Thread (new ThreadStart (Start));
+			Thread t2 = new Thread (new ThreadStart (Start));
+			Thread t3 = new Thread (new ThreadStart (Start));
+
+			Assert.AreEqual (ApartmentState.Unknown, t1.ApartmentState, "Thread1 Default");
+			Assert.AreEqual (ApartmentState.Unknown, t2.ApartmentState, "Thread2 Default");
+			Assert.AreEqual (ApartmentState.Unknown, t3.ApartmentState, "Thread3 Default");
+
+			t1.ApartmentState = ApartmentState.STA;
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "Thread1 Set Once");
+			t1.ApartmentState = ApartmentState.MTA;
+			Assert.AreEqual (ApartmentState.STA, t1.ApartmentState, "Thread1 Set Twice");
+
+			t2.ApartmentState = ApartmentState.MTA;
+			Assert.AreEqual (ApartmentState.MTA, t2.ApartmentState, "Thread2 Set Once");
+			t2.ApartmentState = ApartmentState.STA;
+			Assert.AreEqual (ApartmentState.MTA, t2.ApartmentState, "Thread2 Set Twice");
+
+			bool exception_occured = false;
+			try {
+				t3.ApartmentState = ApartmentState.Unknown;
+			}
+			catch (Exception) {
+				exception_occured = true;
+			}
+			Assert.AreEqual (ApartmentState.Unknown, t3.ApartmentState, "Thread3 Set Invalid");
+#if NET_2_0
+			Assert.IsFalse (exception_occured, "Thread3 Set Invalid Exception Occured");
+#else
+			Assert.IsTrue (exception_occured, "Thread3 Set Invalid Exception Occured");
+#endif
+
+			t1.Start ();
+			exception_occured = false;
+			try {
+				t1.ApartmentState = ApartmentState.STA;
+			}
+			catch (Exception) {
+				exception_occured = true;
+			}
+			Assert.IsTrue (exception_occured, "Thread1 Started Invalid Exception Occured");
+		}
+	}
+
 	public class TestUtil
 	{
 		public static void WaitForNotAlive (Thread t, string s)
@@ -666,8 +920,8 @@ namespace MonoTests.System.Threading {
 			DateTime ti = DateTime.Now;
 			while (t.IsAlive == alive) {
 				if ((DateTime.Now - ti).TotalSeconds > 10) {
-					if (alive) Assertion.Fail ("Timeout while waiting for not alive state. " + s);
-					else Assertion.Fail ("Timeout while waiting for alive state. " + s);
+					if (alive) Assert.Fail ("Timeout while waiting for not alive state. " + s);
+					else Assert.Fail ("Timeout while waiting for alive state. " + s);
 				}
 			}
 		}
@@ -680,8 +934,8 @@ namespace MonoTests.System.Threading {
 					return false;
 
 				if ((DateTime.Now - ti).TotalSeconds > 10) {
-					if (alive) Assertion.Fail ("Timeout while waiting for not alive state. " + s);
-					else Assertion.Fail ("Timeout while waiting for alive state. " + s);
+					if (alive) Assert.Fail ("Timeout while waiting for not alive state. " + s);
+					else Assert.Fail ("Timeout while waiting for alive state. " + s);
 				}
 			}
 
