@@ -22,8 +22,10 @@ namespace Mono.AssemblyCompare
 	{
 		static int Main (string [] args)
 		{
-			if (args.Length != 2)
+			if (args.Length != 2) {
+				Console.WriteLine ("Usage: mono mono-api-diff.exe <assembly 1 xml> <assembly 2 xml>");
 				return 1;
+			}
 
 			XMLAssembly ms = CreateXMLAssembly (args [0]);
 			XMLAssembly mono = CreateXMLAssembly (args [1]);
@@ -178,6 +180,29 @@ namespace Mono.AssemblyCompare
 			return (object []) list.ToArray (type);
 		}
 
+		public static bool IsMeaninglessAttribute (string s)
+		{
+			if (s == null)
+				return false;
+			if (s == "System.Runtime.CompilerServices.CompilerGeneratedAttribute")
+				return true;
+			return false;
+		}
+
+		public static bool IsMonoTODOAttribute (string s)
+		{
+			if (s == null)
+				return false;
+			if (//s.EndsWith ("MonoTODOAttribute") ||
+			    s.EndsWith ("MonoDocumentationNoteAttribute") ||
+			    s.EndsWith ("MonoExtensionAttribute") ||
+//			    s.EndsWith ("MonoInternalNoteAttribute") ||
+			    s.EndsWith ("MonoLimitationAttribute") ||
+			    s.EndsWith ("MonoNotSupportedAttribute"))
+				return true;
+			return s.EndsWith ("TODOAttribute");
+		}
+
 		protected void AddAttribute (XmlNode node, string name, string value)
 		{
 			XmlAttribute attr = document.CreateAttribute (name);
@@ -449,19 +474,17 @@ namespace Mono.AssemblyCompare
 				newNS.Add (node);
 				AddAttribute (node, "name", xns.Name);
 
-				if (oh.ContainsKey (xns.Name)) {
-					int idx = (int) oh [xns.Name];
-					xns.CompareTo (document, node, other [idx]);
+				int idx = -1;
+				if (oh.ContainsKey (xns.Name))
+					idx = (int) oh [xns.Name];
+				XMLNamespace ons = idx >= 0 ? (XMLNamespace) other [idx] : null;
+				xns.CompareTo (document, node, ons);
+				if (idx >= 0)
 					other [idx] = null;
-					xns.AddCountersAttributes (node);
-					counters.Present++;
-					counters.PresentTotal++;
-					counters.AddPartialToTotal (xns.Counters);
-				} else {
-					AddAttribute (node, "presence", "missing");
-					counters.Missing++;
-					counters.MissingTotal++;
-				}
+				xns.AddCountersAttributes (node);
+				counters.Present++;
+				counters.PresentTotal++;
+				counters.AddPartialToTotal (xns.Counters);
 			}
 
 			if (other != null) {
@@ -548,7 +571,7 @@ namespace Mono.AssemblyCompare
 			XmlNode childA = doc.CreateElement ("classes", null);
 			parent.AppendChild (childA);
 
-			CompareTypes (childA, nspace.types);
+			CompareTypes (childA, nspace != null ? nspace.types : new XMLClass [0]);
 		}
 
 		void CompareTypes (XmlNode parent, XMLClass [] other)
@@ -565,23 +588,20 @@ namespace Mono.AssemblyCompare
 				AddAttribute (node, "name", xclass.Name);
 				AddAttribute (node, "type", xclass.Type);
 
-				if (oh.ContainsKey (xclass.Name)) {
-					int idx = (int) oh [xclass.Name];
-					xclass.CompareTo (document, node, other [idx]);
+				int idx = -1;
+				if (oh.ContainsKey (xclass.Name))
+					idx = (int) oh [xclass.Name];
+				xclass.CompareTo (document, node, idx >= 0 ? other [idx] : new XMLClass ());
+				if (idx >= 0)
 					other [idx] = null;
-					counters.AddPartialToPartial (xclass.Counters);
-				} else {
-					AddAttribute (node, "presence", "missing");
-					counters.Missing++;
-					counters.MissingTotal++;
-				}
+				counters.AddPartialToPartial (xclass.Counters);
 			}
 
 			if (other != null) {
 				count = other.Length;
 				for (int i = 0; i < count; i++) {
 					XMLClass c = other [i];
-					if (c == null || c.Name.EndsWith ("TODOAttribute"))
+					if (c == null || IsMonoTODOAttribute (c.Name))
 						continue;
 
 					node = document.CreateElement ("class", null);
@@ -850,7 +870,7 @@ namespace Mono.AssemblyCompare
 			for (int i = 0; i < count; i++) {
 				XMLClass xclass = nested [i];
 
-				node = document.CreateElement ("nestedclass", null);
+				node = document.CreateElement ("class", null);
 				newNodes.Add (node);
 				AddAttribute (node, "name", xclass.Name);
 				AddAttribute (node, "type", xclass.Type);
@@ -872,12 +892,13 @@ namespace Mono.AssemblyCompare
 				count = other.Length;
 				for (int i = 0; i < count; i++) {
 					XMLClass c = other [i];
-					if (c == null || c.Name.EndsWith ("TODOAttribute"))
+					if (c == null || IsMonoTODOAttribute (c.Name))
 						continue;
 
-					node = document.CreateElement ("nestedclass", null);
+					node = document.CreateElement ("class", null);
 					newNodes.Add (node);
 					AddAttribute (node, "name", c.Name);
+					AddAttribute (node, "type", c.Type);
 					AddExtra (node);
 					counters.Extra++;
 					counters.ExtraTotal++;
@@ -960,6 +981,9 @@ namespace Mono.AssemblyCompare
 
 			XMLParameter oparm = (XMLParameter) other;
 
+			if (name != oparm.name)
+				AddWarning (parent, "Parameter name is wrong: {0} != {1}", name, oparm.name);
+
 			if (type != oparm.type)
 				AddWarning (parent, "Parameter type is wrong: {0} != {1}", type, oparm.type);
 			
@@ -1095,17 +1119,17 @@ namespace Mono.AssemblyCompare
 
 		protected override bool CheckIfAdd (string value, XmlNode node)
 		{
-			if (value.EndsWith ("TODOAttribute")) {
+			if (IsMonoTODOAttribute (value)) {
 				isTodo = true;
 
 				XmlNode pNode = node.SelectSingleNode ("properties");
-				if (pNode.ChildNodes [0].Attributes ["value"] != null) {
+				if (pNode != null && pNode.ChildNodes.Count > 0 && pNode.ChildNodes [0].Attributes ["value"] != null) {
 					comment = pNode.ChildNodes [0].Attributes ["value"].Value;
 				}
 				return false;
 			}
 
-			return true;
+			return !IsMeaninglessAttribute (value);
 		}
 
 		protected override void CompareToInner (string name, XmlNode node, XMLNameGroup other)
@@ -1166,7 +1190,7 @@ namespace Mono.AssemblyCompare
 		{
 			XmlNode pNode = node.SelectSingleNode ("properties");
 
-			if (name.EndsWith ("TODOAttribute")) {
+			if (IsMonoTODOAttribute (name)) {
 				isTodo = true;
 				if (pNode.ChildNodes [0].Attributes ["value"] != null) {
 					comment = pNode.ChildNodes [0].Attributes ["value"].Value;
@@ -1551,6 +1575,7 @@ namespace Mono.AssemblyCompare
 	class XMLEvents : XMLMember
 	{
 		Hashtable eventTypes;
+		Hashtable nameToMethod = new Hashtable ();
 
 		protected override void LoadExtraData (string name, XmlNode node)
 		{
@@ -1562,6 +1587,19 @@ namespace Mono.AssemblyCompare
 				eventTypes [name] = xatt.Value;
 			}
 
+			XmlNode child = node.FirstChild;
+			while (child != null) {
+				if (child != null && child.Name == "methods") {
+					XMLMethods m = new XMLMethods ();
+					XmlNode parent = child.ParentNode;
+					string key = GetNodeKey (name, parent);
+					m.LoadData (child);
+					nameToMethod [key] = m;
+					break;
+				}
+				child = child.NextSibling;
+			}
+
 			base.LoadExtraData (name, node);
 		}
 
@@ -1584,6 +1622,16 @@ namespace Mono.AssemblyCompare
 
 				if (etype != oetype)
 					AddWarning (parent, "Event type is {0} and should be {1}", oetype, etype);
+
+				XMLMethods m = nameToMethod [name] as XMLMethods;
+				XMLMethods om = evt.nameToMethod [name] as XMLMethods;
+				if (m != null || om != null) {
+					if (m == null)
+						m = new XMLMethods ();
+
+					m.CompareTo (document, parent, om);
+					counters.AddPartialToPartial (m.Counters);
+				}
 			} finally {
 				AddCountersAttributes (parent);
 				copy.AddPartialToPartial (counters);
@@ -1671,6 +1719,23 @@ namespace Mono.AssemblyCompare
 			base.LoadExtraData (name, node);
 		}
 
+		public override string GetNodeKey (string name, XmlNode node)
+		{
+			// for explicit/implicit operators we need to include the return
+			// type in the key to allow matching; as a side-effect, differences
+			// in return types will be reported as extra/missing methods
+			//
+			// for regular methods we do not need to take into account the
+			// return type for matching methods; differences in return types
+			// will be reported as a warning on the method
+			if (name.StartsWith ("op_")) {
+				XmlAttribute xatt = node.Attributes ["returntype"];
+				string returnType = xatt != null ? xatt.Value + " " : string.Empty;
+				return returnType + name;
+			}
+			return name;
+		}
+
 		protected override void CompareToInner (string name, XmlNode parent, XMLNameGroup other)
 		{
 			// create backup of actual counters
