@@ -15,20 +15,23 @@
 #include <stdlib.h>
 #include "assembly.h"
 #include "image.h"
-#include "cil-coff.h"
 #include "rawbuffer.h"
+#include "object-internals.h"
 #include <mono/metadata/loader.h>
 #include <mono/metadata/tabledefs.h>
 #include <mono/metadata/metadata-internals.h>
+#include <mono/metadata/profiler-private.h>
 #include <mono/metadata/class-internals.h>
 #include <mono/metadata/domain-internals.h>
 #include <mono/metadata/mono-endian.h>
+#include <mono/metadata/mono-debug.h>
 #include <mono/io-layer/io-layer.h>
 #include <mono/utils/mono-uri.h>
 #include <mono/metadata/mono-config.h>
 #include <mono/utils/mono-digest.h>
 #include <mono/utils/mono-logger.h>
 #include <mono/metadata/reflection.h>
+#include <mono/metadata/coree.h>
 
 #ifndef PLATFORM_WIN32
 #include <sys/types.h>
@@ -72,7 +75,6 @@ static const AssemblyVersionMap framework_assemblies [] = {
 	{"Mono.Cairo", 0},
 	{"Mono.CompilerServices.SymbolWriter", 0},
 	{"Mono.Data", 0},
-	{"Mono.Data.SqliteClient", 0},
 	{"Mono.Data.SybaseClient", 0},
 	{"Mono.Data.Tds", 0},
 	{"Mono.Data.TdsClient", 0},
@@ -120,12 +122,6 @@ static MonoAssembly *corlib;
 #define mono_assemblies_unlock() LeaveCriticalSection (&assemblies_mutex)
 static CRITICAL_SECTION assemblies_mutex;
 
-/* A hastable of thread->assembly list mappings */
-static GHashTable *assemblies_loading;
-
-/* A hashtable of reflection only load thread->assemblies mappings */
-static GHashTable *assemblies_refonly_loading;
-
 /* If defined, points to the bundled assembly information */
 const MonoBundledAssembly **bundles;
 
@@ -151,6 +147,20 @@ encode_public_tok (const guchar *token, gint32 len)
 	return res;
 }
 
+/**
+ * mono_public_tokens_are_equal:
+ * @pubt1: first public key token
+ * @pubt2: second public key token
+ *
+ * Compare two public key tokens and return #TRUE is they are equal and #FALSE
+ * otherwise.
+ */
+gboolean
+mono_public_tokens_are_equal (const unsigned char *pubt1, const unsigned char *pubt2)
+{
+	return g_strcasecmp ((char*)pubt1, (char*)pubt2) == 0;
+}
+
 static void
 check_path_env (void)
 {
@@ -229,7 +239,7 @@ assembly_binding_maps_name (MonoAssemblyBindingInfo *info, MonoAssemblyName *ana
 	if (info->culture && strcmp (info->culture, aname->culture))
 		return FALSE;
 	
-	if (strcmp ((const char *)info->public_key_token, (const char *)aname->public_key_token))
+	if (!mono_public_tokens_are_equal (info->public_key_token, aname->public_key_token))
 		return FALSE;
 
 	return TRUE;
@@ -371,39 +381,12 @@ mono_assembly_names_equal (MonoAssemblyName *l, MonoAssemblyName *r)
 	if (!l->public_key_token [0] || !r->public_key_token [0])
 		return TRUE;
 
-	if (strcmp ((char*)l->public_key_token, (char*)r->public_key_token))
+	if (!mono_public_tokens_are_equal (l->public_key_token, r->public_key_token))
 		return FALSE;
 
 	return TRUE;
 }
 
-static MonoAssembly*
-search_loaded (MonoAssemblyName* aname, gboolean refonly)
-{
-	GList *tmp;
-	MonoAssembly *ass;
-	GList *loading;
-
-	ass = mono_assembly_invoke_search_hook_internal (aname, refonly, FALSE);
-	if (ass)
-		return ass;
-	
-	/*
-	 * The assembly might be under load by this thread. In this case, it is
-	 * safe to return an incomplete instance to prevent loops.
-	 */
-	loading = g_hash_table_lookup (refonly ? assemblies_refonly_loading : assemblies_loading, (gpointer)GetCurrentThreadId ());
-	for (tmp = loading; tmp; tmp = tmp->next) {
-		ass = tmp->data;
-		if (!mono_assembly_names_equal (aname, &ass->aname))
-			continue;
-
-		return ass;
-	}
-
-	return NULL;
-}
-
 static MonoAssembly *
 load_in_path (const char *basename, const char** search_path, MonoImageOpenStatus *status, MonoBoolean refonly)
 {
@@ -556,24 +539,6 @@ set_dirs (char *exe)
 
 #endif /* PLATFORM_WIN32 */
 
-#ifdef UNDER_CE
-#undef GetModuleFileName
-#define GetModuleFileName ceGetModuleFileNameA
-
-DWORD ceGetModuleFileNameA(HMODULE hModule, char* lpFilename, DWORD nSize)
-{
-	DWORD res = 0;
-	wchar_t* wbuff = (wchar_t*)LocalAlloc(LPTR, nSize*2);
-	res = GetModuleFileNameW(hModule, wbuff, nSize);
-	if (res) {
-		int len = wcslen(wbuff);
-		WideCharToMultiByte(CP_ACP, 0, wbuff, len, lpFilename, len, NULL, NULL);
-	}
-	LocalFree(wbuff);
-	return res;
-}
-#endif
-
 /**
  * mono_set_rootdir:
  *
@@ -584,12 +549,10 @@ void
 mono_set_rootdir (void)
 {
 #ifdef PLATFORM_WIN32
-	gunichar2 moddir [MAXPATHLEN];
-	gchar *bindir, *installdir, *root, *utf8name, *config;
+	gchar *bindir, *installdir, *root, *name, *config;
 
-	GetModuleFileNameW (NULL, moddir, MAXPATHLEN);
-	utf8name = g_utf16_to_utf8 (moddir, -1, NULL, NULL, NULL);
-	bindir = g_path_get_dirname (utf8name);
+	name = mono_get_module_file_name ((HMODULE) &__ImageBase);
+	bindir = g_path_get_dirname (name);
 	installdir = g_path_get_dirname (bindir);
 	root = g_build_path (G_DIR_SEPARATOR_S, installdir, "lib", NULL);
 
@@ -600,7 +563,7 @@ mono_set_rootdir (void)
 	g_free (root);
 	g_free (installdir);
 	g_free (bindir);
-	g_free (utf8name);
+	g_free (name);
 #else
 	char buf [4096];
 	int  s;
@@ -647,9 +610,6 @@ mono_assemblies_init (void)
 	check_extra_gac_path_env ();
 
 	InitializeCriticalSection (&assemblies_mutex);
-
-	assemblies_loading = g_hash_table_new (NULL, NULL);
-	assemblies_refonly_loading = g_hash_table_new (NULL, NULL);
 }
 
 gboolean
@@ -718,12 +678,12 @@ char*
 mono_stringify_assembly_name (MonoAssemblyName *aname)
 {
 	return g_strdup_printf (
-		"%s, Version=%d.%d.%d.%d, Culture=%s%s%s",
+		"%s, Version=%d.%d.%d.%d, Culture=%s, PublicKeyToken=%s%s",
 		aname->name,
 		aname->major, aname->minor, aname->build, aname->revision,
 		aname->culture && *aname->culture? aname->culture: "neutral",
-		aname->public_key_token [0] ? ", PublicKeyToken=" : "",
-		aname->public_key_token [0] ? (char *)aname->public_key_token : "");
+		aname->public_key_token [0] ? (char *)aname->public_key_token : "null",
+		(aname->flags & ASSEMBLYREF_RETARGETABLE_FLAG) ? ", Retargetable=Yes" : "");
 }
 
 static gchar*
@@ -855,6 +815,11 @@ mono_assembly_load_reference (MonoImage *image, int index)
 	 * it inside a critical section.
 	 */
 	mono_assemblies_lock ();
+	if (!image->references) {
+		MonoTableInfo *t = &image->tables [MONO_TABLE_ASSEMBLYREF];
+	
+		image->references = g_new0 (MonoAssembly *, t->rows + 1);
+	}
 	reference = image->references [index];
 	mono_assemblies_unlock ();
 	if (reference)
@@ -866,8 +831,13 @@ mono_assembly_load_reference (MonoImage *image, int index)
 		/* We use the loaded corlib */
 		if (!strcmp (aname.name, "mscorlib"))
 			reference = mono_assembly_load_full (&aname, image->assembly->basedir, &status, FALSE);
-		else
+		else {
 			reference = mono_assembly_loaded_full (&aname, TRUE);
+			if (!reference)
+				/* Try a postload search hook */
+				reference = mono_assembly_invoke_search_hook_internal (&aname, TRUE, TRUE);
+		}
+
 		/*
 		 * Here we must advice that the error was due to
 		 * a non loaded reference using the ReflectionOnly api
@@ -881,7 +851,7 @@ mono_assembly_load_reference (MonoImage *image, int index)
 		char *extra_msg = g_strdup ("");
 
 		if (status == MONO_IMAGE_ERROR_ERRNO && errno == ENOENT) {
-			extra_msg = g_strdup_printf ("The assembly was not found in the Global Assembly Cache, a path listed in the MONO_PATH environment variable, or in the location of the executing assembly (%s).\n", image->assembly->basedir);
+			extra_msg = g_strdup_printf ("The assembly was not found in the Global Assembly Cache, a path listed in the MONO_PATH environment variable, or in the location of the executing assembly (%s).\n", image->assembly != NULL ? image->assembly->basedir : "" );
 		} else if (status == MONO_IMAGE_ERROR_ERRNO) {
 			extra_msg = g_strdup_printf ("System error: %s\n", strerror (errno));
 		} else if (status == MONO_IMAGE_MISSING_ASSEMBLYREF) {
@@ -931,22 +901,8 @@ mono_assembly_load_reference (MonoImage *image, int index)
 void
 mono_assembly_load_references (MonoImage *image, MonoImageOpenStatus *status)
 {
-	MonoTableInfo *t;
-	int i;
-
+	/* This is a no-op now but it is part of the embedding API so we can't remove it */
 	*status = MONO_IMAGE_OK;
-
-	t = &image->tables [MONO_TABLE_ASSEMBLYREF];
-	
-	image->references = g_new0 (MonoAssembly *, t->rows + 1);
-
-	/* resolve assembly references for modules */
-	for (i = 0; i < image->module_count; i++){
-		if (image->modules [i]) {
-			image->modules [i]->assembly = image->assembly;
-			mono_assembly_load_references (image->modules [i], status);
-		}
-	}
 }
 
 typedef struct AssemblyLoadHook AssemblyLoadHook;
@@ -1266,6 +1222,7 @@ mono_assembly_open_full (const char *filename, MonoImageOpenStatus *status, gboo
 	MonoAssembly *ass;
 	MonoImageOpenStatus def_status;
 	gchar *fname;
+	gchar *new_fname;
 	
 	g_return_val_if_fail (filename != NULL, NULL);
 
@@ -1303,7 +1260,15 @@ mono_assembly_open_full (const char *filename, MonoImageOpenStatus *status, gboo
 	}
 
 	mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_ASSEMBLY,
-			"Assembly Loader probing location: '%s'.", filename);
+			"Assembly Loader probing location: '%s'.", fname);
+	new_fname = mono_make_shadow_copy (fname);
+	if (new_fname && new_fname != fname) {
+		g_free (fname);
+		fname = new_fname;
+		mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_ASSEMBLY,
+			    "Assembly Loader shadow-copied assembly to: '%s'.", fname);
+	}
+	
 	image = NULL;
 
 	if (bundles != NULL)
@@ -1348,20 +1313,44 @@ mono_assembly_open_full (const char *filename, MonoImageOpenStatus *status, gboo
 }
 
 /*
- * load_friend_assemblies:
+ * mono_load_friend_assemblies:
  * @ass: an assembly
  *
  * Load the list of friend assemblies that are allowed to access
  * the assembly's internal types and members. They are stored as assembly
  * names in custom attributes.
+ *
+ * This is an internal method, we need this because when we load mscorlib
+ * we do not have the mono_defaults.internals_visible_class loaded yet,
+ * so we need to load these after we initialize the runtime. 
+ *
+ * LOCKING: Acquires the assemblies lock plus the loader lock.
  */
-static void
-load_friend_assemblies (MonoAssembly* ass)
+void
+mono_assembly_load_friends (MonoAssembly* ass)
 {
 	int i;
-	MonoCustomAttrInfo* attrs = mono_custom_attrs_from_assembly (ass);
-	if (!attrs)
+	MonoCustomAttrInfo* attrs;
+
+	if (ass->friend_assembly_names_inited)
 		return;
+
+	attrs = mono_custom_attrs_from_assembly (ass);
+	if (!attrs) {
+		ass->friend_assembly_names_inited = TRUE;
+		return;
+	}
+
+	mono_assemblies_lock ();
+	if (ass->friend_assembly_names_inited) {
+		mono_assemblies_unlock ();
+		return;
+	}
+
+	/* 
+	 * assemblies_lock () is a low level lock so the code below should not take 
+	 * any locks.
+	 */
 	for (i = 0; i < attrs->num_attrs; ++i) {
 		MonoCustomAttrEntry *attr = &attrs->attrs [i];
 		MonoAssemblyName *aname;
@@ -1379,13 +1368,18 @@ load_friend_assemblies (MonoAssembly* ass)
 		slen = mono_metadata_decode_value (data + 2, &data);
 		aname = g_new0 (MonoAssemblyName, 1);
 		/*g_print ("friend ass: %s\n", data);*/
-		if (mono_assembly_name_parse_full (data, aname, TRUE, NULL)) {
+		if (mono_assembly_name_parse_full (data, aname, TRUE, NULL, NULL)) {
 			ass->friend_assembly_names = g_slist_prepend (ass->friend_assembly_names, aname);
 		} else {
 			g_free (aname);
 		}
 	}
 	mono_custom_attrs_free (attrs);
+
+	/* Because of the double checked locking pattern above */
+	mono_memory_barrier ();
+	ass->friend_assembly_names_inited = TRUE;
+	mono_assemblies_unlock ();
 }
 
 /**
@@ -1412,8 +1406,6 @@ mono_assembly_load_from_full (MonoImage *image, const char*fname,
 {
 	MonoAssembly *ass, *ass2;
 	char *base_dir;
-	GList *loading;
-	GHashTable *ass_loading;
 
 	if (!image->tables [MONO_TABLE_ASSEMBLY].rows) {
 		/* 'image' doesn't have a manifest -- maybe someone is trying to Assembly.Load a .netmodule */
@@ -1421,7 +1413,6 @@ mono_assembly_load_from_full (MonoImage *image, const char*fname,
 		return NULL;
 	}
 
-
 #if defined (PLATFORM_WIN32)
 	{
 		gchar *tmp_fn;
@@ -1440,14 +1431,6 @@ mono_assembly_load_from_full (MonoImage *image, const char*fname,
 	base_dir = absolute_dir (fname);
 #endif
 
-	/*
-	 * To avoid deadlocks and scalability problems, we load assemblies outside
-	 * the assembly lock. This means that multiple threads might try to load
-	 * the same assembly at the same time. The first one to load it completely
-	 * "wins", the other threads free their copy and use the one loaded by
-	 * the winning thread.
-	 */
-
 	/*
 	 * Create assembly struct, and enter it into the assembly cache
 	 */
@@ -1456,22 +1439,31 @@ mono_assembly_load_from_full (MonoImage *image, const char*fname,
 	ass->ref_only = refonly;
 	ass->image = image;
 
-	/* Add a non-temporary reference because of ass->image */
-	mono_image_addref (image);
+	mono_profiler_assembly_event (ass, MONO_PROFILE_START_LOAD);
 
 	mono_assembly_fill_assembly_name (image, &ass->aname);
 
+	if (mono_defaults.corlib && strcmp (ass->aname.name, "mscorlib") == 0) {
+		// MS.NET doesn't support loading other mscorlibs
+		g_free (ass);
+		g_free (base_dir);
+		mono_image_addref (mono_defaults.corlib);
+		*status = MONO_IMAGE_OK;
+		return mono_defaults.corlib->assembly;
+	}
+
+	/* Add a non-temporary reference because of ass->image */
+	mono_image_addref (image);
+
 	mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_ASSEMBLY, "Image addref %s %p -> %s %p: %d\n", ass->aname.name, ass, image->name, image, image->ref_count);
 
 	/* 
-	 * Atomically search the loaded list and add ourselves to it if necessary.
+	 * The load hooks might take locks so we can't call them while holding the
+	 * assemblies lock.
 	 */
-	mono_assemblies_lock ();
 	if (ass->aname.name) {
-		/* avoid loading the same assembly twice for now... */
-		ass2 = search_loaded (&ass->aname, refonly);
+		ass2 = mono_assembly_invoke_search_hook_internal (&ass->aname, refonly, FALSE);
 		if (ass2) {
-			mono_assemblies_unlock ();
 			g_free (ass);
 			g_free (base_dir);
 			mono_image_close (image);
@@ -1479,49 +1471,37 @@ mono_assembly_load_from_full (MonoImage *image, const char*fname,
 			return ass2;
 		}
 	}
-	ass_loading = refonly ? assemblies_refonly_loading : assemblies_loading;
-	loading = g_hash_table_lookup (ass_loading, (gpointer)GetCurrentThreadId ());
-	loading = g_list_prepend (loading, ass);
-	g_hash_table_insert (ass_loading, (gpointer)GetCurrentThreadId (), loading);
-	mono_assemblies_unlock ();
-
-	g_assert (image->assembly == NULL);
-	image->assembly = ass;
-
-	mono_assembly_load_references (image, status);
 
 	mono_assemblies_lock ();
 
-	loading = g_hash_table_lookup (ass_loading, (gpointer)GetCurrentThreadId ());
-	loading = g_list_remove (loading, ass);
-	if (loading == NULL)
-		/* Prevent memory leaks */
-		g_hash_table_remove (ass_loading, (gpointer)GetCurrentThreadId ());
-	else
-		g_hash_table_insert (ass_loading, (gpointer)GetCurrentThreadId (), loading);
-	if (*status != MONO_IMAGE_OK) {
+	if (image->assembly) {
+		/* 
+		 * This means another thread has already loaded the assembly, but not yet
+		 * called the load hooks so the search hook can't find the assembly.
+		 */
 		mono_assemblies_unlock ();
-		mono_assembly_close (ass);
-		return NULL;
+		ass2 = image->assembly;
+		g_free (ass);
+		g_free (base_dir);
+		mono_image_close (image);
+		*status = MONO_IMAGE_OK;
+		return ass2;
 	}
 
-	if (ass->aname.name) {
-		ass2 = search_loaded (&ass->aname, refonly);
-		if (ass2) {
-			/* Somebody else has loaded the assembly before us */
-			mono_assemblies_unlock ();
-			mono_assembly_close (ass);
-			return ass2;
-		}
-	}
+	image->assembly = ass;
 
 	loaded_assemblies = g_list_prepend (loaded_assemblies, ass);
-	if (mono_defaults.internals_visible_class)
-		load_friend_assemblies (ass);
 	mono_assemblies_unlock ();
 
+#ifdef PLATFORM_WIN32
+	if (image->is_module_handle)
+		mono_image_fixup_vtable (image);
+#endif
+
 	mono_assembly_invoke_load_hook (ass);
 
+	mono_profiler_assembly_loaded (ass, MONO_PROFILE_OK);
+	
 	return ass;
 }
 
@@ -1625,7 +1605,7 @@ parse_public_key (const gchar *key, gchar** pubkey)
 }
 
 static gboolean
-build_assembly_name (const char *name, const char *version, const char *culture, const char *token, const char *key, MonoAssemblyName *aname, gboolean save_public_key)
+build_assembly_name (const char *name, const char *version, const char *culture, const char *token, const char *key, guint32 flags, MonoAssemblyName *aname, gboolean save_public_key)
 {
 	gint major, minor, build, revision;
 	gint len;
@@ -1654,6 +1634,7 @@ build_assembly_name (const char *name, const char *version, const char *culture,
 			aname->revision = 0;
 	}
 	
+	aname->flags = flags;
 	aname->name = g_strdup (name);
 	
 	if (culture) {
@@ -1664,13 +1645,20 @@ build_assembly_name (const char *name, const char *version, const char *culture,
 	}
 	
 	if (token && strncmp (token, "null", 4) != 0) {
-		char *lower = g_ascii_strdown (token, MONO_PUBLIC_KEY_TOKEN_LENGTH);
+		char *lower;
+
+		/* the constant includes the ending NULL, hence the -1 */
+		if (strlen (token) != (MONO_PUBLIC_KEY_TOKEN_LENGTH - 1)) {
+			mono_assembly_name_free (aname);
+			return FALSE;
+		}
+		lower = g_ascii_strdown (token, MONO_PUBLIC_KEY_TOKEN_LENGTH);
 		g_strlcpy ((char*)aname->public_key_token, lower, MONO_PUBLIC_KEY_TOKEN_LENGTH);
 		g_free (lower);
 	}
 
-	if (key && strncmp (key, "null", 4) != 0) {
-		if (!parse_public_key (key, &pkey)) {
+	if (key) {
+		if (strcmp (key, "null") == 0 || !parse_public_key (key, &pkey)) {
 			mono_assembly_name_free (aname);
 			return FALSE;
 		}
@@ -1687,7 +1675,7 @@ build_assembly_name (const char *name, const char *version, const char *culture,
 		else
 			g_free (pkey);
 	}
-	
+
 	return TRUE;
 }
 
@@ -1703,30 +1691,36 @@ parse_assembly_directory_name (const char *name, const char *dirname, MonoAssemb
 		return FALSE;
 	}
 	
-	res = build_assembly_name (name, parts[0], parts[1], parts[2], NULL, aname, FALSE);
+	res = build_assembly_name (name, parts[0], parts[1], parts[2], NULL, 0, aname, FALSE);
 	g_strfreev (parts);
 	return res;
 }
 
 gboolean
-mono_assembly_name_parse_full (const char *name, MonoAssemblyName *aname, gboolean save_public_key, gboolean *is_version_defined)
+mono_assembly_name_parse_full (const char *name, MonoAssemblyName *aname, gboolean save_public_key, gboolean *is_version_defined, gboolean *is_token_defined)
 {
 	gchar *dllname;
 	gchar *version = NULL;
 	gchar *culture = NULL;
 	gchar *token = NULL;
 	gchar *key = NULL;
+	gchar *retargetable = NULL;
 	gboolean res;
 	gchar *value;
 	gchar **parts;
 	gchar **tmp;
 	gboolean version_defined;
+	gboolean token_defined;
+	guint32 flags = 0;
 
 	if (!is_version_defined)
 		is_version_defined = &version_defined;
 	*is_version_defined = FALSE;
+	if (!is_token_defined)
+		is_token_defined = &token_defined;
+	*is_token_defined = FALSE;
 	
-	parts = tmp = g_strsplit (name, ",", 4);
+	parts = tmp = g_strsplit (name, ",", 6);
 	if (!tmp || !*tmp) {
 		g_strfreev (tmp);
 		return FALSE;
@@ -1741,35 +1735,78 @@ mono_assembly_name_parse_full (const char *name, MonoAssemblyName *aname, gboole
 		if (!g_ascii_strncasecmp (value, "Version=", 8)) {
 			*is_version_defined = TRUE;
 			version = g_strstrip (value + 8);
+			if (strlen (version) == 0) {
+				goto cleanup_and_fail;
+			}
 			tmp++;
 			continue;
 		}
 
 		if (!g_ascii_strncasecmp (value, "Culture=", 8)) {
 			culture = g_strstrip (value + 8);
+			if (strlen (culture) == 0) {
+				goto cleanup_and_fail;
+			}
 			tmp++;
 			continue;
 		}
 
 		if (!g_ascii_strncasecmp (value, "PublicKeyToken=", 15)) {
+			*is_token_defined = TRUE;
 			token = g_strstrip (value + 15);
+			if (strlen (token) == 0) {
+				goto cleanup_and_fail;
+			}
 			tmp++;
 			continue;
 		}
 
 		if (!g_ascii_strncasecmp (value, "PublicKey=", 10)) {
 			key = g_strstrip (value + 10);
+			if (strlen (key) == 0) {
+				goto cleanup_and_fail;
+			}
 			tmp++;
 			continue;
 		}
-		
+
+		if (!g_ascii_strncasecmp (value, "Retargetable=", 13)) {
+			retargetable = g_strstrip (value + 13);
+			if (strlen (retargetable) == 0) {
+				goto cleanup_and_fail;
+			}
+			if (!g_ascii_strcasecmp (retargetable, "yes")) {
+				flags |= ASSEMBLYREF_RETARGETABLE_FLAG;
+			} else if (g_ascii_strcasecmp (retargetable, "no")) {
+				goto cleanup_and_fail;
+			}
+			tmp++;
+			continue;
+		}
+
+		if (!g_ascii_strncasecmp (value, "ProcessorArchitecture=", 22)) {
+			/* this is ignored for now, until we can change MonoAssemblyName */
+			tmp++;
+			continue;
+		}
+
 		g_strfreev (parts);
 		return FALSE;
 	}
 
-	res = build_assembly_name (dllname, version, culture, token, key, aname, save_public_key);
+	/* if retargetable flag is set, then we must have a fully qualified name */
+	if (retargetable != NULL && (version == NULL || culture == NULL || (key == NULL && token == NULL))) {
+		goto cleanup_and_fail;
+	}
+
+	res = build_assembly_name (dllname, version, culture, token, key, flags,
+		aname, save_public_key);
 	g_strfreev (parts);
 	return res;
+
+cleanup_and_fail:
+	g_strfreev (parts);
+	return FALSE;
 }
 
 /**
@@ -1785,7 +1822,7 @@ mono_assembly_name_parse_full (const char *name, MonoAssemblyName *aname, gboole
 gboolean
 mono_assembly_name_parse (const char *name, MonoAssemblyName *aname)
 {
-	return mono_assembly_name_parse_full (name, aname, FALSE, NULL);
+	return mono_assembly_name_parse_full (name, aname, FALSE, NULL, NULL);
 }
 
 static MonoAssembly*
@@ -1807,13 +1844,14 @@ probe_for_partial_name (const char *basepath, const char *fullname, MonoAssembly
 	while ((direntry = g_dir_read_name (dirhandle))) {
 		gboolean match = TRUE;
 		
-		parse_assembly_directory_name (aname->name, direntry, &gac_aname);
+		if(!parse_assembly_directory_name (aname->name, direntry, &gac_aname))
+			continue;
 		
 		if (aname->culture != NULL && strcmp (aname->culture, gac_aname.culture) != 0)
 			match = FALSE;
 			
 		if (match && strlen ((char*)aname->public_key_token) > 0 && 
-				strcmp ((char*)aname->public_key_token, (char*)gac_aname.public_key_token) != 0)
+				!mono_public_tokens_are_equal (aname->public_key_token, gac_aname.public_key_token))
 			match = FALSE;
 		
 		if (match) {
@@ -1917,7 +1955,13 @@ mono_assembly_load_with_partial_name (const char *name, MonoImageOpenStatus *sta
 
 	if (res)
 		res->in_gac = TRUE;
-
+	else {
+		MonoDomain *domain = mono_domain_get ();
+		MonoReflectionAssembly *refasm = mono_try_assembly_resolve (domain, mono_string_new (domain, name), FALSE);
+		if (refasm)
+			res = refasm->assembly;
+	}
+	
 	g_free (fullname);
 	mono_assembly_name_free (aname);
 
@@ -2079,6 +2123,7 @@ mono_assembly_load_from_gac (MonoAssemblyName *aname,  gchar *filename, MonoImag
 	gchar *name, *version, *culture, *fullpath, *subpath;
 	gint32 len;
 	gchar **paths;
+	char *pubtok;
 
 	if (aname->public_key_token [0] == 0) {
 		return NULL;
@@ -2099,9 +2144,11 @@ mono_assembly_load_from_gac (MonoAssemblyName *aname,  gchar *filename, MonoImag
 		culture = g_strdup ("");
 	}
 
+	pubtok = g_ascii_strdown ((char*)aname->public_key_token, MONO_PUBLIC_KEY_TOKEN_LENGTH);
 	version = g_strdup_printf ("%d.%d.%d.%d_%s_%s", aname->major,
 			aname->minor, aname->build, aname->revision,
-			culture, aname->public_key_token);
+			culture, pubtok);
+	g_free (pubtok);
 	
 	subpath = g_build_path (G_DIR_SEPARATOR_S, name, version, filename, NULL);
 	g_free (name);
@@ -2170,24 +2217,11 @@ mono_assembly_load_corlib (const MonoRuntimeInfo *runtime, MonoImageOpenStatus *
 	return corlib;
 }
 
-/**
- * mono_assembly_load_full:
- * @aname: A MonoAssemblyName with the assembly name to load.
- * @basedir: A directory to look up the assembly at.
- * @status: a pointer to a MonoImageOpenStatus to return the status of the load operation
- * @refonly: Whether this assembly is being opened in "reflection-only" mode.
- *
- * Loads the assembly referenced by @aname, if the value of @basedir is not NULL, it
- * attempts to load the assembly from that directory before probing the standard locations.
- *
- * If the assembly is being opened in reflection-only mode (@refonly set to TRUE) then no 
- * assembly binding takes place.
- *
- * Returns: the assembly referenced by @aname loaded or NULL on error.   On error the
- * value pointed by status is updated with an error code.
- */
 MonoAssembly*
-mono_assembly_load_full (MonoAssemblyName *aname, const char *basedir, MonoImageOpenStatus *status, gboolean refonly)
+mono_assembly_load_full_nosearch (MonoAssemblyName *aname, 
+								  const char       *basedir, 
+								  MonoImageOpenStatus *status,
+								  gboolean refonly)
 {
 	MonoAssembly *result;
 	char *fullpath, *filename;
@@ -2255,9 +2289,33 @@ mono_assembly_load_full (MonoAssemblyName *aname, const char *basedir, MonoImage
 			return result;
 	}
 
-	/* Try a postload search hook */
-	result = mono_assembly_invoke_search_hook_internal (aname, refonly, TRUE);
+	return result;
+}
 
+/**
+ * mono_assembly_load_full:
+ * @aname: A MonoAssemblyName with the assembly name to load.
+ * @basedir: A directory to look up the assembly at.
+ * @status: a pointer to a MonoImageOpenStatus to return the status of the load operation
+ * @refonly: Whether this assembly is being opened in "reflection-only" mode.
+ *
+ * Loads the assembly referenced by @aname, if the value of @basedir is not NULL, it
+ * attempts to load the assembly from that directory before probing the standard locations.
+ *
+ * If the assembly is being opened in reflection-only mode (@refonly set to TRUE) then no 
+ * assembly binding takes place.
+ *
+ * Returns: the assembly referenced by @aname loaded or NULL on error.   On error the
+ * value pointed by status is updated with an error code.
+ */
+MonoAssembly*
+mono_assembly_load_full (MonoAssemblyName *aname, const char *basedir, MonoImageOpenStatus *status, gboolean refonly)
+{
+	MonoAssembly *result = mono_assembly_load_full_nosearch (aname, basedir, status, refonly);
+	
+	if (!result)
+		/* Try a postload search hook */
+		result = mono_assembly_invoke_search_hook_internal (aname, refonly, TRUE);
 	return result;
 }
 
@@ -2287,9 +2345,7 @@ mono_assembly_loaded_full (MonoAssemblyName *aname, gboolean refonly)
 
 	aname = mono_assembly_remap_version (aname, &maped_aname);
 
-	mono_assemblies_lock ();
-	res = search_loaded (aname, refonly);
-	mono_assemblies_unlock ();
+	res = mono_assembly_invoke_search_hook_internal (aname, refonly, FALSE);
 
 	return res;
 }
@@ -2327,24 +2383,16 @@ mono_assembly_close (MonoAssembly *assembly)
 	if (InterlockedDecrement (&assembly->ref_count) > 0)
 		return;
 
+	mono_profiler_assembly_event (assembly, MONO_PROFILE_START_UNLOAD);
+
 	mono_trace (G_LOG_LEVEL_INFO, MONO_TRACE_ASSEMBLY, "Unloading assembly %s [%p].", assembly->aname.name, assembly);
 
+	mono_debug_close_image (assembly->image);
+
 	mono_assemblies_lock ();
 	loaded_assemblies = g_list_remove (loaded_assemblies, assembly);
 	mono_assemblies_unlock ();
 
-	if (assembly->image->references) {
-		int i;
-
-		for (i = 0; assembly->image->references [i]; i++) {
-			if (assembly->image->references [i])
-				mono_assembly_close (assembly->image->references [i]);
-		}
-
-		g_free (assembly->image->references);
-		assembly->image->references = NULL;
-	}
-
 	assembly->image->assembly = NULL;
 
 	mono_image_close (assembly->image);
@@ -2361,19 +2409,14 @@ mono_assembly_close (MonoAssembly *assembly)
 	} else {
 		g_free (assembly);
 	}
+
+	mono_profiler_assembly_event (assembly, MONO_PROFILE_END_UNLOAD);
 }
 
 MonoImage*
 mono_assembly_load_module (MonoAssembly *assembly, guint32 idx)
 {
-	MonoImageOpenStatus status;
-	MonoImage *module;
-
-	module = mono_image_load_file_for_image (assembly->image, idx);
-	if (module)
-		mono_assembly_load_references (module, &status);
-
-	return module;
+	return mono_image_load_file_for_image (assembly->image, idx);
 }
 
 void
@@ -2406,9 +2449,6 @@ mono_assemblies_cleanup (void)
 
 	DeleteCriticalSection (&assemblies_mutex);
 
-	g_hash_table_destroy (assemblies_loading);
-	g_hash_table_destroy (assemblies_refonly_loading);
-
 	for (l = loaded_assembly_bindings; l; l = l->next) {
 		MonoAssemblyBindingInfo *info = l->data;
 
@@ -2462,3 +2502,125 @@ mono_register_bundled_assemblies (const MonoBundledAssembly **assemblies)
 {
 	bundles = assemblies;
 }
+
+#define MONO_DECLSEC_FORMAT_20		0x2E
+#define MONO_DECLSEC_FIELD		0x53
+#define MONO_DECLSEC_PROPERTY		0x54
+
+#define SKIP_VISIBILITY_ATTRIBUTE_NAME ("System.Security.Permissions.SecurityPermissionAttribute")
+#define SKIP_VISIBILITY_ATTRIBUTE_SIZE (sizeof (SKIP_VISIBILITY_ATTRIBUTE_NAME) - 1)
+#define SKIP_VISIBILITY_PROPERTY_NAME ("SkipVerification")
+#define SKIP_VISIBILITY_PROPERTY_SIZE (sizeof (SKIP_VISIBILITY_PROPERTY_NAME) - 1)
+
+static gboolean
+mono_assembly_try_decode_skip_verification_param (const char *p, const char **resp, gboolean *abort_decoding)
+{
+	int len;
+	switch (*p++) {
+	case MONO_DECLSEC_PROPERTY:
+		break;
+	case MONO_DECLSEC_FIELD:
+	default:
+		*abort_decoding = TRUE;
+		return FALSE;
+		break;
+	}
+
+	if (*p++ != MONO_TYPE_BOOLEAN) {
+		*abort_decoding = TRUE;
+		return FALSE;
+	}
+		
+	/* property name length */
+	len = mono_metadata_decode_value (p, &p);
+
+	if (len >= SKIP_VISIBILITY_PROPERTY_SIZE && !memcmp (p, SKIP_VISIBILITY_PROPERTY_NAME, SKIP_VISIBILITY_PROPERTY_SIZE)) {
+		p += len;
+		return *p;
+	}
+	p += len + 1;
+
+	*resp = p;
+	return FALSE;
+}
+
+static gboolean
+mono_assembly_try_decode_skip_verification (const char *p, const char *endn)
+{
+	int i, j, num, len, params_len;
+
+	if (*p++ != MONO_DECLSEC_FORMAT_20)
+		return FALSE;
+
+	/* number of encoded permission attributes */
+	num = mono_metadata_decode_value (p, &p);
+	for (i = 0; i < num; ++i) {
+		gboolean is_valid = FALSE;
+		gboolean abort_decoding = FALSE;
+
+		/* attribute name length */
+		len =  mono_metadata_decode_value (p, &p);
+
+		/* We don't really need to fully decode the type. Comparing the name is enough */
+		is_valid = len >= SKIP_VISIBILITY_ATTRIBUTE_SIZE && !memcmp (p, SKIP_VISIBILITY_ATTRIBUTE_NAME, SKIP_VISIBILITY_ATTRIBUTE_SIZE);
+
+		p += len;
+
+		/*size of the params table*/
+		params_len =  mono_metadata_decode_value (p, &p);
+		if (is_valid) {
+			const char *params_end = p + params_len;
+			
+			/* number of parameters */
+			len = mono_metadata_decode_value (p, &p);
+	
+			for (j = 0; j < len; ++j) {
+				if (mono_assembly_try_decode_skip_verification_param (p, &p, &abort_decoding))
+					return TRUE;
+				if (abort_decoding)
+					break;
+			}
+			p = params_end;
+		} else {
+			p += params_len;
+		}
+	}
+	
+	return FALSE;
+}
+
+
+gboolean
+mono_assembly_has_skip_verification (MonoAssembly *assembly)
+{
+	MonoTableInfo *t;	
+	guint32 cols [MONO_DECL_SECURITY_SIZE];
+	const char *blob;
+	int i, len;
+
+	if (MONO_SECMAN_FLAG_INIT (assembly->skipverification))
+		return MONO_SECMAN_FLAG_GET_VALUE (assembly->skipverification);
+
+	t = &assembly->image->tables [MONO_TABLE_DECLSECURITY];
+
+	for (i = 0; i < t->rows; ++i) {
+		mono_metadata_decode_row (t, i, cols, MONO_DECL_SECURITY_SIZE);
+		if ((cols [MONO_DECL_SECURITY_PARENT] & MONO_HAS_DECL_SECURITY_MASK) != MONO_HAS_DECL_SECURITY_ASSEMBLY)
+			continue;
+		if (cols [MONO_DECL_SECURITY_ACTION] != SECURITY_ACTION_REQMIN)
+			continue;
+
+		blob = mono_metadata_blob_heap (assembly->image, cols [MONO_DECL_SECURITY_PERMISSIONSET]);
+		len = mono_metadata_decode_blob_size (blob, &blob);
+		if (!len)
+			continue;
+
+		if (mono_assembly_try_decode_skip_verification (blob, blob + len)) {
+			MONO_SECMAN_FLAG_SET_VALUE (assembly->skipverification, TRUE);
+			return TRUE;
+		}
+	}
+
+	MONO_SECMAN_FLAG_SET_VALUE (assembly->skipverification, FALSE);
+	return FALSE;
+}
