@@ -12,11 +12,6 @@
 #include <config.h>
 #include <glib.h>
 
-#ifdef PLATFORM_WIN32
-#define WINVER 0x0500
-#define _WIN32_WINNT 0x0500
-#endif
-
 #define THREADS_PER_CPU	5 /* 20 + THREADS_PER_CPU * number of CPUs */
 
 #include <mono/metadata/domain-internals.h>
@@ -31,12 +26,17 @@
 #include <mono/metadata/marshal.h>
 #include <mono/metadata/socket-io.h>
 #include <mono/io-layer/io-layer.h>
-#include <mono/os/gc_wrapper.h>
+#include <mono/metadata/gc-internal.h>
+#include <mono/utils/mono-time.h>
 #include <errno.h>
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
+#endif
 #include <sys/types.h>
 #include <fcntl.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 #include <string.h>
 
 #include <mono/utils/mono-poll.h>
@@ -218,11 +218,11 @@ get_events_from_list (MonoMList *list)
 }
 
 #define ICALL_RECV(x)	ves_icall_System_Net_Sockets_Socket_Receive_internal (\
-				(SOCKET) x->handle, x->buffer, x->offset, x->size,\
+				(SOCKET)(gssize)x->handle, x->buffer, x->offset, x->size,\
 				 x->socket_flags, &x->error);
 
 #define ICALL_SEND(x)	ves_icall_System_Net_Sockets_Socket_Send_internal (\
-				(SOCKET) x->handle, x->buffer, x->offset, x->size,\
+				(SOCKET)(gssize)x->handle, x->buffer, x->offset, x->size,\
 				 x->socket_flags, &x->error);
 
 static void
@@ -231,7 +231,6 @@ async_invoke_io_thread (gpointer data)
 	MonoDomain *domain;
 	MonoThread *thread;
 	thread = mono_thread_current ();
-	thread->threadpool_thread = TRUE;
 	ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
 
 	for (;;) {
@@ -275,14 +274,14 @@ async_invoke_io_thread (gpointer data)
 		if (!data) {
 			guint32 wr;
 			int timeout = 10000;
-			guint32 start_time = GetTickCount ();
+			guint32 start_time = mono_msec_ticks ();
 			
 			do {
 				wr = WaitForSingleObjectEx (io_job_added, (guint32)timeout, TRUE);
 				if (THREAD_WANTS_A_BREAK (thread))
 					mono_thread_interruption_checkpoint ();
 			
-				timeout -= GetTickCount () - start_time;
+				timeout -= mono_msec_ticks () - start_time;
 			
 				if (wr != WAIT_TIMEOUT)
 					data = dequeue_job (&io_queue_lock, &async_io_queue);
@@ -320,7 +319,7 @@ start_io_thread_or_queue (MonoSocketAsyncResult *ares)
 		InterlockedIncrement (&busy_io_worker_threads);
 		InterlockedIncrement (&io_worker_threads);
 		domain = ((ares) ? ((MonoObject *) ares)->vtable->domain : mono_domain_get ());
-		mono_thread_create (mono_get_root_domain (), async_invoke_io_thread, ares);
+		mono_thread_create_internal (mono_get_root_domain (), async_invoke_io_thread, ares, TRUE);
 	} else {
 		append_job (&io_queue_lock, &async_io_queue, (MonoObject*)ares);
 		ReleaseSemaphore (io_job_added, 1, NULL);
@@ -392,7 +391,6 @@ socket_io_poll_main (gpointer p)
 	MonoThread *thread;
 
 	thread = mono_thread_current ();
-	thread->threadpool_thread = TRUE;
 	ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
 
 	allocated = INITIAL_POLLFD_SIZE;
@@ -630,11 +628,10 @@ socket_io_epoll_main (gpointer p)
 void
 mono_thread_pool_remove_socket (int sock)
 {
-#ifdef HAVE_EPOLL
 	MonoMList *list, *next;
 	MonoSocketAsyncResult *state;
 
-	if (socket_io_data.epoll_disabled == TRUE || socket_io_data.inited == FALSE)
+	if (socket_io_data.inited == FALSE)
 		return;
 
 	EnterCriticalSection (&socket_io_data.io_lock);
@@ -658,7 +655,6 @@ mono_thread_pool_remove_socket (int sock)
 
 		list = next;
 	}
-#endif
 }
 
 #ifdef PLATFORM_WIN32
@@ -762,11 +758,11 @@ socket_io_init (SocketIOData *data)
 	g_assert (io_job_added != NULL);
 	InitializeCriticalSection (&io_queue_lock);
 	if (data->epoll_disabled) {
-		mono_thread_create (mono_get_root_domain (), socket_io_poll_main, data);
+		mono_thread_create_internal (mono_get_root_domain (), socket_io_poll_main, data, TRUE);
 	}
 #ifdef HAVE_EPOLL
 	else {
-		mono_thread_create (mono_get_root_domain (), socket_io_epoll_main, data);
+		mono_thread_create_internal (mono_get_root_domain (), socket_io_epoll_main, data, TRUE);
 	}
 #endif
 	InterlockedCompareExchange (&data->inited, 1, 0);
@@ -781,7 +777,7 @@ socket_io_add_poll (MonoSocketAsyncResult *state)
 	MonoMList *list;
 	SocketIOData *data = &socket_io_data;
 
-#if defined(PLATFORM_MACOSX) || defined(PLATFORM_BSD6) || defined(PLATFORM_WIN32)
+#if defined(PLATFORM_MACOSX) || defined(PLATFORM_BSD6) || defined(PLATFORM_WIN32) || defined(PLATFORM_SOLARIS)
 	/* select() for connect() does not work well on the Mac. Bug #75436. */
 	/* Bug #77637 for the BSD 6 case */
 	/* Bug #78888 for the Windows case */
@@ -1049,7 +1045,7 @@ start_thread_or_queue (MonoAsyncResult *ares)
 	    worker < mono_max_worker_threads) {
 		InterlockedIncrement (&mono_worker_threads);
 		InterlockedIncrement (&busy_worker_threads);
-		mono_thread_create (mono_get_root_domain (), async_invoke_thread, ares);
+		mono_thread_create_internal (mono_get_root_domain (), async_invoke_thread, ares, TRUE);
 	} else {
 		append_job (&mono_delegate_section, &async_call_queue, (MonoObject*)ares);
 		ReleaseSemaphore (job_added, 1, NULL);
@@ -1218,14 +1214,14 @@ async_invoke_thread (gpointer data)
 		if (!data) {
 			guint32 wr;
 			int timeout = 10000;
-			guint32 start_time = GetTickCount ();
+			guint32 start_time = mono_msec_ticks ();
 			
 			do {
 				wr = WaitForSingleObjectEx (job_added, (guint32)timeout, TRUE);
 				if (THREAD_WANTS_A_BREAK (thread))
 					mono_thread_interruption_checkpoint ();
 			
-				timeout -= GetTickCount () - start_time;
+				timeout -= mono_msec_ticks () - start_time;
 			
 				if (wr != WAIT_TIMEOUT)
 					data = dequeue_job (&mono_delegate_section, &async_call_queue);
