@@ -26,3 +26,5 @@ WRAPPER(WRITE_BARRIER, "write-barrier")
 WRAPPER(UNKNOWN, "unknown")
 WRAPPER(COMINTEROP_INVOKE, "cominterop-invoke")
 WRAPPER(COMINTEROP, "cominterop")
+WRAPPER(ALLOC, "alloc")
+WRAPPER(STATIC_RGCTX_INVOKE, "static-rgctx-invoke")
