@@ -702,9 +702,9 @@ mono_cprop_invalidate_values (MonoInst *tree, TreeMover *tree_mover, MonoInst **
 		}
 
 		break;
-	case CEE_CALL:
+	case OP_CALL:
 	case OP_CALL_REG:
-	case CEE_CALLVIRT:
+	case OP_CALLVIRT:
 	case OP_LCALL_REG:
 	case OP_LCALLVIRT:
 	case OP_LCALL:
@@ -716,7 +716,23 @@ mono_cprop_invalidate_values (MonoInst *tree, TreeMover *tree_mover, MonoInst **
 	case OP_VCALL:
 	case OP_VOIDCALL_REG:
 	case OP_VOIDCALLVIRT:
-	case OP_VOIDCALL: {
+	case OP_VOIDCALL:
+	case OP_TRAMPCALL_VTABLE:
+	case OP_CALL_RGCTX:
+	case OP_FCALL_RGCTX:
+	case OP_VOIDCALL_RGCTX:
+	case OP_LCALL_RGCTX:
+	case OP_VCALL_RGCTX:
+	case OP_CALL_REG_RGCTX:
+	case OP_FCALL_REG_RGCTX:
+	case OP_VOIDCALL_REG_RGCTX:
+	case OP_LCALL_REG_RGCTX:
+	case OP_VCALL_REG_RGCTX:
+	case OP_CALLVIRT_IMT:
+	case OP_VOIDCALLVIRT_IMT:
+	case OP_FCALLVIRT_IMT:
+	case OP_LCALLVIRT_IMT:
+	case OP_VCALLVIRT_IMT: {
 		MonoCallInst *call = (MonoCallInst *)tree;
 		MonoMethodSignature *sig = call->signature;
 		int i, byref = FALSE;
@@ -801,10 +817,10 @@ mono_cprop_invalidate_values (MonoInst *tree, TreeMover *tree_mover, MonoInst **
 static void
 mono_local_cprop_bb (MonoCompile *cfg, TreeMover *tree_mover, MonoBasicBlock *bb, MonoInst **acp, int acp_size)
 {
-	MonoInst *tree = bb->code;
+	MonoInst *tree;
 	int i;
 
-	if (!tree)
+	if (MONO_INST_LIST_EMPTY (&bb->ins_list))
 		return;
 
 	if (tree_mover != NULL) {
@@ -813,7 +829,7 @@ mono_local_cprop_bb (MonoCompile *cfg, TreeMover *tree_mover, MonoBasicBlock *bb
 			printf ("Running tree mover on BB%d\n", bb->block_num);
 		}
 	}
-	for (; tree; tree = tree->next) {
+	MONO_BB_FOR_EACH_INS (bb, tree) {
 		if (tree_mover != NULL) {
 			if (MONO_DEBUG_TREE_MOVER) {
 				printf ("Running tree mover on tree ");
@@ -839,7 +855,7 @@ mono_local_cprop_bb (MonoCompile *cfg, TreeMover *tree_mover, MonoBasicBlock *bb
 
 			acp [tree->inst_i0->inst_c0] = NULL;
 			if (MONO_DEBUG_TREE_MOVER) {
-				printf ("Assignment clears ACP[%d] at tree ", tree->inst_i0->inst_c0);
+				printf ("Assignment clears ACP[%d] at tree ", (int)tree->inst_i0->inst_c0);
 				mono_print_tree (tree);
 				printf ("\n");
 			}
@@ -858,7 +874,7 @@ mono_local_cprop_bb (MonoCompile *cfg, TreeMover *tree_mover, MonoBasicBlock *bb
 				acp [tree->inst_i0->inst_c0] = i1;
 				tree_can_be_moved = FALSE;
 				if (MONO_DEBUG_LOCAL_PROP) {
-					printf ("  Consequently, ACP[%d] becomes constant ", tree->inst_i0->inst_c0);
+					printf ("  Consequently, ACP[%ld] becomes constant ", (long)tree->inst_i0->inst_c0);
 					mono_print_tree (i1);
 					printf ("\n");
 				}
@@ -877,7 +893,7 @@ mono_local_cprop_bb (MonoCompile *cfg, TreeMover *tree_mover, MonoBasicBlock *bb
 				acp [tree->inst_i0->inst_c0] = i1->inst_i0;
 				tree_can_be_moved = FALSE;
 				if (MONO_DEBUG_LOCAL_PROP) {
-					printf ("  Consequently, ACP[%d] becomes local ", tree->inst_i0->inst_c0);
+					printf ("  Consequently, ACP[%d] becomes local ", (int)tree->inst_i0->inst_c0);
 					mono_print_tree (i1->inst_i0);
 					printf ("\n");
 				}
@@ -1012,7 +1028,7 @@ mono_local_cprop_bb (MonoCompile *cfg, TreeMover *tree_mover, MonoBasicBlock *bb
 		  if (tree->inst_i0->inst_i0->opcode == OP_ICONST &&
 		  tree->inst_i0->inst_i1->opcode == OP_ICONST) {
 
-		  tree->opcode = CEE_BR;
+		  tree->opcode = OP_BR;
 		  if (tree->inst_i0->inst_i0->opcode == tree->inst_i0->inst_i1->opcode) {
 		  tree->inst_target_bb = tree->inst_true_bb;
 		  } else {
@@ -1122,7 +1138,7 @@ apply_tree_mover (TreeMover *tree_mover, TreeMoverTreeMove *move) {
 	}
 	/* All tests passed, apply move */
 	*(move->use) = move->definition->inst_i1;
-	move->definition->opcode = CEE_NOP;
+	move->definition->opcode = OP_NOP;
 	move->definition->ssa_op = MONO_SSA_NOP;
 
 	/* Then disable moves affected by this move */
@@ -1142,7 +1158,7 @@ apply_tree_mover (TreeMover *tree_mover, TreeMoverTreeMove *move) {
 				printf ("  Consequently, kill dependent definition %d: ", tree_mover_slot_to_index (tree_mover, dependency->defined_slot));
 				mono_print_tree_nl (dependency->dead_definition);
 			}
-			dependency->dead_definition->opcode = CEE_NOP;
+			dependency->dead_definition->opcode = OP_NOP;
 			dependency->dead_definition->ssa_op = MONO_SSA_NOP;
 		}
 	}
