@@ -15,8 +15,11 @@
 #include "mono/metadata/class-internals.h"
 #include "mono/metadata/object-internals.h"
 #include <mono/metadata/profiler-private.h>
+#include <mono/metadata/debug-helpers.h>
 #include <mono/utils/mono-compiler.h>
 
+#define MONO_BREAKPOINT_ARRAY_SIZE 64
+
 #include "mini-arch.h"
 #include "regalloc.h"
 #include "declsec.h"
@@ -26,31 +29,44 @@
 #define G_UNLIKELY(a) (a)
 #endif
 
+#ifndef G_MAXINT32
+#define G_MAXINT32 2147483647
+#endif
+
+#ifndef G_MININT32
+#define G_MININT32 (-G_MAXINT32 - 1)
+#endif
+
 #if DISABLE_LOGGING
 #define MINI_DEBUG(level,limit,code)
 #else
 #define MINI_DEBUG(level,limit,code) do {if (G_UNLIKELY ((level) >= (limit))) code} while (0)
 #endif
 
+#define NOT_IMPLEMENTED do { g_assert_not_reached (); } while (0)
+
 #ifndef DISABLE_AOT
 #define MONO_USE_AOT_COMPILER
 #endif
 
 /* for 32 bit systems */
 #if G_BYTE_ORDER == G_LITTLE_ENDIAN
-#define MINI_LS_WORD_OFFSET 0
-#define MINI_MS_WORD_OFFSET 4
-#define inst_ls_word data.op[0].const_val
-#define inst_ms_word data.op[1].const_val
+#define MINI_LS_WORD_IDX 0
+#define MINI_MS_WORD_IDX 1
 #else
-#define MINI_LS_WORD_OFFSET 4
-#define MINI_MS_WORD_OFFSET 0
-#define inst_ls_word data.op[1].const_val
-#define inst_ms_word data.op[0].const_val
+#define MINI_LS_WORD_IDX 1
+#define MINI_MS_WORD_IDX 0
 #endif
+#define MINI_LS_WORD_OFFSET (MINI_LS_WORD_IDX * 4)
+#define MINI_MS_WORD_OFFSET (MINI_MS_WORD_IDX * 4)
+#define inst_ls_word data.op[MINI_LS_WORD_IDX].const_val
+#define inst_ms_word data.op[MINI_MS_WORD_IDX].const_val
+
+#define MONO_FAKE_IMT_METHOD ((MonoMethod*)GINT_TO_POINTER(-1))
+#define MONO_FAKE_VTABLE_METHOD ((MonoMethod*)GINT_TO_POINTER(-2))
 
 /* Version number of the AOT file format */
-#define MONO_AOT_FILE_VERSION "30"
+#define MONO_AOT_FILE_VERSION "37"
 
 #if 0
 #define mono_bitset_foreach_bit(set,b,n) \
@@ -79,16 +95,19 @@ enum {
 };
 #undef OPDEF
 
-#define MONO_VARINFO(cfg,varnum) ((cfg)->vars [varnum])
+#define MONO_VARINFO(cfg,varnum) (&(cfg)->vars [varnum])
 
 #define MONO_INST_NEW(cfg,dest,op) do {	\
 		(dest) = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoInst));	\
 		(dest)->opcode = (op);	\
+        (dest)->cil_code = (cfg)->ip; \
 	} while (0)
 
 #define MONO_INST_NEW_CALL(cfg,dest,op) do {	\
 		(dest) = mono_mempool_alloc0 ((cfg)->mempool, sizeof (MonoCallInst));	\
 		(dest)->inst.opcode = (op);	\
+        (dest)->inst.cil_code = (cfg)->ip; \
+		MONO_INST_LIST_INIT (&(dest)->out_args); \
 	} while (0)
 
 #define MONO_INST_NEW_CALL_ARG(cfg,dest,op) do {	\
@@ -96,20 +115,22 @@ enum {
 		(dest)->ins.opcode = (op);	\
 	} while (0)
 
-#define MONO_ADD_INS(b,inst) do {	\
-		if ((b)->last_ins) {	\
-			(b)->last_ins->next = (inst);	\
-			(b)->last_ins = (inst);	\
-		} else {	\
-			(b)->code = (b)->last_ins = (inst);	\
-		}	\
-	} while (0)
-
 #define NULLIFY_INS(ins) do { \
-        (ins)->opcode = CEE_NOP; \
+        (ins)->opcode = OP_NOP; \
         (ins)->dreg = (ins)->sreg1 = (ins)->sreg2 = -1; \
+		(ins)->ssa_op = MONO_SSA_NOP; \
     } while (0)
 
+/* 
+ * this is used to determine when some branch optimizations are possible: we exclude FP compares
+ * because they have weird semantics with NaNs.
+ */
+#define MONO_IS_COND_BRANCH_OP(ins) (((ins)->opcode >= CEE_BEQ && (ins)->opcode <= CEE_BLT_UN) || ((ins)->opcode >= OP_LBEQ && (ins)->opcode <= OP_LBLT_UN) || ((ins)->opcode >= OP_FBEQ && (ins)->opcode <= OP_FBLT_UN) || ((ins)->opcode >= OP_IBEQ && (ins)->opcode <= OP_IBLT_UN))
+#define MONO_IS_COND_BRANCH_NOFP(ins) (MONO_IS_COND_BRANCH_OP(ins) && (ins)->inst_left->inst_left->type != STACK_R8)
+
+#define MONO_IS_BRANCH_OP(ins) (MONO_IS_COND_BRANCH_OP(ins) || ((ins)->opcode == OP_BR) || ((ins)->opcode == OP_BR_REG) || ((ins)->opcode == OP_SWITCH))
+
+typedef struct MonoInstList MonoInstList;
 typedef struct MonoInst MonoInst;
 typedef struct MonoCallInst MonoCallInst;
 typedef struct MonoCallArgParm MonoCallArgParm;
@@ -129,7 +150,156 @@ extern int mono_exc_esp_offset;
 #else
 extern gboolean mono_compile_aot;
 #endif
-extern gboolean mono_use_security_manager;
+extern gboolean mono_aot_only;
+extern gboolean mono_use_imt;
+extern MonoMethodDesc *mono_inject_async_exc_method;
+extern int mono_inject_async_exc_pos;
+extern MonoMethodDesc *mono_break_at_bb_method;
+extern int mono_break_at_bb_bb_num;
+extern gboolean check_for_pending_exc;
+extern gboolean mono_verify_all;
+
+struct MonoInstList {
+	struct MonoInstList *next, *prev;
+};
+
+#define MONO_INST_LIST_INIT(list) do {	\
+		(list)->next = (list);		\
+		(list)->prev = (list);		\
+	} while (0)
+
+static inline int
+MONO_INST_LIST_EMPTY (MonoInstList *list)
+{
+	return list->next == list;
+}
+
+static inline void
+__MONO_INST_LIST_ADD (MonoInstList *new, MonoInstList *prev, MonoInstList *next)
+{
+	next->prev = new;
+	new->next = next;
+	new->prev = prev;
+	prev->next = new;
+}
+
+
+static inline void
+MONO_INST_LIST_ADD (MonoInstList *new, MonoInstList *head)
+{
+	__MONO_INST_LIST_ADD (new, head, head->next);
+}
+
+static inline void
+MONO_INST_LIST_ADD_TAIL (MonoInstList *new, MonoInstList *head)
+{
+	__MONO_INST_LIST_ADD (new, head->prev, head);
+}
+
+static inline void
+__MONO_INST_LIST_DEL (MonoInstList *prev, MonoInstList *next)
+{
+	next->prev = prev;
+	prev->next = next;
+}
+
+static inline void
+__MONO_INST_LIST_SPLICE (MonoInstList *list, MonoInstList *head)
+{
+	MonoInstList *first = list->next;
+	MonoInstList *last = list->prev;
+	MonoInstList *at = head->next;
+
+	first->prev = head;
+	head->next = first;
+
+	last->next = at;
+	at->prev = last;
+}
+
+static inline void
+MONO_INST_LIST_SPLICE (MonoInstList *list, MonoInstList *head) 
+{
+	if (!MONO_INST_LIST_EMPTY (list))
+		__MONO_INST_LIST_SPLICE (list, head);
+}
+
+static inline void
+MONO_INST_LIST_SPLICE_TAIL (MonoInstList *list, MonoInstList *head) 
+{
+	if (!MONO_INST_LIST_EMPTY (list))
+		__MONO_INST_LIST_SPLICE (list, head->prev);
+}
+
+static inline void
+MONO_INST_LIST_SPLICE_INIT (MonoInstList *list, MonoInstList *head)
+{
+	if (!MONO_INST_LIST_EMPTY (list)) {
+		__MONO_INST_LIST_SPLICE (list, head);
+		MONO_INST_LIST_INIT (list);
+	}
+}
+
+static inline void
+MONO_INST_LIST_SPLICE_TAIL_INIT (MonoInstList *list, MonoInstList *head)
+{
+	if (!MONO_INST_LIST_EMPTY (list)) {
+		__MONO_INST_LIST_SPLICE (list, head->prev);
+		MONO_INST_LIST_INIT (list);
+	}
+}
+
+/*#define mono_container_of(ptr, type, member) ({			\
+	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
+	(type *)( (char *)__mptr - offsetof(type,member) );})
+
+#define MONO_INST_LIST_ENTRY(ptr, type, member) \
+	mono_container_of(ptr, type, member)*/
+
+#define MONO_INST_LIST_ENTRY(ptr, type, member)	\
+	((type *)(gpointer)(ptr))
+
+#define MONO_INST_LIST_FIRST_ENTRY(ptr, type, member) \
+	MONO_INST_LIST_ENTRY((ptr)->next, type, member)
+
+#define MONO_INST_LIST_LAST_ENTRY(ptr, type, member) \
+	MONO_INST_LIST_ENTRY((ptr)->prev, type, member)
+
+#define MONO_INST_LIST_FOR_EACH(pos, head) \
+	for (pos = (head)->next; pos != (head); pos = pos->next)
+
+#define MONO_INST_LIST_FOR_EACH_PREV(pos, head) \
+	for (pos = (head)->prev; pos != (head); pos = pos->prev)
+
+#define MONO_INST_LIST_FOR_EACH_SAFE(pos, n, head) \
+	for (pos = (head)->next, n = pos->next; pos != (head); \
+		pos = n, n = pos->next)
+
+#define MONO_INST_LIST_FOR_EACH_PREV_SAFE(pos, n, head) \
+	for (pos = (head)->prev, n = pos->prev; pos != (head); \
+		pos = n, n = pos->prev)
+
+#define MONO_INST_LIST_FOR_EACH_ENTRY(pos, head, member) \
+	for (pos = MONO_INST_LIST_ENTRY ((head)->next, MonoInst, member);\
+	     &pos->member != (head);\
+	     pos = MONO_INST_LIST_ENTRY (pos->member.next, MonoInst, member))
+
+#define MONO_INST_LIST_FOR_EACH_ENTRY_REVERSE(pos, head, member) \
+	for (pos = MONO_INST_LIST_ENTRY ((head)->prev, MonoInst, member);\
+	     &pos->member != (head);\
+	     pos = MONO_INST_LIST_ENTRY (pos->member.prev, MonoInst, member))
+
+#define MONO_INST_LIST_FOR_EACH_ENTRY_SAFE(pos, n, head, member) \
+	for (pos = MONO_INST_LIST_ENTRY ((head)->next, MonoInst, member),\
+		n = MONO_INST_LIST_ENTRY (pos->member.next, MonoInst, member);\
+	     &pos->member != (head); 					\
+	     pos = n, n = MONO_INST_LIST_ENTRY (n->member.next, MonoInst, member))
+
+#define MONO_BB_FOR_EACH_INS(bb, ins) MONO_INST_LIST_FOR_EACH_ENTRY ((ins), &((bb)->ins_list), node)
+
+#define MONO_BB_FOR_EACH_INS_SAFE(bb, next, ins) MONO_INST_LIST_FOR_EACH_ENTRY_SAFE ((ins), (next), &((bb)->ins_list), node)
+
+#define MONO_BB_FOR_EACH_INS_REVERSE(bb, ins) MONO_INST_LIST_FOR_EACH_ENTRY_REVERSE ((ins), &((bb)->ins_list), node)
 
 struct MonoEdge {
 	MonoEdge *next;
@@ -138,23 +308,9 @@ struct MonoEdge {
 };
 
 struct MonoSpillInfo {
-#ifndef MONO_ARCH_HAS_XP_REGALLOC
-	MonoSpillInfo *next;
-#endif
 	int offset;
 };
 
-/*
- * This structure contains the information maintained by the verifier for each CIL
- * stack slot. This information is also available in MonoInst, but the verifier needs to
- * update the type information during stack merges, which could lead to problems if done
- * on MonoInsts, so we use a dedicated structure instead.
- */
-typedef struct {
-	int type;
-	MonoClass *klass;
-} MonoStackSlot;
-
 /*
  * The IR-level basic block.  
  *
@@ -166,19 +322,11 @@ typedef struct {
  * at the beginning of the block, never in the middle.
  */
 struct MonoBasicBlock {
-	MonoInst *last_ins;
+	MonoInstList ins_list;
 
 	/* the next basic block in the order it appears in IL */
 	MonoBasicBlock *next_bb;
 
-	/*
-	 * Before instruction selection it is the first tree in the
-	 * forest and the first item in the list of trees. After
-	 * instruction selection it is the first instruction and the
-	 * first item in the list of instructions.
-	 */
-	MonoInst *code;
-
 	/* unique block number identification */
 	gint32 block_num;
 	
@@ -238,14 +386,13 @@ struct MonoBasicBlock {
 	guint16 out_scount, in_scount;
 	MonoInst **out_stack;
 	MonoInst **in_stack;
-	MonoStackSlot *stack_state; /* Verification stack state on enter to bblock */
 
-	/* we use that to prevent merging of bblock covered by different clauses*/
+	/* we use that to prevent merging of bblocks covered by different clauses*/
 	guint real_offset;
 
 	/*
 	 * The region encodes whether the basic block is inside
-	 * a finally, catch, filter or none of thoese.
+	 * a finally, catch, filter or none of these.
 	 *
 	 * If the value is -1, then it is neither finally, catch nor filter
 	 *
@@ -271,7 +418,13 @@ enum {
 	BB_EXCEPTION_HANDLER  = 1 << 4
 };
 
+typedef struct MonoMemcpyArgs {
+	int size, align;
+} MonoMemcpyArgs;
+
 struct MonoInst {
+	MonoInstList node; /* this must be the first field in this struct */
+
 	union {
 		union {
 			MonoInst *src;
@@ -302,32 +455,75 @@ struct MonoInst {
 	union {
 		gint32 reg3;
 		gint32 arg_info;
-		gint32 size; /* in OP_MEMSET and OP_MEMCPY */
+		gint32 size;
+		MonoMemcpyArgs *memcpy_args; /* in OP_MEMSET and OP_MEMCPY */
 		gint shift_amount;
 		gboolean is_pinvoke; /* for variables in the unmanaged marshal format */
+		gboolean record_cast_details; /* For CEE_CASTCLASS */
 		gpointer data;
 	} backend;
 	
-	MonoInst *next;
 	MonoClass *klass;
 	const unsigned char* cil_code; /* for debugging and bblock splitting */
 };
 	
+static inline void
+MONO_ADD_INS (MonoBasicBlock *bb, MonoInst *inst)
+{
+	MONO_INST_LIST_ADD_TAIL (&inst->node, &bb->ins_list);
+}
+
+static inline void
+MONO_DELETE_INS (MonoBasicBlock *bb, MonoInst *inst)
+{
+	__MONO_INST_LIST_DEL (inst->node.prev, inst->node.next);
+}
+
+static inline MonoInst *
+mono_inst_list_first (MonoInstList *head)
+{
+	if (MONO_INST_LIST_EMPTY (head))
+		return NULL;
+	return MONO_INST_LIST_FIRST_ENTRY (head, MonoInst, node);
+}
+
+static inline MonoInst *
+mono_inst_list_last (MonoInstList *head)
+{
+	if (MONO_INST_LIST_EMPTY (head))
+		return NULL;
+	return MONO_INST_LIST_LAST_ENTRY (head, MonoInst, node);
+}
+
+static inline MonoInst *
+mono_inst_list_next (MonoInstList *ins, MonoInstList *head)
+{
+	if (ins->next == head)
+		return NULL;
+	return MONO_INST_LIST_ENTRY (ins->next, MonoInst, node);
+}
+
+static inline MonoInst *
+mono_inst_list_prev (MonoInstList *ins, MonoInstList *head)
+{
+	if (ins->prev == head)
+		return NULL;
+	return MONO_INST_LIST_ENTRY (ins->prev, MonoInst, node);
+}
+
 struct MonoCallInst {
 	MonoInst inst;
 	MonoMethodSignature *signature;
 	MonoMethod *method;
 	MonoInst **args;
-	MonoInst *out_args;
+	MonoInstList out_args;
 	gconstpointer fptr;
 	guint stack_usage;
 	gboolean virtual;
 	regmask_t used_iregs;
 	regmask_t used_fregs;
-#if defined(MONO_ARCH_HAS_XP_LOCAL_REGALLOC)
 	GSList *out_ireg_args;
 	GSList *out_freg_args;
-#endif
 };
 
 struct MonoCallArgParm {
@@ -444,7 +640,11 @@ typedef struct {
 	MonoLMF          *first_lmf;
 	gpointer         signal_stack;
 	guint32          signal_stack_size;
+	gpointer         stack_ovf_guard_base;
+	guint32          stack_ovf_guard_size;
 	void            (*abort_func) (MonoObject *object);
+	/* Used to implement --debug=casts */
+	MonoClass       *class_cast_from, *class_cast_to;
 } MonoJitTlsData;
 
 typedef enum {
@@ -500,9 +700,11 @@ struct MonoJumpInfo {
 };
 
 typedef enum {
-	MONO_TRAMPOLINE_GENERIC,
+	MONO_TRAMPOLINE_JIT,
 	MONO_TRAMPOLINE_JUMP,
 	MONO_TRAMPOLINE_CLASS_INIT,
+	MONO_TRAMPOLINE_GENERIC_CLASS_INIT,
+	MONO_TRAMPOLINE_RGCTX_LAZY_FETCH,
 	MONO_TRAMPOLINE_AOT,
 	MONO_TRAMPOLINE_AOT_PLT,
 	MONO_TRAMPOLINE_DELEGATE,
@@ -532,13 +734,13 @@ typedef struct {
 	MonoMethod      *method;
 	MonoMemPool     *mempool;
 	MonoInst       **varinfo;
-	MonoMethodVar  **vars;
+	MonoMethodVar   *vars;
 	MonoInst        *ret;
 	MonoBasicBlock  *bb_entry;
 	MonoBasicBlock  *bb_exit;
 	MonoBasicBlock  *bb_init;
 	MonoBasicBlock **bblocks;
-	GHashTable      *bb_hash;
+	MonoBasicBlock **cil_offset_to_bb;
 	MonoMemPool     *state_pool; /* used by instruction selection */
 	MonoBasicBlock  *cbb;        /* used by instruction selection */
 	MonoInst        *prev_ins;   /* in decompose */
@@ -551,6 +753,7 @@ typedef struct {
 	guint            varinfo_count; /* total storage in varinfo */
 	gint             stack_offset;
 	gint             max_ireg;
+	gint             cil_offset_to_bb_len;
 	MonoRegState    *rs;
 	MonoSpillInfo   *spill_info; /* machine register spills */
 	MonoSpillInfo   *spill_info_float; /* fp register spills */
@@ -560,6 +763,20 @@ typedef struct {
 	MonoMethod      *inlined_method; /* the method which is currently inlined */
 	MonoInst        *domainvar; /* a cache for the current domain */
 	MonoInst        *got_var; /* Global Offset Table variable */
+	MonoInst	*rgctx_var; /* Runtime generic context variable (for static generic methods) */
+	MonoInst        **args;
+
+	/* 
+	 * This variable represents the hidden argument holding the vtype
+	 * return address. If the method returns something other than a vtype, or
+	 * the vtype is returned in registers this is NULL.
+	 */
+	MonoInst        *vret_addr;
+
+	/*
+	 * This is used to initialize the cil_code field of MonoInst's.
+	 */
+	const unsigned char *ip;
 	
 	struct MonoAliasingInformation *aliasing_info;
 
@@ -575,6 +792,9 @@ typedef struct {
 	
 	MonoDomain      *domain;
 
+	MonoGenericSharingContext *generic_sharing_context;
+
+	unsigned char   *cil_start;
 	unsigned char   *native_code;
 	guint            code_size;
 	guint            code_len;
@@ -597,13 +817,17 @@ typedef struct {
 	guint            compile_aot : 1;
 	guint            got_var_allocated : 1;
 	guint            ret_var_is_local : 1;
-	guint            dont_verify_stack_merge : 1;
 	guint            unverifiable : 1;
 	guint            skip_visibility : 1;
+	guint            disable_reuse_registers : 1;
+	guint            disable_reuse_stack_slots : 1;
+	guint            disable_initlocals_opt : 1;
+	guint            disable_omit_fp : 1;
+	guint            has_got_slots : 1;
 	gpointer         debug_info;
 	guint32          lmf_offset;
-	guint16          *intvars;
 	MonoProfileCoverageInfo *coverage_info;
+	GHashTable       *token_info_hash;
 	MonoCompileArch  arch;
 	guint32          exception_type;	/* MONO_EXCEPTION_* */
 	guint32          exception_data;
@@ -613,9 +837,7 @@ typedef struct {
 	/* Fields used by the local reg allocator */
 	void*            reginfo;
 	void*            reginfof;
-	void*            reverse_inst_list;
 	int              reginfo_len, reginfof_len;
-	int              reverse_inst_list_len;
 } MonoCompile;
 
 typedef enum {
@@ -652,8 +874,9 @@ typedef struct {
 	gulong cas_linkdemand_aptc;
 	gulong cas_linkdemand;
 	gulong cas_demand_generation;
-	MonoMethod *max_ratio_method;
-	MonoMethod *biggest_method;
+	gulong generic_virtual_invocations;
+	char *max_ratio_method;
+	char *biggest_method;
 	gboolean enabled;
 } MonoJitStats;
 
@@ -672,13 +895,6 @@ enum {
 	MONO_SSA_LOAD|MONO_SSA_STORE|MONO_SSA_ADDRESS_TAKEN
 };
 
-#define OP_CEQ    (256+CEE_CEQ)
-#define OP_CLT    (256+CEE_CLT)
-#define OP_CLT_UN (256+CEE_CLT_UN)
-#define OP_CGT    (256+CEE_CGT)
-#define OP_CGT_UN (256+CEE_CGT_UN)
-#define OP_LOCALLOC (256+CEE_LOCALLOC)
-
 /* opcodes: value assigned after all the CIL opcodes */
 #ifdef MINI_OP
 #undef MINI_OP
@@ -695,6 +911,9 @@ enum {
 #define OP_PCONST OP_I8CONST
 #define OP_PADD OP_LADD
 #define OP_PNEG OP_LNEG
+#define OP_PCONV_TO_I1 OP_LCONV_TO_I1
+#define OP_PCONV_TO_U1 OP_LCONV_TO_U1
+#define OP_PCONV_TO_I2 OP_LCONV_TO_I2
 #define OP_PCONV_TO_U2 OP_LCONV_TO_U2
 #define OP_PCONV_TO_OVF_I1_UN OP_LCONV_TO_OVF_I1_UN
 #define OP_PCONV_TO_OVF_I1 OP_LCONV_TO_OVF_I1
@@ -705,6 +924,9 @@ enum {
 #define OP_PCONST OP_ICONST
 #define OP_PADD CEE_ADD
 #define OP_PNEG CEE_NEG
+#define OP_PCONV_TO_I1 OP_ICONV_TO_I1
+#define OP_PCONV_TO_U1 OP_ICONV_TO_U1
+#define OP_PCONV_TO_I2 OP_ICONV_TO_I2
 #define OP_PCONV_TO_U2 CEE_CONV_U2
 #define OP_PCONV_TO_OVF_I1_UN CEE_CONV_OVF_I1_UN
 #define OP_PCONV_TO_OVF_I1 CEE_CONV_OVF_I1
@@ -779,6 +1001,9 @@ typedef struct {
 	gboolean keep_delegates;
 	gboolean collect_pagefault_stats;
 	gboolean break_on_unverified;
+	gboolean better_cast_details;
+	gboolean mdb_optimizations;
+	gboolean no_gdb_backtrace;
 } MonoDebugOptions;
 
 enum {
@@ -787,6 +1012,25 @@ enum {
 	BRANCH_UNDEF
 };
 
+typedef enum {
+	CMP_EQ,
+	CMP_NE,
+	CMP_LE,
+	CMP_GE,
+	CMP_LT,
+	CMP_GT,
+	CMP_LE_UN,
+	CMP_GE_UN,
+	CMP_LT_UN,
+	CMP_GT_UN
+} CompRelation;
+
+typedef enum {
+	CMP_TYPE_L,
+	CMP_TYPE_I,
+	CMP_TYPE_F
+} CompType;
+
 /* Implicit exceptions */
 enum {
 	MONO_EXC_INDEX_OUT_OF_RANGE,
@@ -799,6 +1043,12 @@ enum {
 	MONO_EXC_INTRINS_NUM
 };
 
+enum {
+	MINI_TOKEN_SOURCE_CLASS,
+	MINI_TOKEN_SOURCE_METHOD,
+	MINI_TOKEN_SOURCE_FIELD
+};
+
 typedef void (*MonoInstFunc) (MonoInst *tree, gpointer data);
 
 /* main function */
@@ -806,6 +1056,7 @@ int         mono_main                      (int argc, char* argv[]);
 void        mono_set_defaults              (int verbose_level, guint32 opts);
 MonoDomain* mini_init                      (const char *filename, const char *runtime_version) MONO_INTERNAL;
 void        mini_cleanup                   (MonoDomain *domain) MONO_INTERNAL;
+MonoDebugOptions *mini_get_debug_options   (void) MONO_INTERNAL;
 
 /* helper methods */
 MonoJumpInfoToken * mono_jump_info_token_new (MonoMemPool *mp, MonoImage *image, guint32 token) MONO_INTERNAL;
@@ -838,13 +1089,17 @@ void      mono_remove_patch_info            (MonoCompile *cfg, int ip) MONO_INTE
 MonoJumpInfo* mono_patch_info_dup_mp        (MonoMemPool *mp, MonoJumpInfo *patch_info) MONO_INTERNAL;
 guint     mono_patch_info_hash (gconstpointer data) MONO_INTERNAL;
 gint      mono_patch_info_equal (gconstpointer ka, gconstpointer kb) MONO_INTERNAL;
+MonoJumpInfo *mono_patch_info_list_prepend  (MonoJumpInfo *list, int ip, MonoJumpInfoType type, gconstpointer target) MONO_INTERNAL;
 gpointer  mono_resolve_patch_target         (MonoMethod *method, MonoDomain *domain, guint8 *code, MonoJumpInfo *patch_info, gboolean run_cctors) MONO_INTERNAL;
+gpointer  mono_jit_find_compiled_method     (MonoDomain *domain, MonoMethod *method) MONO_INTERNAL;
 MonoLMF * mono_get_lmf                      (void) MONO_INTERNAL;
 MonoLMF** mono_get_lmf_addr                 (void) MONO_INTERNAL;
 void      mono_jit_thread_attach            (MonoDomain *domain);
 guint32   mono_get_jit_tls_key              (void) MONO_INTERNAL;
+gint32    mono_get_jit_tls_offset           (void) MONO_INTERNAL;
 gint32    mono_get_lmf_tls_offset           (void) MONO_INTERNAL;
 gint32    mono_get_lmf_addr_tls_offset      (void) MONO_INTERNAL;
+MonoInst* mono_get_jit_tls_intrinsic        (MonoCompile *cfg) MONO_INTERNAL;
 GList    *mono_varlist_insert_sorted        (MonoCompile *cfg, GList *list, MonoMethodVar *mv, gboolean sort_end) MONO_INTERNAL;
 GList    *mono_varlist_sort                 (MonoCompile *cfg, GList *list, int sort_type) MONO_INTERNAL;
 void      mono_analyze_liveness             (MonoCompile *cfg) MONO_INTERNAL;
@@ -855,15 +1110,17 @@ int       mono_compile_assembly             (MonoAssembly *ass, guint32 opts, co
 MonoCompile *mini_method_compile            (MonoMethod *method, guint32 opts, MonoDomain *domain, gboolean run_cctors, gboolean compile_aot, int parts) MONO_INTERNAL;
 void      mono_destroy_compile              (MonoCompile *cfg) MONO_INTERNAL;
 MonoJitICallInfo *mono_find_jit_opcode_emulation (int opcode) MONO_INTERNAL;
-void	  mono_print_ins (int i, MonoInst *ins) MONO_INTERNAL;
+void	  mono_print_ins_index (int i, MonoInst *ins) MONO_INTERNAL;
+void	  mono_print_ins (MonoInst *ins) MONO_INTERNAL;
 
+/* AOT */
 void      mono_aot_init                     (void) MONO_INTERNAL;
 gpointer  mono_aot_get_method               (MonoDomain *domain,
 											 MonoMethod *method) MONO_INTERNAL;
 gpointer  mono_aot_get_method_from_token    (MonoDomain *domain, MonoImage *image, guint32 token) MONO_INTERNAL;
 gboolean  mono_aot_is_got_entry             (guint8 *code, guint8 *addr) MONO_INTERNAL;
 guint8*   mono_aot_get_plt_entry            (guint8 *code) MONO_INTERNAL;
-gboolean  mono_aot_init_vtable              (MonoVTable *vtable) MONO_INTERNAL;
+guint32   mono_aot_get_plt_info_offset      (gssize *regs, guint8 *code) MONO_INTERNAL;
 gboolean  mono_aot_get_cached_class_info    (MonoClass *klass, MonoCachedClassInfo *res) MONO_INTERNAL;
 gboolean  mono_aot_get_class_from_name      (MonoImage *image, const char *name_space, const char *name, MonoClass **klass) MONO_INTERNAL;
 MonoJitInfo* mono_aot_find_jit_info         (MonoDomain *domain, MonoImage *image, gpointer addr) MONO_INTERNAL;
@@ -872,6 +1129,14 @@ gboolean mono_aot_is_pagefault              (void *ptr) MONO_INTERNAL;
 void mono_aot_handle_pagefault              (void *ptr) MONO_INTERNAL;
 guint32 mono_aot_get_n_pagefaults           (void) MONO_INTERNAL;
 gpointer mono_aot_plt_resolve               (gpointer aot_module, guint32 plt_info_offset, guint8 *code) MONO_INTERNAL;
+gpointer mono_aot_get_method_from_vt_slot   (MonoDomain *domain, MonoVTable *vtable, int slot) MONO_INTERNAL;
+gpointer mono_aot_create_specific_trampoline   (MonoImage *image, gpointer arg1, MonoTrampolineType tramp_type, MonoDomain *domain, guint32 *code_len) MONO_INTERNAL;
+gpointer mono_aot_get_named_code            (const char *name) MONO_INTERNAL;
+gpointer mono_aot_get_unbox_trampoline      (MonoMethod *method) MONO_INTERNAL;
+/* This is an exported function */
+void     mono_aot_register_globals          (gpointer *globals);
+/* This too */
+void     mono_aot_register_module           (gpointer *aot_info);
 
 gboolean  mono_method_blittable             (MonoMethod *method) MONO_INTERNAL;
 gboolean  mono_method_same_domain           (MonoJitInfo *caller, MonoJitInfo *callee) MONO_INTERNAL;
@@ -888,23 +1153,30 @@ MonoJitICallInfo *mono_find_jit_icall_by_addr  (gconstpointer addr) MONO_INTERNA
 MonoJitICallInfo *mono_register_jit_icall      (gconstpointer func, const char *name, MonoMethodSignature *sig, gboolean is_save) MONO_INTERNAL;
 gconstpointer     mono_icall_get_wrapper       (MonoJitICallInfo* callinfo) MONO_INTERNAL;
 
+void              mono_trampolines_init (void) MONO_INTERNAL;
+void              mono_trampolines_cleanup (void) MONO_INTERNAL;
 guint8 *          mono_get_trampoline_code (MonoTrampolineType tramp_type) MONO_INTERNAL;
+gpointer          mono_create_specific_trampoline (gpointer arg1, MonoTrampolineType tramp_type, MonoDomain *domain, guint32 *code_len) MONO_INTERNAL;
 gpointer          mono_create_jump_trampoline (MonoDomain *domain, 
 											   MonoMethod *method, 
 											   gboolean add_sync_wrapper) MONO_INTERNAL;
 gpointer          mono_create_class_init_trampoline (MonoVTable *vtable) MONO_INTERNAL;
 gpointer          mono_create_jit_trampoline (MonoMethod *method) MONO_INTERNAL;
 gpointer          mono_create_jit_trampoline_from_token (MonoImage *image, guint32 token) MONO_INTERNAL;
+gpointer          mono_create_jit_trampoline_in_domain (MonoDomain *domain, MonoMethod *method, gboolean add_sync_wrapper) MONO_INTERNAL;
+gpointer          mono_create_delegate_trampoline (MonoClass *klass) MONO_INTERNAL;
 MonoVTable*       mono_find_class_init_trampoline_by_addr (gconstpointer addr) MONO_INTERNAL;
+MonoClass*        mono_find_delegate_trampoline_by_addr (gconstpointer addr) MONO_INTERNAL;
 gpointer          mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp) MONO_INTERNAL;
-gpointer          mono_delegate_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp) MONO_INTERNAL;
+gpointer          mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guint8* tramp) MONO_INTERNAL;
 gpointer          mono_aot_trampoline (gssize *regs, guint8 *code, guint8 *token_info, 
 									   guint8* tramp) MONO_INTERNAL;
 gpointer          mono_aot_plt_trampoline (gssize *regs, guint8 *code, guint8 *token_info, 
 										   guint8* tramp) MONO_INTERNAL;
 void              mono_class_init_trampoline (gssize *regs, guint8 *code, MonoVTable *vtable, guint8 *tramp) MONO_INTERNAL;
-gpointer          mono_debugger_create_notification_function (void) MONO_INTERNAL;
-
+void              mono_generic_class_init_trampoline (gssize *regs, guint8 *code, MonoVTable *vtable, guint8 *tramp) MONO_INTERNAL;
+gconstpointer     mono_get_trampoline_func (MonoTrampolineType tramp_type);
+gpointer          mini_get_vtable_trampoline (void) MONO_INTERNAL;
 
 gboolean          mono_running_on_valgrind (void) MONO_INTERNAL;
 void*             mono_global_codeman_reserve (int size) MONO_INTERNAL;
@@ -914,8 +1186,15 @@ gint32*           mono_allocate_stack_slots (MonoCompile *cfg, guint32 *stack_si
 void              mono_local_regalloc (MonoCompile *cfg, MonoBasicBlock *bb) MONO_INTERNAL;
 MonoInst         *mono_branch_optimize_exception_target (MonoCompile *cfg, MonoBasicBlock *bb, const char * exname) MONO_INTERNAL;
 gboolean          mono_is_regsize_var (MonoType *t) MONO_INTERNAL;
+CompRelation      mono_opcode_to_cond (int opcode) MONO_INTERNAL;
+CompType          mono_opcode_to_type (int opcode, int cmp_opcode) MONO_INTERNAL;
+int               mono_op_imm_to_op (int opcode) MONO_INTERNAL;
+void              mono_decompose_op_imm (MonoCompile *cfg, MonoInst *ins) MONO_INTERNAL;
+void              mono_peephole_ins (MonoBasicBlock *bb, MonoInst *ins) MONO_INTERNAL;
 
 /* methods that must be provided by the arch-specific port */
+void      mono_arch_init                        (void) MONO_INTERNAL;
+void      mono_arch_cleanup                     (void) MONO_INTERNAL;
 void      mono_arch_cpu_init                    (void) MONO_INTERNAL;
 guint32   mono_arch_cpu_optimizazions           (guint32 *exclude_mask) MONO_INTERNAL;
 void      mono_arch_instrument_mem_needs        (MonoMethod *method, int *stack, int *code) MONO_INTERNAL;
@@ -932,9 +1211,10 @@ gpointer  mono_arch_get_rethrow_exception       (void) MONO_INTERNAL;
 gpointer  mono_arch_get_throw_exception_by_name (void) MONO_INTERNAL;
 gpointer  mono_arch_get_throw_corlib_exception  (void) MONO_INTERNAL;
 guchar*   mono_arch_create_trampoline_code      (MonoTrampolineType tramp_type) MONO_INTERNAL;
-gpointer  mono_arch_create_jit_trampoline       (MonoMethod *method) MONO_INTERNAL;
-MonoJitInfo *mono_arch_create_jump_trampoline      (MonoMethod *method) MONO_INTERNAL;
-gpointer  mono_arch_create_class_init_trampoline(MonoVTable *vtable) MONO_INTERNAL;
+guchar*   mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
+gpointer  mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot) MONO_INTERNAL;
+guint32	  mono_arch_get_rgctx_lazy_fetch_offset (gpointer *regs) MONO_INTERNAL;
+gpointer  mono_arch_get_nullified_class_init_trampoline (guint32 *code_len) MONO_INTERNAL;
 GList    *mono_arch_get_allocatable_int_vars    (MonoCompile *cfg) MONO_INTERNAL;
 GList    *mono_arch_get_global_int_regs         (MonoCompile *cfg) MONO_INTERNAL;
 guint32   mono_arch_regalloc_cost               (MonoCompile *cfg, MonoMethodVar *vmv) MONO_INTERNAL;
@@ -944,7 +1224,9 @@ int       mono_arch_max_epilog_size             (MonoCompile *cfg) MONO_INTERNAL
 guint8   *mono_arch_emit_prolog                 (MonoCompile *cfg) MONO_INTERNAL;
 void      mono_arch_emit_epilog                 (MonoCompile *cfg) MONO_INTERNAL;
 void      mono_arch_emit_exceptions             (MonoCompile *cfg) MONO_INTERNAL;
-void      mono_arch_local_regalloc              (MonoCompile *cfg, MonoBasicBlock *bb) MONO_INTERNAL;
+void      mono_arch_lowering_pass               (MonoCompile *cfg, MonoBasicBlock *bb) MONO_INTERNAL;
+void      mono_arch_peephole_pass_1             (MonoCompile *cfg, MonoBasicBlock *bb) MONO_INTERNAL;
+void      mono_arch_peephole_pass_2             (MonoCompile *cfg, MonoBasicBlock *bb) MONO_INTERNAL;
 void      mono_arch_output_basic_block          (MonoCompile *cfg, MonoBasicBlock *bb) MONO_INTERNAL;
 gboolean  mono_arch_has_unwind_info             (gconstpointer addr) MONO_INTERNAL;
 void      mono_arch_setup_jit_tls_data          (MonoJitTlsData *tls) MONO_INTERNAL;
@@ -965,10 +1247,18 @@ MonoJitInfo *mono_arch_find_jit_info            (MonoDomain *domain,
 						 gboolean *managed) MONO_INTERNAL;
 gpointer mono_arch_get_call_filter              (void) MONO_INTERNAL;
 gpointer mono_arch_get_restore_context          (void) MONO_INTERNAL;
+gpointer mono_arch_get_call_filter_full         (guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
+gpointer mono_arch_get_restore_context_full     (guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
+gpointer  mono_arch_get_throw_exception_full    (guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
+gpointer  mono_arch_get_rethrow_exception_full  (guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
+gpointer  mono_arch_get_throw_exception_by_name_full (guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
+gpointer  mono_arch_get_throw_corlib_exception_full (guint32 *code_size, MonoJumpInfo **ji, gboolean aot) MONO_INTERNAL;
 gboolean mono_arch_handle_exception             (void *sigctx, gpointer obj, gboolean test_only) MONO_INTERNAL;
+void     mono_arch_handle_altstack_exception    (void *sigctx, gpointer fault_addr, gboolean stack_ovf) MONO_INTERNAL;
 gpointer mono_arch_ip_from_context              (void *sigctx) MONO_INTERNAL;
 void     mono_arch_sigctx_to_monoctx            (void *sigctx, MonoContext *ctx) MONO_INTERNAL;
 void     mono_arch_monoctx_to_sigctx            (MonoContext *mctx, void *ctx) MONO_INTERNAL;
+gpointer mono_arch_context_get_int_reg		(MonoContext *ctx, int reg) MONO_INTERNAL;
 void     mono_arch_flush_register_windows       (void) MONO_INTERNAL;
 gboolean mono_arch_is_inst_imm                  (gint64 imm) MONO_INTERNAL;
 MonoInst* mono_arch_get_domain_intrinsic        (MonoCompile* cfg) MONO_INTERNAL;
@@ -977,19 +1267,30 @@ gboolean mono_arch_is_int_overflow              (void *sigctx, void *info) MONO_
 void     mono_arch_invalidate_method            (MonoJitInfo *ji, void *func, gpointer func_arg) MONO_INTERNAL;
 guint32  mono_arch_get_patch_offset             (guint8 *code) MONO_INTERNAL;
 gpointer*mono_arch_get_vcall_slot_addr          (guint8* code, gpointer *regs) MONO_INTERNAL;
+gpointer mono_arch_get_vcall_slot               (guint8 *code, gpointer *regs, int *displacement) MONO_INTERNAL;
 gpointer*mono_arch_get_delegate_method_ptr_addr (guint8* code, gpointer *regs) MONO_INTERNAL;
 void     mono_arch_create_vars                  (MonoCompile *cfg) MONO_INTERNAL;
 void     mono_arch_save_unwind_info             (MonoCompile *cfg) MONO_INTERNAL;
 void     mono_arch_register_lowlevel_calls      (void) MONO_INTERNAL;
-gpointer mono_arch_get_unbox_trampoline         (MonoMethod *m, gpointer addr) MONO_INTERNAL;
-void     mono_arch_patch_callsite               (guint8 *code, guint8 *addr) MONO_INTERNAL;
+gpointer mono_arch_get_unbox_trampoline         (MonoGenericSharingContext *gsctx, MonoMethod *m, gpointer addr) MONO_INTERNAL;
+void     mono_arch_patch_callsite               (guint8 *method_start, guint8 *code, guint8 *addr) MONO_INTERNAL;
 void     mono_arch_patch_plt_entry              (guint8 *code, guint8 *addr) MONO_INTERNAL;
 void     mono_arch_nullify_class_init_trampoline(guint8 *code, gssize *regs) MONO_INTERNAL;
 void     mono_arch_nullify_plt_entry            (guint8 *code) MONO_INTERNAL;
-void     mono_arch_patch_delegate_trampoline    (guint8 *code, guint8 *tramp, gssize *regs, guint8 *addr) MONO_INTERNAL;
+int      mono_arch_get_this_arg_reg             (MonoMethodSignature *sig, MonoGenericSharingContext *gsctx, guint8 *code) MONO_INTERNAL;
+gpointer mono_arch_get_this_arg_from_call       (MonoGenericSharingContext *gsctx, MonoMethodSignature *sig, gssize *regs, guint8 *code) MONO_INTERNAL;
+MonoObject* mono_arch_find_this_argument        (gpointer *regs, MonoMethod *method, MonoGenericSharingContext *gsctx) MONO_INTERNAL;
+gpointer mono_arch_get_delegate_invoke_impl     (MonoMethodSignature *sig, gboolean has_target) MONO_INTERNAL;
 gpointer mono_arch_create_specific_trampoline   (gpointer arg1, MonoTrampolineType tramp_type, MonoDomain *domain, guint32 *code_len) MONO_INTERNAL;
+void        mono_arch_emit_imt_argument         (MonoCompile *cfg, MonoCallInst *call) MONO_INTERNAL;
+MonoMethod* mono_arch_find_imt_method           (gpointer *regs, guint8 *code) MONO_INTERNAL;
+MonoVTable* mono_arch_find_static_call_vtable   (gpointer *regs, guint8 *code) MONO_INTERNAL;
+gpointer    mono_arch_build_imt_thunk           (MonoVTable *vtable, MonoDomain *domain, MonoIMTCheckItem **imt_entries, int count) MONO_INTERNAL;
+void    mono_arch_notify_pending_exc            (void) MONO_INTERNAL;
+void    mono_arch_fixup_jinfo                   (MonoCompile *cfg) MONO_INTERNAL;
 
 /* Exception handling */
+void     mono_exceptions_init                   (void) MONO_INTERNAL;
 gboolean mono_handle_exception                  (MonoContext *ctx, gpointer obj,
 						 gpointer original_ip, gboolean test_only) MONO_INTERNAL;
 void     mono_handle_native_sigsegv             (int signal, void *sigctx) MONO_INTERNAL;
@@ -998,6 +1299,12 @@ void     mono_jit_walk_stack                    (MonoStackWalk func, gboolean do
 void     mono_jit_walk_stack_from_ctx           (MonoStackWalk func, MonoContext *ctx, gboolean do_il_offset, gpointer user_data) MONO_INTERNAL;
 void     mono_setup_altstack                    (MonoJitTlsData *tls) MONO_INTERNAL;
 void     mono_free_altstack                     (MonoJitTlsData *tls) MONO_INTERNAL;
+gpointer mono_get_throw_exception               (void) MONO_INTERNAL;
+gpointer mono_get_rethrow_exception             (void) MONO_INTERNAL;
+gpointer mono_get_call_filter                   (void) MONO_INTERNAL;
+gpointer mono_get_restore_context               (void) MONO_INTERNAL;
+gpointer mono_get_throw_exception_by_name       (void) MONO_INTERNAL;
+gpointer mono_get_throw_corlib_exception        (void) MONO_INTERNAL;
 
 /* the new function to do stack walks */
 typedef gboolean (*MonoStackFrameWalk)          (MonoDomain *domain, MonoContext *ctx, MonoJitInfo *ji, gpointer data);
@@ -1036,13 +1343,17 @@ void      mono_debug_add_icall_wrapper          (MonoMethod *method, MonoJitICal
 void      mono_debug_print_vars                 (gpointer ip, gboolean only_arguments);
 void      mono_debugger_run_finally             (MonoContext *start_ctx);
 
+extern gssize mono_breakpoint_info_index [MONO_BREAKPOINT_ARRAY_SIZE];
+
+gboolean mono_breakpoint_clean_code (guint8 *method_start, guint8 *code, int offset, guint8 *buf, int size);
+
 /* Mono Debugger support */
 void      mono_debugger_init                    (void);
 int       mono_debugger_main                    (MonoDomain *domain, MonoAssembly *assembly, int argc, char **argv);
 
 
 /* Tracing */
-MonoTraceSpec *mono_trace_parse_options         (char *options) MONO_INTERNAL;
+MonoTraceSpec *mono_trace_parse_options         (const char *options) MONO_INTERNAL;
 void           mono_trace_set_assembly          (MonoAssembly *assembly) MONO_INTERNAL;
 gboolean       mono_trace_eval                  (MonoMethod *method) MONO_INTERNAL;
 
@@ -1057,8 +1368,39 @@ mono_local_cprop (MonoCompile *cfg) MONO_INTERNAL;
 MonoSecurityFrame* ves_icall_System_Security_SecurityFrame_GetSecurityFrame (gint32 skip) MONO_INTERNAL;
 MonoArray* ves_icall_System_Security_SecurityFrame_GetSecurityStack (gint32 skip) MONO_INTERNAL;
 
-int mini_wapi_hps     (int argc, char **argv);
-int mini_wapi_semdel  (int argc, char **argv);
-int mini_wapi_seminfo (int argc, char **argv);
+/* Generic sharing */
+
+MonoGenericSharingContext* mono_get_generic_context_from_code (guint8 *code) MONO_INTERNAL;
+
+MonoGenericContext* mini_method_get_context (MonoMethod *method) MONO_INTERNAL;
+
+int mono_method_check_context_used (MonoMethod *method) MONO_INTERNAL;
+
+gboolean mono_generic_context_equal_deep (MonoGenericContext *context1, MonoGenericContext *context2) MONO_INTERNAL;
+
+gboolean mono_generic_context_is_sharable (MonoGenericContext *context, gboolean allow_type_vars) MONO_INTERNAL;
+
+gboolean mono_method_is_generic_impl (MonoMethod *method) MONO_INTERNAL;
+gboolean mono_method_is_generic_sharable_impl (MonoMethod *method, gboolean allow_type_vars) MONO_INTERNAL;
+
+gpointer mono_helper_get_rgctx_other_ptr (MonoClass *caller_class, MonoVTable *vtable,
+					  guint32 token, guint32 token_source, guint32 rgctx_type,
+					  gint32 rgctx_index) MONO_INTERNAL;
+
+void mono_generic_sharing_init (void) MONO_INTERNAL;
+
+MonoClass* mini_class_get_container_class (MonoClass *class) MONO_INTERNAL;
+MonoGenericContext* mini_class_get_context (MonoClass *class) MONO_INTERNAL;
+
+MonoType* mini_get_basic_type_from_generic (MonoGenericSharingContext *gsctx, MonoType *type) MONO_INTERNAL;
+
+int mini_type_stack_size (MonoGenericSharingContext *gsctx, MonoType *t, int *align) MONO_INTERNAL;
+
+/* wapihandles.c */
+int mini_wapi_hps (int argc, char **argv) MONO_INTERNAL;
+
+int mini_wapi_semdel (int argc, char **argv) MONO_INTERNAL;
+
+int mini_wapi_seminfo (int argc, char **argv) MONO_INTERNAL;
 
 #endif /* __MONO_MINI_H__ */
